// memory allocator for AllocArray and other templates
//---------------------------------------------------------------------------
// holds the handle for allocated data shit (overload for spec)
// holds data to allocation
class AllocManager;

class AllocHolder
{
	public:
	AllocHolder(){mgr=NULL;valid=false,size=0,memmode=0;data=NULL;exdata=NULL;datahdl=NULL;}

	AllocManager	*mgr;	// manager which handles objects of this type
	MultithreadLock  lock;  // locking (when locked ldata is valid)

	// manager specificly used data
	bool			 valid;     // valid check
	long		 	 size;      // size in elements
	long			 memmode;	// different alloc types (Plain Mem - Spec Mem)
	void		    *data;      // valid when locked
	void			*exdata;	// more data
	void			*datahdl;    // data handle
};

//---------------------------------------------------------------------------
// If you want to offer a new data type, overload the alloc manager
class AllocManager
{
	public:
	AllocManager(){};
	virtual ~AllocManager(){};

	// creation/destruction functions
	virtual bool		 New(AllocHolder &handle,long size=-1,long istruct=0)=0;
	virtual void         Delete(AllocHolder &handle)=0;
	virtual void		*Lock(AllocHolder &handle)=0;	  // gets linear pointer to array
	virtual void		 Unlock(AllocHolder &handle)=0; 
	virtual bool		 Valid(AllocHolder &handle)=0;  // validity check
	virtual void		 Operation(AllocHolder &handle,long op1=0,long op2=0){};	// perform operation on array (optimize)

	PtrList<AllocHolder> objects;	// all objects belonging to this manager.

	void				 ObjAdd(AllocHolder *handle){objects.Add(handle);}
	void				 ObjRem(AllocHolder *handle){Delete(*handle);objects.Remove(objects.Find(handle));}
};

//---------------------------------------------------------------------------
// manages access to object of this type (smart pointer)
template <class T> 
class AllocObject
{
	public:
	AllocObject(AllocManager *mgr=NULL)
	{
		internal.mgr=mgr;
		if(mgr!=NULL){mgr->ObjAdd(&internal);}
	}
	~AllocObject()
	{
		if (internal.mgr!=NULL){internal.mgr->ObjRem(&internal);}  
	}

	void	SetManager(AllocManager *mgr)
	{
		if (internal.mgr!=NULL){internal.mgr->ObjRem(&internal);}  
		if(mgr!=NULL){mgr->ObjAdd(&internal);}
		internal.mgr=mgr;
	}

	AllocHolder &GetAllocHolder(){return internal;}

	bool New(long istruct=0)
	{
		if (internal.mgr!=NULL)
		{
			internal.mgr->New(internal,-1,istruct);
		}
		return false;
	};
	void Delete()
	{
		if (internal.mgr!=NULL)internal.mgr->Delete(internal);
	}
	
	// validity check
	bool	Valid()
	{
		return ( (internal.mgr!=NULL && internal.mgr->Valid(internal)) );
	};

	// lock data
	T*	Lock()
	{
		if (internal.mgr==NULL)return NULL;
		return((T*)internal.mgr->Lock(internal));
	}

	// unlock
	void Unlock()
	{
		if (internal.mgr!=NULL)
		internal.mgr->Unlock(internal);
	}

	// Access operator overloads
	inline T* operator->() 		
	{ 
		assert(internal.lock.IsLocked()) 
		assert(internal.mgr!=NULL && internal.data!=NULL);
		return((T*)internal.data);
	}
	inline operator	T&()			{ return o(); }
	inline operator	T*()			{ return &o(); }
	T &operator=(T &other)			{ T &t=o();t=other;return t;}

	inline T & O(){return o();}
	inline T & o()
	{
		assert(internal.lock.IsLocked());
		assert(internal.mgr!=NULL && internal.data!=NULL);
		return(*(T*)internal.data);
	}
	AllocHolder internal;
};

//---------------------------------------------------------------------------
// array of custom 'allocated' elements of type 't'

template <class T>
class AllocArray
{
	public:
	// Initialization
	AllocArray(AllocManager *mgr=NULL):Len(0)	{internal.mgr=mgr;if(mgr!=NULL){mgr->ObjAdd(&internal);}}
	~AllocArray(){if (internal.mgr!=NULL)	{internal.mgr->ObjRem(&internal);} }

	void	SetManager(AllocManager *mgr)
	{
		if (internal.mgr!=NULL){internal.mgr->ObjRem(&internal);}  
		if(mgr!=NULL){mgr->ObjAdd(&internal);}
		internal.mgr=mgr;
	}

	AllocHolder &GetAllocHolder(){return internal;}


	// Properties													
	inline bool		IsEmpty() const									{ return Len==0; }
	inline long		MaxIndex() const								{ return Len-1; }
	inline long		MemoryUsage() const								{ return Len*sizeof(T); }
	inline long		Length() const									{ return Len; }

	// Locking/unlocking
	inline T		*Lock()										    
	{
		if (internal.mgr==NULL)return NULL;
		return((T*)internal.mgr->Lock(internal));
	}

	inline void		 Unlock()		
	{
		if (internal.mgr!=NULL)
		internal.mgr->Unlock(internal);
	}

	// Simple operations
	inline void     Clear(){SetSize(0);}
	inline void		SetSize(long n,long istruct=0)									
	{ 
		if (Len!=n && internal.mgr!=NULL)
		{
			Unlock();
			internal.mgr->Delete(internal);
			if (n>0)
			{
				internal.mgr->New(internal,n,istruct);
				Len=n;
			}
			else Len=0;
		}
	}

	// Element access
	inline T&		GetElem(long pos){	return (*this)[pos]; }
	inline T&		operator[](long pos)						
	{
		bounds(pos, 0, Len-1); 
		assert(internal.lock.IsLocked()); 
		assert(internal.mgr!=NULL && internal.data!=NULL);
		return(((T*)internal.data)[pos]);
	}

	// Serialization
	inline friend class Archive& operator<<(class Archive& file, AllocArray<T>& arr) { if (arr.Lock()!=NULL){file << true; file << arr.Len; for (long i=0; i<arr.Len; i++) file << arr[i]; arr.Unlock();} else file<<false;return file;}
	inline friend class Archive& operator>>(class Archive& file, AllocArray<T>& arr) { bool a; file>>a;if (a){long len; file >> len; arr.SetSize(len); assert(arr.Lock()!=NULL);for (long i=0; i<arr.Len; i++) file >> arr[i];arr.Unlock();} return file; }

protected:
	
	// Data
	long		 Len;
	AllocHolder  internal;
};
