/*
---------------------------------------------------------------------------
|					File functions/serialization
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/

#include "Library.h"

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---	 File IO class
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------


bool FileIO::Open(cString filename,long mode)// euh?
{
	if (file!=NULL)Close();
	// reset buffer data..
	buf.size=FILEBUFFERSIZE;buf.used=0;buf.start=0;buf.readoffset=0;
	String type;
	switch(mode)
	{
		case mode_read:       type="rb"; break;
		case mode_write:      type="wb"; break;
		case mode_readwrite:  type="w+b";break;
		case mode_streamread: type="rb"; break;
		case mode_streamwrite:type="wb"; break;
	}
	Mode=mode;
	FILE *f=fopen(filename,type);	
	if (f!=NULL)
	{
		file=f;
		if ( (mode&mode_read) )ReadBuffer(); 
		Opened=true;
		return(true);
	}
	return(false);
}

//---------------------------------------------------------------------------

bool FileIO::Close()// guess
{
	if (file!=NULL)
	{
		FlushBuffer(); // flush data before closing.

		fclose ( (FILE *)file);
		file=NULL;
		Opened=false;
		return(true);
	}
	return(false);
}

//---------------------------------------------------------------------------

bool FileIO::Write(const void *t, long len)// write
{
	if (t!=NULL && file!=NULL && (Mode&mode_write) ) // writing is possible
	{
		if (len>buf.size) // block is larger then buffer
		{
			FlushBuffer(); // flushes buffer..
			fwrite(t,1,len,(FILE *)file); // write data plain to disk
			cur_position+=len;
		}
		else
		{
			// see if it fits in the buffer
			if (len+buf.used>=buf.size)
			{ // doesn't fit in buffer.. flush and write
				FlushBuffer();
				memcpy(&buf.mem[buf.used],t,len);
				buf.used=len;
				cur_position+=len;
			}
			else // copy straight to buffer
			{
				memcpy(&buf.mem[buf.used],t,len);
				buf.used+=len;
				cur_position+=len;
			}
		}
		return(true);
	}
	return(false);
}

//---------------------------------------------------------------------------

bool FileIO::WriteAt(long offset,const void *t, long len)// write at offset NOT BUFFERED
{
   // unbuffered write to file..
	if (file!=NULL && (Mode&mode_write) ) // writing is possible
	{
		if (buf.start<offset+len)FlushBuffer(); // it is still in the buffer make sure it is all written to disk..
		long curfilepos=GetFilePos();
		fseek ( (FILE *)file,offset,SEEK_SET);
		fwrite(t,1,len,(FILE *)file); // write data plain to disk
		fseek ( (FILE *)file,curfilepos,SEEK_SET);
		return(true);
	}
	return(false);
}

//---------------------------------------------------------------------------

bool FileIO::Read(void *d,long len)			
{					// reads bytes..
	char *t=(char *)d;
	if (file!=NULL && (Mode&mode_read) ) // writing is possible
	{
		if (buf.readoffset>=buf.used) // buffer is 'read out'
		{
			ReadBuffer();
		}
		if (len>buf.used-buf.readoffset) // object is half in buffer
		{
			// copy first part..
			long partlen=buf.used-buf.readoffset;
			memcpy(t,&buf.mem[buf.readoffset],partlen);
			buf.readoffset=buf.used;len-=partlen;
			// if next part is larger then buffer, copy directly else use buffer
			if (len>buf.size)
			{
				fread(&t[partlen],1,len,(FILE *)file);
				cur_position+=len;
			}
			else // use buffer
			{
				ReadBuffer();	
				memcpy(&t[partlen],buf.mem,len);
				buf.readoffset+=len;
				cur_position+=len;
			}
		}
		else // object is in buffer
		{

			memcpy(t,&buf.mem[buf.readoffset],len);
			buf.readoffset+=len;
			cur_position+=len;
		}
		return(true);
	}
	return(false);
}

//---------------------------------------------------------------------------

bool FileIO::ReadAt(long offset,void *t,long len)// Not buffered
{
	// unbuffered read from
	if (file!=NULL && (Mode&mode_read) ) // writing is possible
	{
		long curfilepos=GetFilePos();
		fseek( (FILE *)file,offset,SEEK_SET);
		fread(t,1,len,(FILE *)file); // read data plain from disk
		fseek( (FILE *)file,curfilepos,SEEK_SET);
		return(true);
	}
	return(false);
}

//---------------------------------------------------------------------------
	
bool FileIO::IsEnd()// Check for end-of-file
{
	if (file!=NULL) // writing is possible
	{
		if (buf.used==0)return(true);
		return false;
	}
	return(true);
}

//---------------------------------------------------------------------------

long FileIO::GetPos()// Return offset in block (file, mem or network)
{
	if (file!=NULL) // writing is possible
	{
		if (Mode & mode_read) // mode reading
		{
			return(buf.start+buf.readoffset);
		}
		else
		{
			return(GetFilePos()+buf.used);
		}
	}
	return(0);
}

//---------------------------------------------------------------------------

long FileIO::GetFilePos()// Return offset in block (file, mem or network)
{
	if (file!=NULL) // writing is possible
	{
		return(ftell((FILE *)file));
	}
	return(0);
}

//---------------------------------------------------------------------------

void FileIO::Seek(long new_pos)	// Seek from beginning of file
{
	if (file!=NULL) // writing is possible
	{
		FlushBuffer(); // flush whatever you where doing..(write mode)
		fseek( (FILE *)file,new_pos,SEEK_SET);
		cur_position=new_pos;
		ReadBuffer(); // reload buffer on that position.  (read mode)
	}
}

//---------------------------------------------------------------------------

long FileIO::Skip(long amount)
{
	if (file!=NULL) // writing is possible
	{
		FlushBuffer(); // flush whatever you where doing..(write mode)
		fseek( (FILE *)file,amount,SEEK_CUR);
		cur_position+=amount;
		ReadBuffer(); // reload buffer on that position.  (read mode)
		return(true);
	}
	return(false);
}

//---------------------------------------------------------------------------
// Attributes
long FileIO::GetLength()
{										// Get length of file
	long curfilepos=GetFilePos();
	fseek( (FILE *)file,0,SEEK_END);
	filesize=ftell( (FILE *)file);
	fseek( (FILE *)file,curfilepos,SEEK_SET);
	return(filesize);
}

//---------------------------------------------------------------------------
void FileIO::Flush()
{
	FlushBuffer();	
}	

//---------------------------------------------------------------------------

void FileIO::FlushBuffer()// flushes write buffer
{
	if (file!=NULL && (Mode&mode_write) && buf.used>0)
	{
		fwrite(buf.mem,1,buf.used,(FILE *)file);
		buf.start=GetFilePos();
		buf.readoffset=0;
		buf.used=0;
		cur_position=buf.start;
		// buffer written
		filesize=cur_position;
	}
}

//---------------------------------------------------------------------------

void FileIO::ReadBuffer()// Reads next block of BUFFERSIZE in buffer
{
	if (file!=NULL && (Mode&mode_read) )// read next block at current file pos
	{
		buf.start=GetFilePos();
		buf.readoffset=0;
		buf.used=fread(buf.mem,1,buf.size,(FILE *)file);
		cur_position=buf.start;
		// buffer reloaded..
	}
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---	 Archive class
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
bool Archive::OpenOld(SerialIO *i_target, short i_type,long version,bool hasheader) // save and load determines what is done.. read or write
{
	if (Opened)Close();
	HasHeader=hasheader;
	if (i_target!=NULL)
	{
		target=i_target;type=i_type;
		if (type==Load)
		{
			if (HasHeader)
			{
				HeaderOffset=target->GetPos();
				target->Read(&ArcHdr,sizeof(ArcHdr)-sizeof(ArcHdr.ID));
				if (ArcHdr.Tag[0]!='V' ||
					ArcHdr.Tag[1]!='F' ||
					ArcHdr.Tag[2]!='A' ||
					ArcHdr.Tag[3]!='R')return(false); // not right archive
				if (ArcHdr.Version!=version)return(false); // not right version
				//if (ArcHdr.ID!=ID)return(false);
			}
			Opened=true;
			return(true);
		}
		else
		{	// tag= 'VFAR'
			if (HasHeader)
			{
				HeaderOffset=target->GetPos(); 
				ArcHdr.Tag[0]='V',
				ArcHdr.Tag[1]='F',
				ArcHdr.Tag[2]='A',
				ArcHdr.Tag[3]='R';
				ArcHdr.Version=version;
				ArcHdr.ID=0;
				target->Write(&ArcHdr,sizeof(ArcHdr)-sizeof(ArcHdr.ID));
				Opened=true;
			}
			return(true);
		}
	}
	return(false);
}

//---------------------------------------------------------------------------

bool Archive::Open(SerialIO *i_target, short i_type,long version,long ID,bool hasheader) // save and load determines what is done.. read or write
{
	if (Opened)Close();
	HasHeader=hasheader;
	if (i_target!=NULL)
	{
		target=i_target;type=i_type;
		if (type==Load)
		{
			if (HasHeader)
			{
				HeaderOffset=target->GetPos();
				target->Read(&ArcHdr,sizeof(ArcHdr));
				if (ArcHdr.Tag[0]!='V' ||
					ArcHdr.Tag[1]!='F' ||
					ArcHdr.Tag[2]!='A' ||
					ArcHdr.Tag[3]!='R')return(false); // not right archive
				if (ArcHdr.Version!=version)return(false); // not right version
				if (ArcHdr.ID!=ID)return(false);
			}
			Opened=true;
			return(true);
		}
		else
		{	// tag= 'VFAR'
			if (HasHeader)
			{
				HeaderOffset=target->GetPos(); 
				ArcHdr.Tag[0]='V',
				ArcHdr.Tag[1]='F',
				ArcHdr.Tag[2]='A',
				ArcHdr.Tag[3]='R';
				ArcHdr.Version=version;
				ArcHdr.ID=ID;
				target->Write(&ArcHdr,sizeof(ArcHdr));
				Opened=true;
			}
			return(true);
		}
	}
	return(false);
}

//---------------------------------------------------------------------------

void Archive::Close() // writes file size to header (if applicable)
{
	if (Opened)
	{
		target->Flush(); // flush buffer.. and write new fileheader
		if (HasHeader)
		{
			ArcHdr.FileLength=target->GetPos()-HeaderOffset;
			target->WriteAt(HeaderOffset,&ArcHdr,sizeof(ArcHdr));
		}
		target=NULL;
	}
	Opened=false;
}

//---------------------------------------------------------------------------

void Archive::JumpTo(long offset)		// skips to offset(FORWARD ONLY, and offsetted)
{
	if (Opened)
	{
		target->Seek(HeaderOffset+offset);
	}
}

//---------------------------------------------------------------------------

void Archive::OffsetTagJump(long offset)			// jumps the offset tag in the buffer..
{
	if (Opened)
	{
		if (offset==-1)offset=Read_dword();
		target->Seek(HeaderOffset+offset);
	}
}
//---------------------------------------------------------------------------

long Archive::OffsetTagGet()			// returns a read tag
{
	if (Opened)
	{
		long offset=Read_dword();
		return(offset);
	}
	return(0);
}

//---------------------------------------------------------------------------

long Archive::OffsetTagPlace()
{
	if (Opened)
	{
		long offset=target->GetPos();
		Write_dword(0);
		return(offset-HeaderOffset);
	}
	return(0);
}

//---------------------------------------------------------------------------

void Archive::OffsetTagSet(long offset)// writes current offset to previous tag..
{
	if (Opened)
	{
		long newoffset=target->GetPos()-HeaderOffset;
		target->WriteAt(offset+HeaderOffset,&newoffset,sizeof(long)); // write tag
	}
}

//---------------------------------------------------------------------------

void  Archive::Write_byte(byte b)  {	if (Opened)target->Write(&b,sizeof(byte));}
void  Archive::Write_word(word w)  {	if (Opened)target->Write(&w,sizeof(word));}
void  Archive::Write_dword(dword d){	if (Opened)target->Write(&d,sizeof(dword));}
void  Archive::Write_qword(qword q){	if (Opened)target->Write(&q,sizeof(qword));}
byte  Archive::Read_byte()		   {	byte b=0; if (Opened)target->Read(&b,sizeof(byte));return(b);}
word  Archive::Read_word()		   {	word b=0; if (Opened)target->Read(&b,sizeof(word));return(b);}
dword Archive::Read_dword()		   {	dword b=0;if (Opened)target->Read(&b,sizeof(dword));return(b);}
qword Archive::Read_qword()		   {	qword b=0;if (Opened)target->Read(&b,sizeof(qword));return(b);}	
void  Archive::Write(const void *buf, uint len){ Write_byte(FID_BLOCK);if (Opened)target->Write(buf,len);}
bool  Archive::Read(void *buf, uint len)       { byte b=Read_byte();assert(b==FID_BLOCK);if (Opened)return(target->Read(buf,len));return(false);}
dword Archive::GetPos() const			       { if (Opened)return(target->GetPos()-HeaderOffset);return(0);}

