
/*
---------------------------------------------------------------------------
|					Runtime assembler compilation
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/

#ifndef _H_ASMOBJ
#define _H_ASMOBJ

// asm defines
	// Control word flags
enum
{
	FCW_INVALIDOP	= 0x1,
	FCW_DENORMAL	= 0x2,
	FCW_ZERO		= 0x4,
	FCW_OVERFLOW	= 0x8,
	FCW_UNDERFLOW	= 0x10,
	FCW_PRECISION	= 0x20,	
	FCW_ALLEX		= FCW_INVALIDOP|FCW_DENORMAL|FCW_ZERO|FCW_OVERFLOW|FCW_UNDERFLOW|FCW_PRECISION,
	FCW_PRECISION32	= 0x00,
	FCW_PRECISION64 = 0x200,
	FCW_PRECISION80 = 0x300,
	FCW_ROUNDNEAREST= 0x0,
	FCW_ROUNDUP		= 0x800,
	FCW_ROUNDDOWN	= 0x400,
	FCW_ROUNDCHOP	= 0xc00,

	R_EAX			= 0x05,
	R_EBX			= 0x1D,
	R_ECX			= 0x0D,
	R_EDX			= 0x15,

	B_EAX			= 0x0,
	B_ECX			= 0x1,
	B_EDX			= 0x2,
	B_EBX			= 0x3,
	B_ESP			= 0x4,
	B_EBP			= 0x5,
	B_ESI			= 0x6,
	B_EDI			= 0x7,

	SRC_EAX			= 0xC0,
	SRC_ECX			= 0xC8,
	SRC_EDX			= 0xD0,
	SRC_EBX			= 0xD8,
	SRC_ESP			= 0xE0,
	SRC_EBP			= 0xE8,
	SRC_ESI			= 0xF0,
	SRC_EDI			= 0xF8,

	SRCM_EAX		= 0x00,
	SRCM_ECX		= 0x08,
	SRCM_EDX		= 0x10,
	SRCM_EBX		= 0x18,
	SRCM_ESP		= 0x20,
	SRCM_EBP		= 0x28,
	SRCM_ESI		= 0x30,
	SRCM_EDI		= 0x38,

};

#define ASMOBJSTARTSIZE 4096
class  AsmObj
{
	public:
	AsmObj(); 
	~AsmObj();
	void Reset();
	void Done();
	//void *LockFunction(); // locks/unlocks function array and returns function ptr
	//void UnlockFunction(); 
	bool HasMem;			// has memory already
	byte  *fdata;			// virual allocedf
	dword DataOffset;		// offset while writing
	dword FinalSize;		// size of the final function
	bool IsCompiled;		// wether the function is compiled already
	bool IsOverflow;
	// data array helper functions
	inline void AB(byte b )		{	if (DataOffset>=ASMOBJSTARTSIZE)IsOverflow=true;else fdata[DataOffset]=b;DataOffset++;}
	inline void ARW(word w )	{	AB(gLoByte(w));AB(gHiByte(w));	  }			
	inline void ARD(dword d)	{	ARW(gLoWord(d)); ARW(gHiWord(d)); }			
	inline void ARQ(qword q)	{	ARD(gLoDWord(q));ARD(gHiDWord(q));}			
	inline void AW(word w )		{	AB(gHiByte(w));AB(gLoByte(w));	}			
	inline void AD(dword d)		{	AW(gHiWord(d));AW(gLoWord(d));	}			
	inline void AQ(qword q)		{	AD(gHiDWord(q));ARD(gLoDWord(q));}			
	
	// Normal CPU
	inline void ADD32RI(byte reg,dword im){AB(0x81);AB(0xC0+reg);ARD(im);}// 81 /0 id
	inline void ADD32RR(byte src,byte dst){AB(0x03);AB(src+dst);}
	inline void AND32RI(byte reg,dword im){AB(0x81);AB(0xE0+reg);ARD(im);}// 81 /4 id
	inline void AND32RR(byte src,byte dst){AB(0x23);AB(src+dst);}
	inline void CMP32I(byte reg,dword im){AB(0x81);AB(0xF8+reg);ARD(im);}
	inline void DEC32(byte regID)   {AB(0xFF);AB(0xC8+regID);}
	inline void INC32(byte reg)		{AB(0x40+reg);}
	inline void JMP(byte diff)		{AB(0xEB);AB(diff);}
	inline void JA(byte diff)		{AB(0x77);AB(diff);}
	inline void JAE(byte diff)		{AB(0x73);AB(diff);}
	inline void JB(byte diff)		{AB(0x72);AB(diff);}
	inline void JBE(byte diff)		{AB(0x76);AB(diff);}
	inline void JC(byte diff)		{AB(0x72);AB(diff);}
	inline void JCXZ(byte diff)		{AB(0xE3);AB(diff);}
	inline void JE(byte diff)		{AB(0x74);AB(diff);}
	inline void JG(byte diff)		{AB(0x7F);AB(diff);}
	inline void JGE(byte diff)		{AB(0x7D);AB(diff);}
	inline void JL(byte diff)		{AB(0x7C);AB(diff);}
	inline void JLE(byte diff)		{AB(0x7E);AB(diff);}
	inline void JNA(byte diff)		{AB(0x76);AB(diff);}
	inline void JNAE(byte diff)		{AB(0x72);AB(diff);}
	inline void JNB(byte diff)		{AB(0x73);AB(diff);}
	inline void JNBE(byte diff)		{AB(0x77);AB(diff);}
	inline void JNC(byte diff)		{AB(0x73);AB(diff);}
	inline void JNE(byte diff)		{AB(0x75);AB(diff);}
	inline void JNG(byte diff)		{AB(0x7E);AB(diff);}
	inline void JNGE(byte diff)		{AB(0x7C);AB(diff);}
	inline void JNL(byte diff)		{AB(0x7D);AB(diff);}
	inline void JNLE(byte diff)		{AB(0x7F);AB(diff);}
	inline void JNO(byte diff)		{AB(0x71);AB(diff);}
	inline void JNP(byte diff)		{AB(0x7B);AB(diff);}
	inline void JNS(byte diff)		{AB(0x79);AB(diff);}
	inline void JNZ(byte diff)		{AB(0x75);AB(diff);}
	inline void JO(byte diff)		{AB(0x70);AB(diff);}
	inline void JP(byte diff)		{AB(0x7A);AB(diff);}
	inline void JPE(byte diff)		{AB(0x7A);AB(diff);}
	inline void JPO(byte diff)		{AB(0x7B);AB(diff);}
	inline void JS(byte diff)		{AB(0x78);AB(diff);}
	inline void JZ(byte diff)		{AB(0x74);AB(diff);}
	inline void LOOP(byte diff)		{AB(0xE2);AB(diff);}
	inline void MOV32M2R(byte regID,void *addr){AB(0x8B);AB(regID);ARD((dword)addr);}
	inline void MOV32R2M(byte regID,void *addr){AB(0x89);AB(regID);ARD((dword)addr);}
	inline void MOV32I2R(byte regID,dword val) {AB(0xB8+regID);ARD(val);}
	inline void MOV32R2R(byte src,byte dst){AB(0x8b);AB(src+dst);}
	inline void MUL32EAXR(byte reg) {AB(0xF7);AB(0xE0+reg);}
	inline void PUSHAD()			{AB(0x60);}
	inline void POPAD()				{AB(0x61);}
	inline void RET()				{AB(0xC3);}	
	inline void SAHF()				{AB(0x9E);}
	inline void SHL32RI(byte reg,byte d){AB(0xC1);AB(0xE0+reg);AB(d);}// C1 /4 ib
	inline void SHR32RI(byte reg,byte d){AB(0xC1);AB(0xE8+reg);AB(d);}// C1 /5 ib
	inline void XOR32RR(byte src,byte dst){AB(0x31);AB(src+dst);}


	// FPU
	inline void F2XM1()				{AW(0xD9F0);}
	inline void FABS()				{AW(0xD9E1);}
	inline void FADDP()				{AW(0xDEC1);}
	inline void FCHS()				{AW(0xD9E0);}
	inline void FCOM(word i=1)		{AW(0xD8D0+i);}
	inline void FCOMP(word i=1)		{AW(0xD8D8+i);}
	inline void FCOS()				{AW(0xD9FF);}
	inline void FDECSTP()			{AW(0xD9F6);}
	inline void FDIV(word i=1)		{AW(0xD8F0+i);}
	inline void FDIVP()				{AW(0xDEF9);}
	inline void FDIVRP()			{AW(0xDEF1);}
	inline void FINCSTP()			{AW(0xD9F7);}
	inline void FILD32(void *addr)	{AW(0xDB05);ARD((long)addr);}
	inline void FISTP32(void *addr) {AW(0xDB1D);ARD((long)addr);}
	inline void FIST32(void *addr)  {AW(0xDB15);ARD((long)addr);}
	inline void FLD64(void *addr)	{AW(0xDD05);ARD((long)addr);}
	inline void FLD32(void *addr)	{AW(0xD905);ARD((long)addr);}
	inline void FLD64R(byte reg)	{AB(0xDD);AB(reg);} // DD /0
	inline void FLD32R(byte reg)	{AB(0xD9);AB(reg);} // D9 /0
	inline void FLDCW(void *addr)	{AW(0xD92D);ARD((long)addr);}
	inline void FLDZ()				{AW(0xD9EE);}
	inline void FLD1()				{AW(0xD9E8);}
	inline void FLDL2T()			{AW(0xD9E9);}
	inline void FLDL2E()			{AW(0xD9EA);}
	inline void FLDPI()				{AW(0xD9EB);}
	inline void FLDLG2()			{AW(0xD9EC);}
	inline void FLDLN2()			{AW(0xD9ED);}
	inline void FLDST(word i)		{AW(0xD9C0+i);}
	inline void FMUL(word i=1)		{AW(0xD8C8+i);}
	inline void FMULP(word i=1)		{AW(0xDEC8+i);}
	inline void FNSTSWAX()			{AW(0xDFE0);}
	inline void FPATAN()			{AW(0xD9F3);}
	inline void FPREM1()			{AW(0xD9F5);}
	inline void FPREM()				{AW(0xD9F8);}
	inline void FPTAN()				{AW(0xD9F2);}
	inline void FRNDINT()			{AW(0xD9FC);}
	inline void FSIN()				{AW(0xD9FE);}
	inline void FSTP64(void *addr)	{AW(0xDD1D);ARD((long)addr);}
	inline void FSTP32(void *addr)	{AW(0xD91D);ARD((long)addr);}
	inline void FST64(void *addr)	{AW(0xDD15);ARD((long)addr);}
	inline void FSTP(word i)		{AW(0xDDD8+i);}
	inline void FSQRT()				{AW(0xD9FA);}
	inline void FSUBP(word i=1)		{AW(0xDEE8+i);}
	inline void FSUBR(word i=1)		{AW(0xDCE0+i);}
	inline void FSUBRP()			{AW(0xDEE1);}
	inline void FSCALE()			{AW(0xD9FD);}
	inline void FTST()				{AW(0xD9E4);}
	inline void FXCH(word i=1)		{AW(0xD9C8+i);}
	inline void FYL2X()				{AW(0xD9F1);}
};

#endif