/*
---------------------------------------------------------------------------
|			   Chain class. used for linked chains.. Implementation
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
Requirements:
Changes:
	07-08-1998	Initial file setup
*/

#include "Library.h"

//--------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------
// _chain constructor
//--------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------

_Cchain::_Cchain(_Cchain *i_prev,_Cchain *i_next)
{
	prev=i_prev;
	next=i_next;
	if (prev!=NULL){next=prev->next,prev->next=this;}
	if (next!=NULL){prev=next->prev;next->prev=this;}
}

//--------------------------------------------------------------------------------------------
// _chain destructor

_Cchain::~_Cchain()
{
	// Tie the links
	if (prev!=NULL)prev->next=next;
	if (next!=NULL)next->prev=prev;
}

//--------------------------------------------------------------------------------------------
// _chain unlink()

void _Cchain::unlink()
{
	if (prev!=NULL)prev->next=next;
	if (next!=NULL)next->prev=prev;
	prev=NULL;
	next=NULL;
}


//--------------------------------------------------------------------------------------------
// _chain link()

void _Cchain::link(_Cchain *i_prev,_Cchain *i_next)
{
	prev=i_prev;
	next=i_next;
	//if (prev!=NULL && next!=NULL)ASSERT(prev->next==next);
	if (prev!=NULL){next=prev->next,prev->next=this;}
	if (next!=NULL){prev=next->prev,next->prev=this;}
}

//--------------------------------------------------------------------------------------------
// _chain move()

void _Cchain::move(_Cchain *i_prev,_Cchain *i_next)
{
	unlink();
	link(i_prev,i_next);
}

//--------------------------------------------------------------------------------------------
// grabs the chain by its base, and moves it to another one
void _Cchain::moveprev(_Cchain *i_prev)
{
	if (prev!=NULL)prev->next=NULL;
	prev=i_prev;
	if (prev!=NULL)prev->next=this;
}

//--------------------------------------------------------------------------------------------

void _Cchain::slide(bool up)
{
	if (up==true)
	{
		if (prev!=NULL)
		{
			move(prev->prev,prev);
		}
	}
	else
	{
		if (next!=NULL)
		{
			move(next,next->next);
		}
	}
}

//--------------------------------------------------------------------------------------------

void _Cchain::move_to_bottom()
{
	_Cchain *last=this;
	while (last->next!=NULL)last=last->next;
	if (last!=this)
	{
		unlink();
		link(last);
	}
}
	// Combination of unlink and link

//--------------------------------------------------------------------------------------------
// _chain enumlen()

_Cchain *_Cchain::enumelem(short *num)
{
	short i=0;
	_Cchain *p;
	p=this;
	
	if (*num==-1)
	{
		while (p->next!=NULL)
		{
			//ASSERT(p->next!=this); // Cyclic link!
			i++;
			p=p->next;
		};
		*num=i;
		return(p);
	}

	for (;;)
	{
		if (p->prev==NULL)break;
		else p=p->prev;
	}
	
	while (i<*num)
	{
		p=p->next;
		i++;
		if (p==NULL)break;
	}
	*num=i;
	return (p);
}

_Cchain *_Cchain::enumelem(short num){return (enumelem(&num));}

//--------------------------------------------------------------------------------------------
// enum elem by object()

short _Cchain::enumelem(_Cchain *d)
{
	short i=0;
	_Cchain *p=this;
	while(p!=NULL)
	{
		if (p==d)return(i);
		i++;
		p=p->next;
	}
	return(-1);
}

//--------------------------------------------------------------------------------------------
// _chain delete_chain()
void _Cchain::delete_chain()
{
	_Cchain *p,*q;
	p=enumelem(-1);
	while (p!=NULL)
	{
		q=p->prev;
		if (p!=this)delete p;
		p=q;
	}
	delete this;
}

//--------------------------------------------------------------------------------------------
// _chain delete_chain()
void _Cchain::delete_links()
{
	_Cchain *p,*q;
	p=next;
	while (p!=NULL)
	{
		q=p->next;
		delete p;
		p=q;
	}
}

//--------------------------------------------------------------------------------------------

