/*
---------------------------------------------------------------------------
|	Tree class, to be used as a a thing like _Cchain, but then in tree form
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
Requirements:

Changes:
	Thursday 5th august:	Initial file setup
*/

#include "Library.h"

//-----------------------------------------------------------------------------
// Constructor

_ctree::_ctree(_ctree *i_parent,bool inserttop)
{
// Add yourself to the parent
	parent=NULL;
	child=NULL;
	place(i_parent,inserttop);
}

//-----------------------------------------------------------------------------
// Destructor

_ctree::~_ctree()
{
	// Delete children
	while (child!=NULL)delete child;
	// Delete yourself
	if (parent!=NULL && parent->child==this)parent->child=next;
	if (next!=NULL)next->prev=prev;
	if (prev!=NULL)prev->next=next;
}

//-----------------------------------------------------------------------------
// Delete children

void _ctree::delete_children()
{
	while (child!=NULL)delete child;
}

//--------------------------------------------------------------------------------------------

void _ctree::place(_ctree *i_parent,bool inserttop)	// inserts into tree
{
	if (i_parent!=NULL)
	{
		if (parent!=NULL)remove();
		parent=i_parent;
		if (parent->child!=NULL)
		{
			if (inserttop==true)
			{
				// Insert at top
				prev=NULL;
				next=parent->child;
				next->prev=this;
				parent->child=this;
			}
			else
			{
				// get last object
				_ctree *i=parent->child;
				while (i->next!=NULL)i=i->next;
				i->next=this;
				prev=i;
				next=NULL;
			}
		}
		else
		{
			prev=NULL;
			next=NULL;
			parent->child=this;
		}
	}
	else
	{
		prev=NULL;
		next=NULL;
		parent=NULL;
		child=NULL;	
	}
}

//--------------------------------------------------------------------------------------------

void   _ctree::remove()					// removes from tree
{
	if (parent!=NULL)
	{
		if (parent->child==this)
		{
			parent->child=next;
		}
		if (next!=NULL)next->prev=prev;
		if (prev!=NULL)prev->next=next;

		prev=NULL;
		next=NULL;
		parent=NULL;
	}
}

//--------------------------------------------------------------------------------------------

void _ctree::move(_ctree *i_parent,bool inserttop)
{
	remove();
	place(i_parent,inserttop);
}// moves to other parent

//--------------------------------------------------------------------------------------------

void  _ctree::movechildren(_ctree *i_parent) // children to other parent
{
	if (i_parent!=this)
	{
		_ctree *obj=child,*t;
		while (obj!=NULL)
		{
			t=obj->next;
			obj->remove();
			obj->place(i_parent);
			obj=t;
		}
	}
}

//--------------------------------------------------------------------------------------------
// NOTE .. SLIDE is malfunctioning!
void _ctree::slide(bool down)
{
	if (down==false)
	{
		if (prev!=NULL)
		{
			// Tie down old place
			prev->next=next;
			if (next!=NULL)next->prev=prev;

			if (parent!=NULL && parent->child==prev)parent->child=this;

			next=prev;prev=prev->prev;
			
			// set new neigbours
			if (prev!=NULL)prev->next=this;
			if (next!=NULL)next->prev=this;
		}
	}		
	else
	{
		if (next!=NULL)
		{
			if (prev!=NULL)prev->next=next;
			next->prev=prev;

			if (parent!=NULL && parent->child==this)parent->child =next;

			// Set new place
			prev=next;next=next->next;
			
			prev->next=this;
			if (next!=NULL)next->prev=this;
		}
	}
}

//--------------------------------------------------------------------------------------------

void _ctree::move_bottom()
{
	if (next!=NULL)
	{
		// Tie down old place
		if (prev!=NULL)prev->next=next;
		if (next!=NULL)next->prev=prev;

		if (parent!=NULL && parent->child==this)parent->child=next;

		_ctree *obj=next;
		while (obj->next!=NULL)obj=obj->next;
		prev=obj;
		next=NULL;
		prev->next=this;
	}
}

//--------------------------------------------------------------------------------------------

void _ctree::move_top()
{
	if (prev!=NULL)
	{
		// Tie down old place
		if (prev!=NULL)prev->next=next;
		if (next!=NULL)next->prev=prev;

		_ctree *obj=prev;
		while (obj->prev!=NULL)obj=obj->prev;

		prev=NULL;
		next=obj;
		next->prev=this;
		if (parent!=NULL)parent->child=this;
	}
}

//--------------------------------------------------------------------------------------------

_ctree   *_ctree::findlastsibling()
{
	_ctree *obj=next;
	if (next==NULL)return(this);

	while (obj->next!=NULL)
	{
		obj=obj->next;
	}
	return(obj);
}// finds last sibling..
		
//--------------------------------------------------------------------------------------------

_ctree *_ctree::walk(_ctree *startobj,short *direction,bool passbranch) 
{
	_ctree *obj=this;
	if (direction!=NULL)
	{
		(*direction)=0;

		if (obj->child!=NULL && !passbranch)
		{
			(*direction)++;
			return(obj->child);
		}
		if (obj==startobj)return(NULL);

		while (obj->next==NULL)
		{
			obj=obj->parent;(*direction)--;
			if (obj==startobj || obj==NULL)return (NULL);
		}
		return (obj->next);
	}
	else
	{
		if (obj->child!=NULL && !passbranch)return(obj->child);
		if (obj==startobj)return(NULL);
		
		while (obj->next==NULL)
		{
			obj=obj->parent;
			if (obj==startobj || obj==NULL)return (NULL);
		}
		return (obj->next);
	}
}

//--------------------------------------------------------------------------------------------

_ctree *_ctree::copy(_ctree *parent) // returns pointer to new _ctree object, identical to this one
{
	_ctree *clone=new _ctree(parent);

	_ctree *obj=child;
	while (obj!=NULL)
	{
		obj->copy(clone);
		obj=obj->next;
	}
	return(clone);
}

//--------------------------------------------------------------------------------------------

void _ctree::copynb(_ctree *parent) // Copies neigbours too
{
	// find first object
	_ctree *obj=this;while (obj->prev!=NULL)obj=obj->prev;
	// now call copy on all neigbours
	while (obj!=NULL)
	{
		obj->copy(parent);
		obj=obj->next;
	}
}

//--------------------------------------------------------------------------------------------

bool _ctree::IsChildOf(_ctree *object) // checks wether this object is a child of object
{
	if (object!=NULL)
	{
		_ctree *obj=object;
		while (obj!=NULL)
		{
			obj=obj->walk(object);
			if (obj==this)return(true);		
		}
	}
	return(false);
}

//--------------------------------------------------------------------------------------------

_ctree	*_ctree::enumchild(long num)
{
	_ctree *obj=child;
	long j=0;
	while (obj!=NULL)
	{
		if (j==num)return(obj);
		obj=obj->next;j++;
	}
	return(NULL);
}

//--------------------------------------------------------------------------------------------

long _ctree::getenumchild(_ctree *c)
{
	_ctree *obj=child;
	long j=0;
	while (obj!=NULL)
	{
		if (obj==c)return(j);
		obj=obj->next;j++;
	}
	return(-1);
}

//--------------------------------------------------------------------------------------------
