
/*
---------------------------------------------------------------------------
|					Color defines and functions
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
Requirements

Changes:
	Wendsday 20th december:	Initial file setup
*/

//------------------------------------------------------------------------------------------------
//
// RGB Colors
//
#pragma once
enum
{
	// Orange
	cORANGE				= 0xff00a5ff,
	cDARK_ORANGE		= 0xff008cff,

	// Primary colors
	cWHITE				= 0xffffffff,
	cBLACK				= 0xff000000,
	cRED				= 0xff0000ff,
	cGREEN				= 0xff00ff00,
	cBLUE				= 0xffff0000,
	cYELLOW				= 0xff00ffff,
	cMAGENTA			= 0xffff00ff,
	cCYAN				= 0xffffff00,

	cDARK_RED			= 0xff00008b,
	cDARK_GREEN			= 0xff008b00,
	cDARK_BLUE			= 0xff8b0000,
	cDARK_YELLOW		= 0xff008b8b,
	cDARK_MAGENTA		= 0xff8b008b,
	cDARK_CYAN			= 0xff8b8b00,

	// Grey tones
	cLIGHT_GRAY			= 0xffd3d3d3,
	cGRAY				= 0xffbebebe,
	cDARK_GRAY			= 0xff696969,
	cLIGHT_SLATE_GRAY	= 0xff998877,
	cSLATE_GRAY			= 0xff908070,
	cDARK_SLATE_GRAY	= 0xff4f4f2f,

	// Blue tones
	cMIDNIGHT_BLUE		= 0xff701919,
	cCORNFLOWER_BLUE	= 0xffed9564,
	cLIGHT_SLATE_BLUE	= 0xffff7084,
	cSLATE_BLUE			= 0xffcd5a6a,
	cDARK_SLATE_BLUE	= 0xff8b3d48,
	cDEEP_SKY_BLUE		= 0xffffbf00,
	cSTEEL_BLUE			= 0xffb48246,

	// Green tones
	cSEA_GREEN			= 0xff578b2e,
	cLAWN_GREEN			= 0xff00fc7c,

	// Red tones
	cINDIAN_RED			= 0xff5c5ccd,
	cTOMATO				= 0xff4763ff,
	cFIREBRICK			= 0xff2222b2,

	// Colors
	cAQUAMARINE			= 0xffd4ff7f,
	cAZURE				= 0xfffffff0,
	cTURQUOISE			= 0xffd0e040,
	cGOLD				= 0xff00d7ff,
	cKHAKI				= 0xff8ce6f0,
	cDARK_KHAKI			= 0xff6bb7bd,
	cPEACH_PUFF			= 0xffb9daff,
	cLAVENDER			= 0xfffae6e6,
	cWHEAT				= 0xffbae7ff
};

//------------------------------------------------------------------------------------------------
//
// Pixel format ID
//
//------------------------------------------------------------------------------------------------
DECLARE_CLASS(RGBIColor)
DECLARE_CLASS(RGBColor)
class RGBColor;
//
// RGB Color
//

// integer (dword ARGB) version
class  RGBIColor
{
public:
	RGBIColor()														: col(0xffffffff) { }
	RGBIColor(byte r,byte g,byte b,byte a=255)						{ col=(a<<24)|(r<<16)|(g<<8)|b;}
	RGBIColor(cRGBIColor other)										: col(other.col) { }
	explicit RGBIColor(dword argb)									: col(argb){ }
	cRGBIColor operator=(byte other)							    { Set(other,other,other,other);return *this; }
	cRGBIColor operator=(cRGBIColor other)							{ col=other.col; return *this; }
	inline	void	Set(byte rr, byte gg, byte bb,byte aa=255)		{ col=(aa<<24)|(rr<<16)|(gg<<8)|bb;}
	
	// Color packing/unpacking to DWord
	// dword format:  a r g b (high-low)

	// color unpacking
	inline  RGBIColor&		 FromA8(byte d)								{col=(dword)d<<24;return *this;}
	
	inline  RGBIColor&		 FromARGB_8888(dword d)						{col=d; return (*this);}
	inline  RGBIColor&		 FromRGBA_8888(dword d)						{col=/*r*/((d>>8)&0xff0000)|/*g*/((d>>8)&0xff00)|/*b*/((d>>8)& 0xff)|/*a*/((d&0xff)<<24); return (*this);}
	inline  RGBIColor&		 FromBGRA_8888(dword d)						{col=/*b*/((d>>24)&0xff)|/*g*/((d>>8)&0xff00)|/*r*/((d<<8)& 0xff0000)|/*a*/((d&0xff)<<24); return (*this);}
	inline  RGBIColor&		 FromARGB_1555(word d)						{col=/*a*/((d<<16))|/*r*/((d&0x7c00)<<9)|((d&0x3e0)<<6)|((d&0x1f)<<3); return (*this); return (*this);}
	inline  RGBIColor&		 FromRGBA_5551(word d)						{col=/*r*/((d&0xf800)<<8)|/*g*/((d&0x7c0)<<5)|/*b*/((d&0x3e)<<2)|/*a*/(((d&0x1)<<31));  return (*this);}
	inline  RGBIColor&		 FromBGRA_5551(word d)						{col=/*b*/((d&0xf800)>>8)|/*g*/((d&0x7c0)<<5)|/*r*/((d&0x3e)<<18)|/*a*/(((d&0x1)<<31));  return (*this);}
	inline  RGBIColor&		 FromARGB_4444(word d)						{col=/*a*/((d&0xf000)<<16|/*r*/((d&0xf00)<<12)|/*g*/((d&0xf0)<<8))|/*b*/((d&0xf)<<4); return (*this);}

	inline  RGBIColor&		 FromRGB_X888(dword d)						{col= d&0x00ffffff; return (*this);}
	inline  RGBIColor&		 FromBGR_X888(dword d)						{col= /*g*/(d&0xff00)|/*b*/((d&0xff0000)>>16)|/*r*/((d&0xff)<<16); return (*this);}
	
	inline  RGBIColor&		 FromRGB_565(word d)						{col= /*r*/((d&0xf800)<<8)|/*g*/((d&0x7e0)<<5)|/*b*/((d&0x1f)<<3);  return (*this);}
	inline  RGBIColor&		 FromRGB_555(word d)						{col= /*r*/((d&0x7c00)<<9)|/*g*/((d&0x3e0)<<6)|/*b*/((d&0x1f)<<3); return (*this);}
	inline  RGBIColor&		 FromBGR_555(word d)						{col= /*b*/((d&0x7c00)>>7)|/*g*/((d&0x3e0)<<6)|/*r*/((d&0x1f)<<19); return (*this);}
	
		
	// color packing
	inline  byte      ToA8()										{return (col&0xff000000)>>24; }

	inline  dword     ToARGB_8888()									{return col;}
	inline  dword     ToRGBA_8888()									{return ((col&0xff000000)>>24)|((col&0xff0000)<<8)|((col&0xff00)<<8)|((col&0xff)<<8);}
	inline  dword     ToBGRA_8888()									{return ((col&0xff000000)>>24)|((col&0xff0000)>>8)|((col&0xff00)<<8)|((col&0xff)<<24);}
	
	inline  word      ToARGB_1555()									{return ((col>>16)&0x8000)|((col>>9)&0x7c00)|((col>>6)&0x3e0)|((col>>3)&0x1f);}
	inline  word      ToRGBA_5551()									{return ((col>>31)&0x1)|((col>>8)&0xf800)|((col>>5)&0x7c0)|((col>>2)&0x3e);}
	inline  word      ToBGRA_5551()									{return ((col>>31)&0x1)|((col>>18)&0x3e) |((col>>5)&0x7c0)|((col<<8)&0xf800);}
	
	inline  word      ToARGB_4444()									{return ((col>>16)&0xf000)|((col>>12)&0xf00)|((col>>8)&0xf0)|((col>>4)&0xf);}
	
	inline  dword     ToRGB_X888()									{return col&0x00ffffff;}
	inline  dword     ToBGR_X888()									{return (col&0xff00)|((col&0xff)<<16)|((col&0xff0000)>>16);}
	
	inline  word      ToRGB_565()									{return ((col>>8)&0xf800)|((col>>5)&0x7e0)|((col>>3)&0x1f);}
	inline  word      ToRGB_555()									{return ((col>>9)&0x7c00)|((col>>6)&0x3e0)|((col>>3)&0x1f);}
	inline  word      ToBGR_555()									{return ((col<<7)&0x7c00)|((col>>6)&0x3e0)|((col>>19)&0x1f);}

	inline	operator dword ()										{ return col; }

	inline	byte	 a()											{ return (col>>24)&0xff; }
	inline	byte	 r()											{ return (col>>16)&0xff; }
	inline	byte	 g()											{ return (col>>8 )&0xff; }
	inline	byte	 b()											{ return (col    )&0xff; }

	friend	rArchive operator<<(rArchive f, cRGBIColor rgb)	{ return f << rgb.col; }
	friend	rArchive operator>>(rArchive f, rRGBIColor rgb)	{ return f >> rgb.col; }

	// generic interface
	RGBIColor &FromType(long pf,dword d);
	dword ToType(long pf);

	dword	col;
};

class  RGBColor
{
public:
	// Initialization
	RGBColor()														: r(1.0), g(1.0), b(1.0),a(1.0) { assert((&g) == (&r+1));  assert((&b) == (&g+1)); }
	RGBColor(float nr, float ng, float nb,float na=1.0)				{ r=nr, g=ng, b=nb, a=na;}
	RGBColor(cRGBColor other)										: r(other.r), g(other.g), b(other.b), a(other.a) { }
	explicit RGBColor(dword rgba)									: r(((rgba>>0) & 0xff)/255.0f), g(((rgba>>8) & 0xff)/255.0f), b(((rgba>>16) & 0xff)/255.0f) { }
	cRGBColor operator=(float other)							    { r=g=b=a=other;return *this; }
	cRGBColor operator=(cRGBColor other)							{ r=other.r; g=other.g; b=other.b;a=other.a; return *this; }
	inline	void	Set(float rr, float gg, float bb,float aa=0.0f)	{ r=rr; g=gg; b=bb;a=aa; }

	// Access operators
	inline	float&	operator[](int elem)							{ assert(elem<3); return ((float*)&r)[elem]; }
	inline	operator float*()										{ return &r; }
	inline	operator const float*() const							{ return &r; }

	// Operations
	inline	float	Intensity()										{ return r*0.212671f + g*0.715160f + b*0.072169f; }
	inline	RGBColor& Clamp()										{ r=clamp_f(r, 0.0f, 1.0f); g=clamp_f(g, 0.0f, 1.0f); b=clamp_f(b, 0.0f, 1.0f);a=clamp_f(a,0.0f,1.0f); return *this; }
	inline	RGBColor& BlendWith(cRGBColor rhs)						{ r=(rhs.r+r)*0.5f; g=(rhs.g+g)*0.5f; b=(rhs.b+b)*0.5f; return *this; }
	inline	RGBColor& BlendWith(cRGBColor rhs, float frac)			{ bounds(frac, 0.0f, 1.0f); float ifrac = 1.0f-frac; r = (rhs.r*frac) + (r*ifrac); g = (rhs.g*frac) + (g*ifrac); b = (rhs.b*frac) + (b*ifrac);a= (rhs.a*frac) * (a*ifrac); return *this; }
	inline  RGBColor  Interpolate(cRGBColor rhs, float frac)		{ bounds(frac, 0.0f, 1.0f); float ifrac = 1.0f-frac; return RGBColor((rhs.r*frac) + (r*ifrac), (rhs.g*frac) + (g*ifrac), (rhs.b*frac) + (b*ifrac), (rhs.a*frac) + (a*ifrac)); }

	// color unpacking
	inline  RGBColor& FromA8(byte d)								{a=(float)d/255.0f;		return (*this);}
	
	inline  RGBColor& FromARGB_8888(dword d)						{a=((d>>24)&0xff)/255.0f,r=((d>>16)&0xff)/255.0f, g=((d>>8)& 0xff)/255.0f,b=(d&0xff)/255.0f;      return (*this);}
	inline  RGBColor& FromRGBA_8888(dword d)						{r=((d>>24)&0xff)/255.0f,g=((d>>16)&0xff)/255.0f, b=((d>>8)& 0xff)/255.0f,a=(d&0xff)/255.0f;      return (*this);}
	inline  RGBColor& FromBGRA_8888(dword d)						{b=((d>>24)&0xff)/255.0f,g=((d>>16)&0xff)/255.0f, r=((d>>8)& 0xff)/255.0f,a=(d&0xff)/255.0f;      return (*this);}

	inline  RGBColor& FromARGB_1555(word d)							{a=((d>>15)&0x1),		 r=((d>>10)&0x1f)/31.0f ,g=((d>>5)&0x1f)/31.0f,  b=(d&0x1f)/31.0f; 		 return (*this);}
	inline  RGBColor& FromRGBA_5551(word d)							{r=((d>>11)&0x1f)/31.0f ,g=((d>>6)&0x1f)/31.0f,  b=((d>>1)&0x1f)/31.0f,  a=(d&1);				 return (*this);}
	inline  RGBColor& FromBGRA_5551(word d)							{b=((d>>11)&0x1f)/31.0f ,g=((d>>6)&0x1f)/31.0f,  r=((d>>1)&0x1f)/31.0f,  a=(d&1);				 return (*this);}
	inline  RGBColor& FromARGB_4444(word d)							{a=((d>>12)&0x10)/15.0f ,r=((d>>8)&0x10)/15.0f,  g=((d>>4)&0x10)/15.0f,  b=(d&0x10)/15.0f;		 return (*this);}

	inline  RGBColor& FromRGB_X888(dword d)							{r=((d>>16)&0xff)/255.0f, g=((d>>8)& 0xff)/255.0f,b=(d&0xff)/255.0f;							 return (*this);}
	inline  RGBColor& FromBGR_X888(dword d)							{b=((d>>16)&0xff)/255.0f, g=((d>>8)& 0xff)/255.0f,r=(d&0xff)/255.0f;							 return (*this);}
	
	inline  RGBColor& FromRGB_565(word d)				       		{r=((d>>11)&0x1f)/31.0f, g=((d>>5)&0x3f)/63.0f,  b=(d&0x1f)/31.0f;								 return (*this);}
	inline  RGBColor& FromRGB_555(word d)							{r=((d>>10)&0x1f)/31.0f ,g=((d>>5)&0x1f)/31.0f,  b=(d&0x1f)/31.0f;								 return (*this);}
	inline  RGBColor& FromBGR_555(word d)							{b=((d>>10)&0x1f)/31.0f ,g=((d>>5)&0x1f)/31.0f,  r=(d&0x1f)/31.0f;								 return (*this);}
	
		
	// color packing
	inline  byte      ToA8()										{return ( (byte)(a*255.0f) ); }

	inline  dword     ToARGB_8888()									{return ( (((dword)(a*255.0f)&0xff)<<24)) + (((dword)(r*255.0f)&0xff)<<16) + (((dword)(g*255.0f)&0xff)<<8) +  ((dword)(b*255.0f)&0xff); }
	inline  dword     ToRGBA_8888()									{return ( (((dword)(r*255.0f)&0xff)<<24)) + (((dword)(g*255.0f)&0xff)<<16) + (((dword)(b*255.0f)&0xff)<<8) +  ((dword)(a*255.0f)&0xff); }
	inline  dword     ToBGRA_8888()									{return ( (((dword)(b*255.0f)&0xff)<<24)) + (((dword)(g*255.0f)&0xff)<<16) + (((dword)(r*255.0f)&0xff)<<8) +  ((dword)(a*255.0f)&0xff); }
	
	inline  word      ToARGB_1555()									{return ( (((word)a)&0x1)<<15)		  + (((word)(r*31.0f)&0x1f)<<10)  + (((word)(g*31.0f)&0x1f)<<5)   + (((word)(b*31.0f)&0x1f) ); }
	inline  word      ToRGBA_5551()									{return ( (((word)(r*31.0f)&0x1f)<<11)+ (((word)(g*31.0f)&0x1f)<<6)   + (((word)(b*31.0f)&0x1f)<<1)   + (((word)a)&0x1)  ); }
	inline  word      ToBGRA_5551()									{return ( (((word)(b*31.0f)&0x1f)<<11)+ (((word)(g*31.0f)&0x1f)<<6)   + (((word)(r*31.0f)&0x1f)<<1)   + (((word)a)&0x1)  ); }
	
	inline  word      ToARGB_4444()									{return ( (((word)(a*15.0f)&0x10)<<12)+ (((word)(r*15.0f)&0x10)<<8)   + (((word)(g*15.0f)&0x10)<<4)   + ((word)(b*15.0f)&0x10) ); }
	
	inline  dword     ToRGB_X888()									{return ( (((dword)(r*255.0f)&0xff)<<16) + (((dword)(g*255.0f)&0xff)<<8) +  ((dword)(b*255.0f)&0xff) ); }
	inline  dword     ToBGR_X888()									{return ( (((dword)(b*255.0f)&0xff)<<16) + (((dword)(g*255.0f)&0xff)<<8) +  ((dword)(r*255.0f)&0xff) ); }
	
	inline  word      ToRGB_565()									{return ( (((word)(r*31.0f)&0x1f)<<11)+ (((word)(g*63.0f)&0x3f)<<5)   + ((word)(b*31.0f)&0x1f) ); }
	inline  word      ToRGB_555()									{return ( (((word)(r*31.0f)&0x1f)<<10)+ (((word)(g*31.0f)&0x1f)<<5)   + ((word)(b*31.0f)&0x1f) ); }
	inline  word      ToBGR_555()									{return ( (((word)(b*31.0f)&0x1f)<<10)+ (((word)(g*31.0f)&0x1f)<<5)   + ((word)(r*31.0f)&0x1f) ); }


	// generic interface
	RGBColor  &FromType(long pf,dword d);
	dword	  ToType(long pf);


	// Comparison
	inline	bool operator==(cRGBColor rhs)							{ return ((r==rhs.r) && (g==rhs.g) && (b==rhs.b)); }

	// Color addition/substraction
	inline	cRGBColor operator+=(cRGBColor rhs)						{ r+=rhs.r; g+=rhs.g; b+=rhs.b; return *this; }
	inline	cRGBColor operator-=(cRGBColor rhs)						{ r-=rhs.r; g-=rhs.g; b-=rhs.b; return *this; }
	inline	friend RGBColor operator+(cRGBColor lhs, cRGBColor rhs) { return RGBColor(lhs.r+rhs.r, lhs.g+rhs.g, lhs.b+rhs.b); }
	inline	friend RGBColor operator-(cRGBColor lhs, cRGBColor rhs) { return RGBColor(lhs.r-rhs.r, lhs.g-rhs.g, lhs.b-rhs.b); }

	// Color*scalar and vector/scalar
	inline	cRGBColor operator*=(float m)							{ r*=m; g*=m; b*=m; return *this; }
	inline	cRGBColor operator/=(float d)							{ float iMul=1.0f/d; r*=iMul; g*=iMul; b*=iMul; return *this; }
	inline	friend RGBColor operator*(cRGBColor lhs, float m)		{ return RGBColor(lhs.r*m, lhs.g*m, lhs.b*m); }
	inline	friend RGBColor operator*(float m, cRGBColor rhs)		{ return RGBColor(m*rhs.r, m*rhs.g, m*rhs.b); }
	inline	friend RGBColor operator/(cRGBColor lhs, float d)		{ float iMul=1.0f/d; return RGBColor(lhs.r*iMul, lhs.g*iMul, lhs.b*iMul); }
	
	// RGB/HSV conversion
	RGBColor HSV_to_RGB();
	RGBColor RGB_to_HSV();

	// HWB/RGB conversion
	RGBColor HWB_to_RGB();
	RGBColor RGB_to_HWB();
	// RGBColor -> String
	inline	String ToString() const									{ String t; t.Format("(r: %f, g: %f, b: %f a: %f)", r, g, b,a); return t; }

	friend	rArchive operator<<(rArchive f, cRGBColor rgb)	{ return f << rgb.r << rgb.g << rgb.b << rgb.a; }
	friend	rArchive operator>>(rArchive f, rRGBColor rgb)	{ return f >> rgb.r >> rgb.g >> rgb.b >> rgb.a; }

	// Data members
	float	r, g, b, a;
};
//-------------------------------------------------------------------------
// Color conversion/packing helper functions
enum
{
	// Letters of the types occur in the way the format is read
	// like 0xffaabbcc in ARGB mode is 0xff aa bb cc A R G B
	// One-component modes
	PF_INDEX_8			= 0,	// default
	PF_ALPHA_8			= 1,
	
	// ARGB & ABGR modes
	PF_ARGB_8888		= 2,	// Default (DX)
	PF_RGBA_8888		= 3,
	PF_BGRA_8888		= 4,
	PF_ARGB_1555		= 5,	// Default
	PF_RGBA_5551		= 6,	
	PF_BGRA_5551		= 7,	
	PF_ARGB_4444		= 8,	

	// RGB & BGR modes
	PF_RGB_X888			= 9,	// Default
	PF_BGR_X888			= 10,	
	PF_RGB_565			= 11,	// default
	PF_RGB_555			= 12,		
	PF_BGR_555			= 13		
};

// COLOR FORMAT helpers
String			PF_ToString(long type);
inline  long		PF_BPPByte(long bpp)									{return ( (bpp+1)/8 );}// causes 15 bpp to count as 2
long			PF_Size(long type);
long			PF_TypeFromMask(long rm,long gm,long bm,long am=0);
void    		PF_MaskFromType(long type,long &rm,long &gm,long &bm,long &am);
long			PF_TypeOfBpp(long type);
long			PF_BppOfType(long bpp);		// returns default PF for bpp

