/*
---------------------------------------------------------------------------
|		   Dialog manager, featuring other dialog framing and parenting
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/
#include "Library.h"
#include "DialogManager.h"
#include "FrameDlg.h"
#include "PopupMenu.h"

//---------------------------------------------------------------------------

DialogManager::DialogManager()
{
	parent=NULL;
	layout=DML_FULL;
	nFrames=0;
	Frames.Resize(10); // basic 'full' frame
}

//---------------------------------------------------------------------------

DialogManager::~DialogManager()
{

}

//---------------------------------------------------------------------------

void DialogManager::AddWindow(CDialogTool *window,long frameNo) // adds window to view (selected in dialog data)
{
	// first check if window doesn't exist yet
	if (FindWindow(window)!=-1)
	{
		log("DialogManager::AddWindow - window already present");
	}
	else
	{
		if (frameNo>=nFrames)frameNo=0;
		window->dmanager=this;
		Frames[frameNo].windows.Add(window); // add to list		

		if (frameNo==0) // move into tool frame
		{
			CFrameDlg *frdlg=new CFrameDlg(AfxGetMainWnd());
			frdlg->dmanager=this;
			window->framedlg=frdlg;
			window->SetParent(&frdlg->m_Frame); // move inside new frame dlg
			// set size of 'frdlg' to the that of the window
			CRect src;window->GetClientRect(src);
			long x=-1,y=-1;
			if (window->lastframe>0 && window->lastframe<nFrames)
			{
				x=Frames[window->lastframe].frame.left;
				y=Frames[window->lastframe].frame.top;
			}

			move_winptrwindow(frdlg,x,y,x+src.Width(),y+src.Height()+GetSystemMetrics(SM_CYCAPTION)+2*GetSystemMetrics(SM_CYSIZEFRAME));
			// now both are invisible
			frdlg->ShowWin();
			window->ShowWin();
			window->dialogtitle.SetWnd(frdlg);
		}
		else
		{
			// hide previous top window
			CWnd *top=Frames[frameNo].object->GetTopWindow();
			if (top!=NULL)
			{
				((CDialogTool *)top)->HideWin();
			}

			// now, move the window 'inside' the parent, 
			window->SetParent(Frames[frameNo].object);
			// size the window to the frame's size
			move_winptrwindow(window,-1,-1,Frames[frameNo].frame.Width(),Frames[frameNo].frame.Height());

			window->ShowWin();
			
			CRect t;GetButRect(10,t,Frames[frameNo].frame);
			parent->InvalidateRect(t,false);		

			//win->ModifyStyle(0,WS_THICKFRAME|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_MAXIMIZEBOX|WS_CHILD,0);
			//win->ModifyStyleEx(0,WS_EX_MDICHILD,0);   
		}
	}
}
//---------------------------------------------------------------------------

void DialogManager::RemoveWindow(CDialogTool *window)			  // removes window from list
{
	for (long i=0;i<nFrames;i++)
	{
		PtrList<CDialogTool>::Iter iter=Frames[i].windows.Begin();
		while(iter.IsValid())
		{
			if (iter.Obj()==window)
			{

				if (i==0) // is toplevel window
				{
					close_winptrwindow(&window->framedlg);
					// remove from list
					Frames[i].windows.Remove(iter);
					break;
				}
				else
				{
					iter->lastframe=i; // where it came from
					Frames[i].windows.Remove(iter);
					iter=Frames[i].windows.Tail();
					iter->ShowWin();
				}
				CRect t;GetButRect(10,t,Frames[i].frame);
				parent->InvalidateRect(t,false);		

				break;
			}
			iter++;
		}
	}
}

//---------------------------------------------------------------------------	

long DialogManager::FindWindow(CDialogTool *window,bool settop)
{
	for (long i=0;i<nFrames;i++)
	{
		PtrList<CDialogTool>::Iter iter=Frames[i].windows.Begin();
		while(iter.IsValid())
		{
			if (iter.Obj()==window)
			{
				if (settop)
				{
					CDialogTool *obj=GetCurTop(i);
					if (obj!=window){obj->HideWin();window->ShowWin();}
					CRect t;GetButRect(10,t,Frames[i].frame);
					parent->InvalidateRect(t,false);		

				}
				return(i);
				break;
			}
			iter++;
		}
	}
	return(-1);
}

//---------------------------------------------------------------------------	

#define DEF_FRAMETOP   24
#define DEF_FRAMESIDE 	3

bool DialogManager::IsValidWindow(CDialogTool *window)
{
	if (window==NULL)return false;
	for (long i=1;i<nFrames;i++)
	{
		if (Frames[i].windows.Begin().Obj()==window)
		{
			return(false) ; // is bottom window
		}
	}
	return true;
}

//---------------------------------------------------------------------------	


void DialogManager::SetLayout(long type)			 // sets stan
{
	// alright.. 
	// Hide all windows,and move to global frame .. delete all others
	assert(parent!=NULL);

	CRect frame;parent->GetClientRect(frame);

	frame.DeflateRect(0,0,0,0);

	long frametop	=   DEF_FRAMETOP,
		frameside	=   DEF_FRAMESIDE;
	long screenmidx	=   frame.CenterPoint().x,
		 screenmidy =   frame.CenterPoint().y;
	long flags=WS_VISIBLE|WS_CHILD|SS_BLACKFRAME|SS_GRAYRECT;	

	switch(type)
	{
		case DML_4SQ:
		{
			// create 4 CStatic controlls
			CStatic *o;
			nFrames=5;
			
			
			Frames[0].object=parent->GetParent();
			Frames[1].frame.SetRect(0,frametop,screenmidx-frameside,screenmidy);
			Frames[2].frame.SetRect(screenmidx+frameside,frametop,frame.right,screenmidy);
			Frames[3].frame.SetRect(0,screenmidy+frametop,screenmidx-frameside,frame.bottom);
			Frames[4].frame.SetRect(screenmidx+frameside,screenmidy+frametop,frame.right,frame.bottom);
			for (long i=1;i<5;i++)
			{
				o=new CStatic;Frames[i].object=o;o->Create(NULL,flags,Frames[i].frame,parent,15001+i);
				Frames[i].windows.Add(new CNullDialog(o));
			}
		}
		break;
		case DML_4PSQ:
		{
			CStatic *o;
			nFrames=5;
			
			long secxmid=(screenmidx+frameside+frame.right)/2;
			long secymid=(frame.top+screenmidy)/2; 
			
			Frames[0].object=parent->GetParent();
			Frames[1].frame.SetRect(0,frametop,secxmid-frameside,screenmidy);
			Frames[2].frame.SetRect(0,screenmidy+frametop,secxmid-frameside,frame.bottom);
			Frames[3].frame.SetRect(secxmid+frameside,screenmidy+frametop,frame.right,frame.bottom);
			Frames[4].frame.SetRect(secxmid+frameside,frametop,frame.right,screenmidy);

			//Frames[2].frame.SetRect(screenmidx+frameside,frametop,frame.right,screenmidy);
			//Frames[3].frame.SetRect(0,screenmidy+frametop,screenmidx-frameside,frame.bottom);
			//Frames[4].frame.SetRect(screenmidx+frameside,screenmidy+frametop,frame.right,frame.bottom);


			for (long i=1;i<5;i++)
			{
				o=new CStatic;Frames[i].object=o;o->Create(NULL,flags,Frames[i].frame,parent,15001+i);
				Frames[i].windows.Add(new CNullDialog(o));
			}


		}
		break;
		case DML_3SQ:
		{
			// create 4 CStatic controlls
			CStatic *o;
			nFrames=4;
			
			long secxmid=(screenmidx+frameside+frame.right)/2;
			long secymid=(frame.top+screenmidy)/2; 
			
			Frames[0].object=parent->GetParent();
			Frames[1].frame.SetRect(0,frametop,frame.right-frameside,screenmidy);
			Frames[2].frame.SetRect(0,screenmidy+frametop,secxmid-frameside,frame.bottom);
			Frames[3].frame.SetRect(secxmid+frameside,screenmidy+frametop,frame.right,frame.bottom);
			//Frames[4].frame.SetRect(secxmid+frameside,secymid+frametop,frame.right,frame.bottom);

			//Frames[2].frame.SetRect(screenmidx+frameside,frametop,frame.right,screenmidy);
			//Frames[3].frame.SetRect(0,screenmidy+frametop,screenmidx-frameside,frame.bottom);
			//Frames[4].frame.SetRect(screenmidx+frameside,screenmidy+frametop,frame.right,frame.bottom);


			for (long i=1;i<5;i++)
			{
				o=new CStatic;Frames[i].object=o;o->Create(NULL,flags,Frames[i].frame,parent,15001+i);
				Frames[i].windows.Add(new CNullDialog(o));
			}
		}
		break;/*
		case DML_5SP:
		{
			// create 5 window view
			CStatic *o;
			nFrames=6;
				
			Frames[0].object=parent->GetParent();
			long secmid=(screenmidx+frameside+frame.right)/2;
			Frames[1].frame.SetRect(0,frametop,screenmidx-frameside,screenmidy);
			Frames[2].frame.SetRect(screenmidx+frameside,frametop,frame.right,screenmidy);
			Frames[3].frame.SetRect(0,screenmidy+frametop,screenmidx-frameside,frame.bottom);
			Frames[4].frame.SetRect(screenmidx+frameside,screenmidy+frametop,secmid-frameside,frame.bottom);
			Frames[5].frame.SetRect(secmid+frameside,screenmidy+frametop,frame.right,frame.bottom);	
			for(long i=1;i<6;i++)
			{
				o=new CStatic;Frames[i].object=o;o->Create(NULL,flags,Frames[i].frame,parent,15001+i);
				Frames[i].windows.Add(new CNullDialog(o));
			}		
		}
		break;*/
		case DML_5SP:
		{
			// create 5 window view for sequencer
			CStatic *o;
			nFrames=6;
				
			Frames[0].object=parent->GetParent();
			long secxmid=(screenmidx+frameside+frame.right)/2;
			long secymid=(frame.top+screenmidy)/2; 
			Frames[1].frame.SetRect(0,frametop,secxmid-frameside,screenmidy);
			Frames[2].frame.SetRect(secxmid+frameside,frametop,frame.right,secymid);
			Frames[3].frame.SetRect(secxmid+frameside,secymid+frametop,frame.right,screenmidy);
			Frames[4].frame.SetRect(0,screenmidy+frametop,screenmidx-frameside,frame.bottom);
			Frames[5].frame.SetRect(screenmidx+frameside,screenmidy+frametop,frame.right,frame.bottom);
			for(long i=1;i<6;i++)
			{
				o=new CStatic;Frames[i].object=o;o->Create(NULL,flags,Frames[i].frame,parent,15001+i);
				Frames[i].windows.Add(new CNullDialog(o));
			}		
		}
		break;

	}
}

//---------------------------------------------------------------------------	

void DialogManager::GetButRect(long num,CRect &t,CRect &f)
{
	t.bottom= f.bottom-2;
	t.top   = f.top+2;

	switch(num)
	{
	case 0://close
		t.right = f.right - 4;
		t.left  = f.right - 20;
		break;
	case 1://get out
		t.right = f.right - 20;
		t.left  = f.right - 36;
	break;
	case 2:// right
		t.right = f.right - 40;
		t.left  = f.right - 60;
	break;
	case 3:// left
		t.right = f.right - 60;
		t.left  = f.right - 80;
	break;
	case 4:// popup
		t.right = f.right - 80;
		t.left  = f.right - 96;
	break;
	case 5:// caption.. make large for dragging
		t.bottom= f.bottom;
		t.top   = f.top;
		t.right = f.right - 100;
		t.left  = f.left ; 
		break;
	case 6:// mini window view
		t.bottom= f.bottom-2;
		t.top   = f.top+2;
		t.right = f.left + 13;
		t.left  = f.left + 3; // 100 thingies over voor de dialogs
		break;
	case 10:// thing itself
		t=f;
		t.bottom=f.top;
		t.top-=DEF_FRAMETOP-6;
		break;

	}

}

//---------------------------------------------------------------------------	

void DialogManager::DrawSplitters(CDC &dc) // draws window splitters
{
	// draw a blue box on top
	CPen p_line1,p_line2,p_line3,p_axes,p_grid;
	CBrush b_back;
	CFont t_textl,t_texts;
	CGdiObject *oldpen,*oldbrush,*oldfont;
	p_line1.CreatePen(PS_SOLID,1,RGB(255,255,255));		
	p_line2.CreatePen(PS_SOLID,2,RGB(255,0,0));		
	p_line3.CreatePen(PS_SOLID,1,RGB(0,0,0));		
	p_axes.CreatePen(PS_SOLID,1,RGB(192,192,192));		
	p_grid.CreatePen(PS_SOLID,1,RGB(64,64,64));		
	b_back.CreateSolidBrush(RGB(255,255,0));
	t_textl.CreateFont(30,0,0,0,600,false,false,false,0,
						OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
						DEFAULT_QUALITY,DEFAULT_PITCH|FF_SWISS,NULL);
	t_texts.CreateFont(15,0,0,0,500,false,false,false,0,
						OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
						DEFAULT_QUALITY,DEFAULT_PITCH|FF_SWISS,NULL);
	oldpen=dc.SelectObject(&p_axes);oldbrush=dc.SelectObject(&b_back);oldfont=dc.SelectObject(&t_texts);	
	dc.SetBkMode(TRANSPARENT); // transparent text
	dc.SetTextColor(RGB(255,255,255));

	// first drawborders around the windows
	for (long i=1;i<nFrames;i++)
	{
		CRect tgt=Frames[i].frame;tgt.top-=DEF_FRAMETOP-6;
		tgt.InflateRect(1,1,1,1);
		dc.Draw3dRect(tgt,GetSysColor(COLOR_3DDKSHADOW),GetSysColor(COLOR_3DHILIGHT));
		tgt.InflateRect(1,1,1,1);
		dc.Draw3dRect(tgt,GetSysColor(COLOR_3DSHADOW),GetSysColor(COLOR_3DLIGHT));
		tgt.InflateRect(1,1,1,1);
		dc.Draw3dRect(tgt,GetSysColor(COLOR_3DSHADOW),GetSysColor(COLOR_3DLIGHT));

		// now, get the frame where the window shit can be drawn
		CRect drct;
		GetButRect(10,drct,Frames[i].frame);

		// draw shaded window header
		TRIVERTEX        vert[2] ;
		GRADIENT_RECT    gRect;
		vert [0] .x      = drct.left;
		vert [0] .y      = drct.top;
		vert [0] .Red    = 0x0000;
		vert [0] .Green  = 0x0000;
		vert [0] .Blue   = 0x3f00;
		vert [0] .Alpha  = 0x0000;

		vert [1] .x      = drct.right;
		vert [1] .y      = drct.bottom; 
		vert [1] .Red    = 0x8f00;
		vert [1] .Green  = 0x8f00;
		vert [1] .Blue   = 0x8f00;
		vert [1] .Alpha  = 0x0000;

		gRect.UpperLeft  = 0;
		gRect.LowerRight = 1;
		GradientFill(dc.m_hDC,vert,2,&gRect,1,GRADIENT_FILL_RECT_H);

				// draw the dialog title
		CRect but;
		GetButRect(5,but,drct);
		// get window title
		CWnd *top=Frames[i].object->GetTopWindow();
		if (top!=NULL)
		{	
			String d=((CDialogTool *)top)->dialogtitle;
			dc.DrawText((char *)d,drct,DT_CENTER|DT_VCENTER);
		}


		// draw some buttons
		GetButRect(0,but,drct);	dc.DrawFrameControl(but,DFC_CAPTION,DFCS_CAPTIONCLOSE);
		GetButRect(1,but,drct);	dc.DrawFrameControl(but,DFC_SCROLL,DFCS_SCROLLUP);
		GetButRect(2,but,drct);	dc.DrawFrameControl(but,DFC_SCROLL,DFCS_SCROLLRIGHT);
		GetButRect(3,but,drct);	dc.DrawFrameControl(but,DFC_SCROLL,DFCS_SCROLLLEFT);
		GetButRect(4,but,drct);	dc.DrawFrameControl(but,DFC_SCROLL,DFCS_SCROLLDOWN);
				
		// draw some sweet window info here
		GetButRect(6,but,drct);
		//dc.FrameRect(but,&b_back);
		// gather info
		but.DeflateRect(0,1,0,1);
		long windows=0;
		long curwindow=0;
		CDialogTool *twin=GetCurTop(i);
		PtrList<CDialogTool>::Iter iter=Frames[i].windows.Begin()+1;
		while(iter.IsValid()){if (twin==iter.Obj())curwindow=windows;iter++;windows++;}
		// alright now draw sweet cubes in the box, and text
		// cubes are all same size but gradienting from blue to green (10 steps)
		float intp=0;RGBColor a(cSTEEL_BLUE),b(cTOMATO);
		float step=1.0f/(float)windows;
		if (windows>1)
		for (long x=0;x<windows;x++,intp+=step)
		{
			CBrush cube;cube.CreateSolidBrush(a.Interpolate(b,intp).ToBGR_X888());
			dc.FillRect(but,&cube);
			cube.DeleteObject();
			if (x==curwindow)
			{
				CRect d=but;
				d.DeflateRect(2,2,2,2);
				dc.FillRect(d,&b_back);
			}
			but.left+=15;but.right+=15;
		}
	}
	// start drawing that bitch

	// draw the bitchin grid

	dc.SelectObject(oldpen);dc.SelectObject(oldbrush);dc.SelectObject(oldfont);	
	p_line1.DeleteObject();
	p_line2.DeleteObject();
	p_line3.DeleteObject();
	p_axes.DeleteObject();
	p_grid.DeleteObject();
	b_back.DeleteObject();
	t_texts.DeleteObject();
	t_textl.DeleteObject();
	

}

//---------------------------------------------------------------------------	

void OnMouseDown(CPoint &p)		 // Mouse drag modes
{
}

//---------------------------------------------------------------------------	

CDialogTool *DialogManager::GetCurTop(long frame)
{
	CWnd *top=Frames[frame].object->GetTopWindow();
	return(CDialogTool *)top;
}

//---------------------------------------------------------------------------	

void DialogManager::MoveWindow(CDialogTool *window,long targetframe)
{
	if (IsValidWindow(window))
	{
	long srcframe=FindWindow(window);
	if (srcframe==0) // move requested from, window is !=NULL
	{
		window->HideWin();
		window->SetParent(parent); // temporary unparent
		RemoveWindow(window); // window is unparented so remove
		AddWindow(window,targetframe); // add to new parent
	}
	else // go
	{
		window->HideWin();
		window->SetParent(parent); // make sure the dialog is never unparented
		
		RemoveWindow(window); // remove window from list 1
		AddWindow(window,targetframe);
	}
	}
}

//---------------------------------------------------------------------------	

void DialogManager::GetMouse(CPoint &p,long &frame,long &value)
{
	frame=-1;value=-1;
	for (long i=1;i<nFrames;i++)
	{
		for (long j=0;j<7;j++)
		{
			// now, get the frame where the window shit can be drawn
			CRect drct=Frames[i].frame,but;
			drct.bottom=drct.top;
			drct.top-=DEF_FRAMETOP-8;

			GetButRect(j,but,drct);	
			if (but.PtInRect(p)==TRUE)
			{
				frame=i;
				value=j;
				break;
			}

		}
	}
}

//---------------------------------------------------------------------------	

void DialogManager::UpdateCall(CDialogTool *obj)
{
	// find window
	for (long i=0;i<nFrames;i++)
	{
		PtrList<CDialogTool>::Iter iter=Frames[i].windows.Begin();
		while(iter.IsValid())
		{
			if (iter.Obj()==obj)
			{
				if (i==0) // is framedlg, so update title
				{
					obj->dialogtitle.SetWnd(obj->framedlg);
					break;
				}
				else
				{
					CRect t;GetButRect(10,t,Frames[i].frame);
					parent->InvalidateRect(t,false);		
					break;
				}
			}
			iter++;
		}
	}
	
}

//---------------------------------------------------------------------------	

long DialogManager::OnMouseDown(CPoint &p) // checks for mouseclicks somewhere
{
	long i,j;bool redraw=false;
	GetMouse(p,i,j);
	switch(j)
	{
		case -1: // invalid
		break;
		case 0: // post a quit message
		{
			CDialogTool *obj=GetCurTop(i);
			if (IsValidWindow(obj))
				obj->PostMessage(WM_CLOSE); // try to close it
			redraw=true;
		}
		break;
		case 1: // breakout
		{
			CDialogTool *obj=GetCurTop(i);
			if (IsValidWindow(obj))
				MoveWindow(obj,0);
			redraw=true;
		}
		break;
		case 2: // right
		{
			CDialogTool *obj=GetCurTop(i);
			PtrList<CDialogTool>::Iter iter=Frames[i].windows.Find(obj);
			if (iter.IsValid())iter++;
			if (!iter.IsValid()){iter=Frames[i].windows.Begin();iter++;}
			if (iter.IsValid()){obj->HideWin();iter->ShowWin();}
			redraw=true;
		}
		break;
		case 3: // left
		{
			CDialogTool *obj=GetCurTop(i);
			PtrList<CDialogTool>::Iter iter=Frames[i].windows.Find(obj);
			if (iter.IsValid())iter--;
			if (!iter.IsValid() || iter==Frames[i].windows.Begin()){iter=Frames[i].windows.Tail();}
			if (iter.IsValid()){obj->HideWin();iter->ShowWin();}
			redraw=true;
		}
		break;
		case 4: // popdown open popup menu
		{
			// aight now go get them titles
			PtrList<CDialogTool>::Iter iter=Frames[i].windows.Begin()+1;
			long lines=0;PopupMenu m;
			while(iter.IsValid())
			{
				m.AddString(iter->dialogtitle,lines,false);
				iter++,lines++;
			}
			if (lines!=0)
			{
				CPoint lp=p;
				parent->ClientToScreen(&lp);
				long val=m.Show(parent,&lp);
				if (val>=0 && val<lines)
				{
					iter=Frames[i].windows.Begin()+1;
					while (val>0 && iter.IsValid())iter++,val--;
					if (iter.IsValid())
					{
						CDialogTool *obj=GetCurTop(i);
						if (obj!=iter.Obj())
						{
							obj->HideWin();
							iter->ShowWin();
						}
					}
				};
			}
			
		}
		break;
		case 5: // text .. you can start a drag from here
		{
			
			bool found=false;

			// check if the mouse may be on a window 'square'
			CRect drct,but;
			GetButRect(10,drct,Frames[i].frame);
			GetButRect(6,but,drct);
			
			long windows=0;
			PtrList<CDialogTool>::Iter iter=Frames[i].windows.Begin()+1;
			for(;iter.IsValid();iter++,windows++);
			iter=Frames[i].windows.Begin()+1;
			for (long x=0;x<windows;x++)
			{
				if (but.PtInRect(p)==TRUE && iter.IsValid()) // window thingy selected
				{
					CDialogTool *obj=GetCurTop(i);
					if (obj!=iter.Obj()){
						obj->HideWin();iter->ShowWin();}
					redraw=true;
					found=true;break;
				}
				if (iter.IsValid())iter++;
				but.left+=15;but.right+=15;
			}
			if (!found)	return i+10; // start drag
		}
	}
	if (redraw)
	{
		CRect t;GetButRect(10,t,Frames[i].frame);
		parent->InvalidateRect(t,false);		
	}
	
	return(0);
}
//---------------------------------------------------------------------------	

void DialogManager::WindowDrop(CPoint &p,long srcframe,bool swap)
{
	bool found=false;
	for (long i=1;i<nFrames;i++)
	{
		CRect whole=Frames[i].frame;
		whole.top-=DEF_FRAMETOP-8;
		if (whole.PtInRect(p)==TRUE)
		{
			if (i!=srcframe)
			{
				CDialogTool *srcwin=GetCurTop(srcframe);
				if (swap)MoveWindow(GetCurTop(i),srcframe);
				MoveWindow(srcwin,i);
			}
			found=true;
			break;
		}
	}
	if (!found)MoveWindow(GetCurTop(srcframe),0); // move out
}