/*
---------------------------------------------------------------------------
|					Function to assembler compiler
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/
#include "Library.h"
#include "AsmObj.h"
#include "FuncAsm.h"

//---------------------------------------------------------------------------
bool FuncAsm::HighPrec=false;

FuncAsm::FuncAsm()
{
	CalcPtr=NULL;
}; 

//---------------------------------------------------------------------------

FuncAsm::~FuncAsm()
{
};


//---------------------------------------------------------------------------

bool FuncAsm::Create(String &func,long cookie)
{
	if ( cookie!=Cookie || !IsValid() || cookie==-1)
	{
		Cookie=cookie;
		d.IsCompiled=false;
		if (FuncTree::Create(func))
		{
			Compile();
			return(true);
		}
	}
	return(false);
}

//---------------------------------------------------------------------------

void FuncAsm::Copy(FuncTree *target,long cookie)
{
	bool docompile=false;
	if ( cookie!=Cookie || cookie==-1) docompile=true;

	if (IsValid())
	{
		// copy ende compile
		FuncTree::Copy(target,cookie); 
		if(docompile)((FuncAsm *)target)->Compile();
		Cookie=cookie;
	}
}

//---------------------------------------------------------------------------

FuncTreeNode *FuncAsm::CreateSubnode(FuncTreeNode *parent)
{
	return new FuncAsmNode(this,parent);
};

//---------------------------------------------------------------------------

bool FuncAsm::Compile()
{
	Lock.Lock();
	if (IsValid() && !IsVar)
	{
		d.Reset(); // resets function data, also creates memory (if not alread)
		// add init data
		CalcPtr=(CalcPtrType)d.fdata;
		d.PUSHAD();
		//if (!HighPrec)
			td_controlword=FCW_PRECISION32|FCW_ROUNDUP|FCW_ALLEX;
		//else
		//	td_controlword=FCW_PRECISION64|FCW_ROUNDUP|FCW_ALLEX;
		d.FLDCW(&td_controlword); 
		// reset the random counter
		((FuncAsmNode *)basenode)->CompileNode(d,&final);
		d.POPAD();
		d.RET();// ret
		d.Done();
	}
	// set compile to 0
	else {d.Reset();CalcPtr=NULL;}
	Lock.Unlock();
	return(true);
};// compiles the functree into asm

//---------------------------------------------------------------------------

double FuncAsm::Calculate()		// calcs outcome
{
	Lock.Lock();
	if (IsValid())
	{
		if (IsVar)
		{
			Lock.Unlock();
			return variable;
		}
		if (d.IsCompiled && CalcPtr!=NULL)
		{
			CalcPtr();
			Lock.Unlock();
			return(final);
		}
	}
	Lock.Unlock();
	return FuncTree::Calculate();
};

//---------------------------------------------------------------------------
//- Node stuff
bool FuncAsmNode::CompileNode(AsmObj &d,double *target) // the actual node compiling
{
	bool targetloaded=false;
	// if node only contains a value
	FuncAsm *td=(FuncAsm *)base;

	if (IsVariable && specop==-1)
	{
		//String t;t.FormatSpec(variable);t.MsgBox();
		d.FLD64(&base->varlist[varindex].value);
		if (neg)d.FCHS();
		d.FSTP64(target);						
		return(true);
	}
	if (!IsVariable) // means it needs to go to it's children for variables
	{
		if (child!=NULL)
		{
			((FuncAsmNode *)child)->CompileNode(d,&a);		
			if (child->next!=NULL)
			{
				((FuncAsmNode *)child->next)->CompileNode(d,&b);		
			}
			else
			{
				d.FLDZ();	d.FSTP64(&b);
			}
		}
		else
		{
			d.FLDZ();	d.FST64(&a); d.FSTP64(&b);
		}
		// now combine a and b using some kind of op
		switch(otype)
		{
			case _OPT_NOP:	
				d.FLD64(&a);
				break;
			case _OPT_ADD:
				d.FLD64(&b);
				d.FLD64(&a);// load a and b
				d.FADDP();
				targetloaded=true;
				break;
			case _OPT_SUB:
				d.FLD64(&b);
				d.FLD64(&a);
				d.FSUBP();
				targetloaded=true;
				break;
			case _OPT_MUL:
				//String t;t.Format("%f %f",a,b);t.MsgBox();
				d.FLD64(&a);
				d.FLD64(&b);// load a and b
				d.FMULP();
				targetloaded=true;
				break;
			case _OPT_DIV:
				d.FLD64(&b);
				d.FLD64(&a);
				d.FDIVP();
				targetloaded=true;
				break;
			case _OPT_MOD:
				d.FLD64(&a);
				d.FLD64(&b);
				d.FPREM();
				d.FXCH();
				d.FSTP(0);
				
				targetloaded=true;
				break;
			case _OPT_NSQR: // make y 1/y
			case _OPT_POW:
			{
				if (otype==_OPT_NSQR)
				{
	/*6*/		d.FLD64(&a);   // x
	/*6*/		d.FLD64(&b);   // y x
	/*2*/		d.FLD1();	   // 1 y x
	/*2*/ 		d.FDIVRP();	   // 1/y x
				}
				else
				{
	/*6*/		d.FLD64(&a);   // x
	/*6*/		d.FLD64(&b);   // y x
				};

	///*2*/		d.FXCH();       // x y .. check x against zero
	/*2*/		d.FTST();		// Test against zero
	/*2*/		d.FNSTSWAX();   // flags ->ax
	/*1*/		d.SAHF();		// move in ah
	///*2*/		d.FXCH();		// swap back
	/*2*/		d.JBE(/*NEGATIVE*/20);		// jump to NEGATIVE

				// DO POSITIVE
	/*2*/		d.FYL2X();	   // 2log(x^y)
	/*2*/		d.FLD1();	   // 1 n
	/*2*/		d.FLDST(1);	   // n 1 n
	/*2*/		d.FPREM();    // rn 1 n
	/*2*/		d.F2XM1();	   // 2^rn-1 1 n
	/*2*/		d.FADDP();	   // 2^rn n
	/*2*/		d.FSCALE();    // 2^int(n)*2^rn n
	/*2*/		d.FXCH();	   // swaps
	/*2*/		d.FSTP(0);     // loses the 'n'
	/*2*/		d.JMP(71/*END*/);
	// 20			
				// NEGATIVE
				// check power-modularity
	/*2*/		d.FLDST(1);		// x y x
	/*2*/		d.FLD1();       // 1 x y x
	/*2*/		d.FLD1();		// 1 1 x y x
	/*2*/		d.FADDP();		// 2 x y x
	/*2*/		d.FXCH();		// x 2 y x
	/*2*/		d.FPREM();		// xmod2 2 y x
	/*2*/		d.FLD1();		// 1 xmod2 2 y x
	/*2*/		d.FSUBP();		// xmod2-1 2 y x
	/*2*/		d.FTST();		// Test against zero
	/*2*/		d.FNSTSWAX();   // flags ->ax
	/*2*/		d.FSTP(0);		// 2 y x
	/*2*/		d.FSTP(0);		// y x
	/*1*/		d.SAHF();		// move in ah
	/*2*/		d.JAE(22/*NEGFLIP*/);		// if >0, do neg. flip after
	// 27
				// First make it positive
	/*2*/		d.FCHS();
	/*2*/		d.FYL2X();	   // 2log(x^y)
	/*2*/		d.FLD1();	   // 1 n
	/*2*/		d.FLDST(1);	   // n 1 n
	/*2*/		d.FPREM();    // rn 1 n
	/*2*/		d.F2XM1();	   // 2^rn-1 1 n
	/*2*/		d.FADDP();	   // 2^rn n
	/*2*/		d.FSCALE();    // 2^int(n)*2^rn n
	/*2*/		d.FXCH();	   // swaps
	/*2*/		d.FSTP(0);     // loses the 'n'
	/*2*/		d.JMP(22/*END*/);
	//22 
				// NEGFLIP
				// First make it positive
	/*2*/		d.FCHS();
	/*2*/		d.FYL2X();	   // 2log(x^y)
	/*2*/		d.FLD1();	   // 1 n
	/*2*/		d.FLDST(1);	   // n 1 n
	/*2*/		d.FPREM();    // rn 1 n
	/*2*/		d.F2XM1();	   // 2^rn-1 1 n
	/*2*/		d.FADDP();	   // 2^rn n
	/*2*/		d.FSCALE();    // 2^int(n)*2^rn n
	/*2*/		d.FXCH();	   // swaps
	/*2*/		d.FSTP(0);     // loses the 'n'
	/*2*/		d.FCHS();	   // make negative
	//22
				targetloaded=true;
			}
			break;
			case _OPT_FIX:
				d.FLD64(&a); // y
	/*2*/		d.FTST();	   // y x .. check y against zero
	/*2*/		d.FNSTSWAX();  // flags ->ax
	/*1*/		d.SAHF();	
	/*2*/		d.JZ(22);  // jump zero
	/*6*/		d.FLD64(&b);   // x y
	/*2*/		d.FDIV();      // x/y y
	/*6*/		d.FLD64(&ftbl.td_onehalf); // 0.5 x/y y
	/*2*/		d.FSUBP();		// x/y-0.5 y
	/*2*/		d.FRNDINT();    // ceil(x/y-0.5) y
	/*2*/		d.FMULP();		// fixed value
				d.JMP(12);
				d.FSTP64(target); // stackempty
				d.FLD64(&b);
				
				targetloaded=true;
				break;
			case _OPT_DST:
	/*6*/		d.FLD64(&a);   // y
	/*2*/		d.FMUL(0);	   // y^2
	/*6*/		d.FLD64(&b);   // x y^2
	/*2*/		d.FMUL(0);	   // x^2 y^2
	/*2*/		d.FADDP();	   // x^2+y^2

				d.FSTP32(&td->td_longholder); // load in long
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.MOV32R2R(SRC_EBX,B_ECX);
				d.SHR32RI(B_ECX,MF_SQRT_GET_FRACT);
				d.AND32RI(B_ECX,MF_SQRT_MASK);
				d.SHL32RI(B_ECX,2); // float table.. *4
				d.ADD32RI(B_ECX,(long)mfTab.SqrtFrac);
				d.MOV32R2R(SRCM_EAX,B_ECX);
				d.SHR32RI(B_EBX,1);
				d.AND32RI(B_EBX,MF_EXPONENT_MASK);
				d.ADD32RR(SRC_EAX,B_EBX);
				d.MOV32R2M(R_EAX,&td->td_longholder);
				d.FLD32(&td->td_longholder);

//	/*2*/		d.FSQRT();	   // sqrt(x^2+y^2);
				targetloaded=true;			
				break;
			case _OPT_ANG:
				d.FLD64(&b);		// x
				d.FLD64(&a);		// y
				d.FPATAN();			// atan(x/y)
				targetloaded=true;			
				break;
			case _OPT_LOG:
				d.FLD1();	 // 1
				d.FLD64(&a); // a 1
				d.FYL2X();   // 1log(a);
				d.FLD1();	 // 1 1log(a);
				d.FLD64(&b); // b 1 1log(a);
				d.FYL2X();   // 1log(b) 1log(a);
				d.FDIVP();   // log(a)/log(b)
				targetloaded=true;
				break;
			}
	}
	if (specop!=-1)
	{
		if (IsVariable)
		{
			d.FLD64(&base->varlist[varindex].value);
			targetloaded=true;
		}
		else
		{
			if (!targetloaded)d.FLD64(target);
			targetloaded=true;
		}
		switch(specop)
		{
			case _OPO_SIN:
				d.FSIN();
				break;
			case _OPO_COS:
				d.FCOS();
				break;
			case _OPO_TAN:
				d.FPTAN();
				d.FSTP64(target); // remove the 1
				break;
			case _OPO_LOG:
				d.FLD1();
				d.FXCH();
				d.FYL2X();
				d.FLDL2T();
				d.FDIVP();
				break;
			case _OPO_LN:
				d.FLD1();
				d.FXCH();
				d.FYL2X();
				d.FLDL2E();
				d.FDIVP();
				break;
			case _OPO_ABS:
				d.FABS();
				break;
			case _OPO_SQR:
				d.FSQRT();
				break;
			case _OPO_RAN: // table random
				d.FSTP32(&td->td_longholder);
				d.MOV32M2R(R_ECX,&mfTab.RandomCounter);
				d.INC32(B_ECX);
				d.AND32RI(B_ECX,MF_RANDOM-1);
				d.MOV32R2M(R_ECX,&mfTab.RandomCounter);
				d.SHL32RI(B_ECX,2); // 4 bytes per float
				d.ADD32RI(B_ECX,(dword)mfTab.Random); // offset of random table
				d.FLD32R(B_ECX);		// load value from table*/
			break;
			case _OPO_RND: // hash random
				d.FSTP32(&td->td_longholder); // load in long
				d.MOV32M2R(R_EBX,&td->td_longholder);
				d.MOV32I2R(B_EAX,8192);
				d.MUL32EAXR(B_EBX); // s*8192
				d.XOR32RR(SRC_EBX,B_EAX);
				d.MOV32R2R(SRC_EBX,B_EAX);
				d.MUL32EAXR(B_EBX); // s*s
				d.MOV32I2R(B_ECX,15731); 
				d.MUL32EAXR(B_ECX); //s*s*15731
				d.MOV32I2R(B_ECX,789221);
				d.ADD32RR(SRC_EAX,B_ECX); //s*s*15731+789221
				d.MUL32EAXR(B_EBX);//s*(s*s*15731+789221)
				d.MOV32I2R(B_ECX,1376312589);
				d.ADD32RR(SRC_EAX,B_ECX);
				d.AND32RI(B_EAX,0x7fffffff);
				d.MOV32R2M(R_EAX,&td->td_longholder);
				d.FILD32(&td->td_longholder);
				d.FLD64(&ftbl.td_random);
				d.FDIVP();
				d.FLD1();
				d.FSUBRP(); // DONE
				break;
			case _OPO_ATN:
				d.FLD1();
				d.FPATAN();
				break;
			case _OPO_INT:
				d.FLD64(&ftbl.td_onehalf);
				d.FSUBP();
				d.FRNDINT();
				break;		
			case _OPO_RAD:
				d.FLD64(&ftbl.td_trad);
				d.FMULP();
				break;
			case _OPO_DEG:
				d.FLD64(&ftbl.td_tdeg);
				d.FMULP();
				break;
			case _OPO_ASN: 
				d.FLD1();
				d.FADDP();
				d.FLD64(&ftbl.td_arctableindex);	// loads 1024/2
				d.FMULP();
				d.FISTP32(&td->td_longholder); // holds long variable
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.AND32RI(B_ECX,(MF_ANGLES-1));	// leave the ANGLES mask part
				d.SHL32RI(B_ECX,2);			// shift two bits to the left (
				d.ADD32RI(B_ECX,(dword)mfTab.ASine); // add address of arc-sine table
				d.FLD32R(B_ECX);			// load 32 bit float value at offset
				break;
			case _OPO_ACS: 
				d.FLD1();
				d.FADDP();
				d.FLD64(&ftbl.td_arctableindex);	// loads 1024/2
				d.FMULP();
				d.FISTP32(&td->td_longholder); // holds long variable
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.AND32RI(B_ECX,(MF_ANGLES-1));	// leave the ANGLES mask part
				d.SHL32RI(B_ECX,2);			// shift two bits to the left (
				d.ADD32RI(B_ECX,(dword)mfTab.ACosine); // add address of arc-sine table
				d.FLD32R(B_ECX);			// load 32 bit float value at offset
				break;
			case _OPO_FAC: 
				d.FISTP32(&td->td_longholder);
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.SHL32RI(B_ECX,3); // 8 bytes (double);
				d.AND32RI(B_ECX,(FACTABLESIZE-1));	// leave the ANGLES mask part
				d.ADD32RI(B_ECX,(dword)ftbl.factable);
				d.FST64(&td->td_longholder);
				d.FLD64R(B_ECX);
				break;
			case _OPO_SUM: 
				d.FISTP32(&td->td_longholder);
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.SHL32RI(B_ECX,3); // 8 bytes (double);
				d.AND32RI(B_ECX,(SUMTABLESIZE-1));	// leave the ANGLES mask part
				d.ADD32RI(B_ECX,(dword)ftbl.sumtable);
				d.FST64(&td->td_longholder);
				d.FLD64R(B_ECX);
				break;		
			case _OPO_SAW: // saw wave
	/*1*/		d.FLD64(&ftbl.td_tableindex);	// loads 1024/2pi
				d.FMULP();					// re-indexed value..
				d.FISTP32(&td->td_longholder); // holds long variable
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.AND32RI(B_ECX,(MF_ANGLES-1));// leave the 1024 part
											// use ECX to calc index into fsn table
				d.SHL32RI(B_ECX,2);			// shift two bits to the left (
				d.ADD32RI(B_ECX,(dword)mfTab.Saw); // add address of saw table
				d.FLD32R(B_ECX);			// load 32 bit float value at offset
				break;		
			case _OPO_BLK: // block wave
				// check if y < 0 ifso make y > 0 and set some 'invert byte'
				//d.FTST();	   // y x .. check y against zero
				//d.FNSTSWAX();  // flags ->ax
				//d.MOV32R2M(R_EAX,&td->td_signhold);
				//d.SAHF();	
				//d.JAE(58);  // jump SKIPNEG

				//d.FLD64(&ftbl.td_pi116);	// loads 1/pi
				//d.FMULP();			// re-indexed value..
				//d.FISTP32(&td->td_longholder); // holds long variable
				//d.MOV32M2R(R_ECX,&td->td_longholder);
				//d.ADD32RI(B_ECX,1);
				//d.AND32RI(B_ECX,1);// leaves 1 or 0 the 1024 part
				//d.MOV32R2M(R_ECX,&td->td_longholder);
				//d.FILD32(&td->td_longholder);
				//d.FLD64(&ftbl.td_two);
				//d.FMULP();
				//d.FLD1();
				//d.FSUBP();
				//d.FCHS();
				//d.JMP(56);
	// 6 2 6 6 6 6 6 6 6 2 2 2 2 2  8*6 + 6*2
				// SKIPNEG
				d.FLD64(&ftbl.td_pi116);	// loads 1/pi
				d.FMULP();			// re-indexed value..
				d.FISTP32(&td->td_longholder); // holds long variable
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.ADD32RI(B_ECX,1);
				d.AND32RI(B_ECX,1);// leaves 1 or 0 the 1024 part
				d.MOV32R2M(R_ECX,&td->td_longholder);
				d.FILD32(&td->td_longholder);
				d.FLD64(&ftbl.td_two);
				d.FMULP();
				d.FLD1();
				d.FSUBP();
				d.FCHS();
	// 6 2 6 6 6 6 6 6 6 2 2 2 2 2  8*6 + 4*2

				break;		
			case _OPO_FSN: // fast sine ( about twice as fast as normal sine)
	/*1*/		d.FLD64(&ftbl.td_tableindex);	// loads 1024/2pi
				d.FMULP();					// re-indexed value..
				d.FISTP32(&td->td_longholder); // holds long variable
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.AND32RI(B_ECX,(MF_ANGLES-1));// leave the 1024 part
											// use ECX to calc index into fsn table
				d.SHL32RI(B_ECX,2);			// shift two bits to the left (
				d.ADD32RI(B_ECX,(dword)mfTab.Sine); // add address of sine table
				d.FLD32R(B_ECX);			// load 32 bit float value at offset
	//13 clocks total
				break;		
			case _OPO_FCS: // fast co-sine
	/*1*/		d.FLD64(&ftbl.td_tableindex);	// loads 1024/2pi
	/*1*/		d.FMULP();					// re-indexed value..
	/*6*/		d.FISTP32(&td->td_longholder); // holds long variable
	/*1*/		d.MOV32M2R(R_ECX,&td->td_longholder);
				d.ADD32RI(B_ECX,MF_ANGLES/4); // make it a cosine
	/*1*/		d.AND32RI(B_ECX,(MF_ANGLES-1));	// leave the ANGLES mask part
											// use ECX to calc index into fsn table
	/*1*/		d.SHL32RI(B_ECX,2);			// shift two bits to the left (*4)
	/*1*/		d.ADD32RI(B_ECX,(dword)mfTab.Sine); // add address of sine table
	/*1*/		d.FLD32R(B_ECX);			// load 32 bit float value at offset
	//14 clocks total
				break;		
			case _OPO_FRAC:
				d.FLDST(0); // x x
				d.FRNDINT();// xc x
				d.FSUBP();
				d.FLD1();
				d.FADDP();
				break;
			case _OPO_TWO:
				d.FLD1();
				d.FADDP();
				d.FLD64(&ftbl.td_two);
				d.FDIVP();
				break;
			case _OPO_FSQR: // whoo fast square root
				d.FSTP32(&td->td_longholder); // load in long
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.MOV32R2R(SRC_EBX,B_ECX);
				d.SHR32RI(B_ECX,MF_SQRT_GET_FRACT);
				d.AND32RI(B_ECX,MF_SQRT_MASK);
				d.SHL32RI(B_ECX,2); // float table.. *4
				d.ADD32RI(B_ECX,(long)mfTab.SqrtFrac);
				d.MOV32R2R(SRCM_EAX,B_ECX);
				d.SHR32RI(B_EBX,1);
				d.AND32RI(B_EBX,MF_EXPONENT_MASK);
				d.ADD32RR(SRC_EAX,B_EBX);
				d.MOV32R2M(R_EAX,&td->td_longholder);
				d.FLD32(&td->td_longholder);
				//mfFloat(mfTab.SqrtFrac[(mfBin(f) >> MF_SQRT_GET_FRACT) & MF_SQRT_MASK] + ((mfBin(f) >> 1) & MF_EXPONENT_MASK)); }
				break;
			case _OPO_OOSQR: // one over square root
				d.FSTP32(&td->td_longholder); // load in long
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.MOV32R2R(SRC_EBX,B_ECX);
				d.SHR32RI(B_ECX,MF_FRACTION_APPROX_SHIFT);
				d.AND32RI(B_ECX,MF_FRACTION_APPROX_MASK);
				d.SHL32RI(B_ECX,2); // float table.. *4
				d.ADD32RI(B_ECX,(long)mfTab.OoSqrtFrac);
				d.MOV32R2R(SRCM_EAX,B_ECX);
				d.MOV32R2R(SRC_ECX,B_EBX);
				d.SHR32RI(B_ECX,MF_GET_EXPONENT_SHIFT);
				d.AND32RI(B_ECX,MF_EXPONENT_SIGN_MASK);
				d.SHL32RI(B_ECX,2); // float table.. *4
				d.ADD32RI(B_ECX,(long)mfTab.OoSqrtExp);
				d.MOV32R2R(SRCM_EBX,B_ECX);
				d.MOV32R2M(R_EAX,&td->td_longholder);
				d.FLD32(&td->td_longholder);				
				d.MOV32R2M(R_EBX,&td->td_longholder);
				d.FLD32(&td->td_longholder);				
				d.FMULP();
				break;
				//{ return (mfTab.OoSqrtFrac [(mfBin(f) >> MF_FRACTION_APPROX_SHIFT) & MF_FRACTION_APPROX_MASK] * mfTab.OoSqrtExp [(mfBin(f) >> MF_GET_EXPONENT_SHIFT) & MF_EXPONENT_SIGN_MASK]); }
			case _OPO_OO:
				d.FSTP32(&td->td_longholder); // load in long
				d.MOV32M2R(R_ECX,&td->td_longholder);
				d.MOV32R2R(SRC_EBX,B_ECX);
				d.SHR32RI(B_ECX,MF_FRACTION_APPROX_SHIFT);
				d.AND32RI(B_ECX,MF_FRACTION_APPROX_MASK);
				d.SHL32RI(B_ECX,2); // float table.. *4
				d.ADD32RI(B_ECX,(long)mfTab.OoFrac);
				d.MOV32R2R(SRCM_EAX,B_ECX);
				d.MOV32R2R(SRC_ECX,B_EBX);
				d.SHR32RI(B_ECX,MF_GET_EXPONENT_SHIFT);
				d.AND32RI(B_ECX,MF_EXPONENT_SIGN_MASK);
				d.SHL32RI(B_ECX,2); // float table.. *4
				d.ADD32RI(B_ECX,(long)mfTab.OoExp);
				d.MOV32R2R(SRCM_EBX,B_ECX);
				d.MOV32R2M(R_EAX,&td->td_longholder);
				d.FLD32(&td->td_longholder);				
				d.MOV32R2M(R_EBX,&td->td_longholder);
				d.FLD32(&td->td_longholder);				
				d.FMULP();
				break;
			case _OPO_FLOG:
				d.FABS();
				d.FLD1();
				d.FADDP();
				d.FLD1();	 // 1 a
				d.FXCH();	 // a 1
				d.FYL2X();   // 1log(a);
				//d.FLD1();	 // 1 1log(a);
				d.FLD64(&ftbl.td_flog); // b 1 1log(a);
				//d.FYL2X();   // 1log(b) 1log(a);
				d.FDIVRP();   // log(a)/log(b)
				break;
			case _OPO_IN:  // normal integration .. 
				//log("compiling IN");
				d.FLD64(&value);
				d.FADDP();		 // add to current thingy
				d.FST64(&value); // store back to value
				break;
			case _OPO_IT: // time scaled integration (delta t)
			{
				//log("Compiling IT");
				long idx='t'-97;
				assert(td->variableindex[idx]!=-1);
				// substract new 't' from old 't'
				d.FLD64(&lastt);
				d.FLD64(&td->varlist[td->variableindex[idx]].value);
				d.FST64(&lastt); // store new time 
				d.FSUBP();	// t-lastt
				d.FABS();	// abs()
				d.FMULP();	// value*fabs(t-lastt)
				d.FLD64(&value); // load old value
				d.FADDP(); // add up
				d.FST64(&value); // store new value
			}
			break;
			case _OPO_IDIF:
			{
				// compare 	lastval to value, and only integrate if different
				d.FLDST(0);			// make extra copy of the value  
				d.FLD64(&lastval);  // lv v v 
				d.FXCH();			// v lv v
				d.FST64(&lastval);	// store v in lastval
				d.FSUBP();			// v-lv v
				d.FTST();			// lv-v against zero
				d.FNSTSWAX();		// flags ->ax
				d.MOV32R2M(R_EAX,&td->td_signhold);
				d.SAHF();			
				d.JE(18);			// are equal.. don't integrate
				d.FSTP(0);			// lose the difference
				d.FLD64(&value);    // load integrated value
				d.FADDP();			// add to current thingy
				d.FST64(&value);	// store back to value
				d.JMP(10);			// skip over non-integrate code
				d.FSTP(0);			// clear stack
				d.FSTP(0);
				d.FLD64(&value);	// load old value
			}
			break;
			case _OPO_DN: // differencing
			{
				d.FLD64(&lastval);
				d.FXCH();
				d.FST64(&lastval); // write new val in old value
				d.FSUBRP();		  // substract and done
			};
			break;
			case _OPO_DT:
			{
				long idx='t'-97;
				assert(td->variableindex[idx]!=-1);

				d.FLD64(&lastval);
				d.FXCH();
				d.FST64(&lastval); // write new val in old value
				d.FSUBRP();		  // substract and done
				d.FLD64(&lastt);
				d.FLD64(&td->varlist[td->variableindex[idx]].value);
				d.FST64(&lastt); // store new time 
				d.FSUBP();	// t-lastt
				d.FABS();	// abs()
				d.FMULP();	// value*fabs(t-lastt)
			}
			break;
			case _OPO_IZ: // check against zero
			{
				d.FTST();	   // check y against zero
				d.FNSTSWAX();  // flags ->ax
				d.MOV32R2M(R_EAX,&td->td_signhold);
				d.SAHF();	
				d.JE(4);  // jump over
				d.FSTP(0);
				d.FLD1();
			};
			break;
			case _OPO_SIGN:
			{
				d.FTST();	   // check y against zero
				d.FNSTSWAX();  // flags ->ax
				d.MOV32R2M(R_EAX,&td->td_signhold);
				d.SAHF();	
				d.JB(4);  // jump over
				d.FSTP(0); // loose value
				d.FLD1();
			};
			break;
			case _OPO_VK:{long idx='k'-97;if (td->variableindex[idx]!=-1)d.FST64(&td->varlist[td->variableindex[idx]].value);}break;
			case _OPO_VL:{long idx='l'-97;if (td->variableindex[idx]!=-1)d.FST64(&td->varlist[td->variableindex[idx]].value);}break;
			case _OPO_VM:{long idx='m'-97;if (td->variableindex[idx]!=-1)d.FST64(&td->varlist[td->variableindex[idx]].value);}break;
			case _OPO_VG:{long idx='g'-97;if (td->variableindex[idx]!=-1)d.FST64(&td->varlist[td->variableindex[idx]].value);}break;
			case _OPO_VH:{long idx='h'-97;if (td->variableindex[idx]!=-1)d.FST64(&td->varlist[td->variableindex[idx]].value);}break;
			case _OPO_VI:{long idx='i'-97;if (td->variableindex[idx]!=-1)d.FST64(&td->varlist[td->variableindex[idx]].value);}break;
		}
		if (neg)d.FCHS();
		d.FSTP64(target);
		targetloaded=false;
	}
	else if (neg)
	{
		if (!targetloaded)d.FLD64(target);
		d.FCHS();
		d.FSTP64(target);
		targetloaded=false;
	};
	if (targetloaded)d.FSTP64(target);
	return(true);
};

//---------------------------------------------------------------------------

