/*
---------------------------------------------------------------------------
|						Function tree objects
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
Requirements:

Changes:
	11-19-98	Initial file setup
*/

#include "Library.h"


//----------------------------------------------------------
// Table generation
ftable::ftable()
{
	double i,j;
	factable[0]=0;for (i=1,j=1;i<FACTABLESIZE;i++)j*=i,factable[(long)i]=j;
	for (i=0,j=0;i<SUMTABLESIZE;i++)j+=i,sumtable[(long)i]=j;
	td_onehalf=0.5;td_trad=TRAD;td_tdeg=TDEG;
	td_two=2.0;td_three=3.0;td_four=4.0;td_pi2=PI2;td_pi12=PI12;td_pi116=1.0/PI;

	td_tableindex=(float)MF_ANGLES/(PI2);
	td_random=1073741824.0;
	td_flog=log(1.2f)/log(2.0f);
	td_arctableindex=(float)MF_ANGLES/2.0f;
}

ftable ftbl;

//----------------------------------------------------------
//----------------------------------------------------------
//- Function tree stuff

FuncTree::FuncTree(): Lock()
{
	Cookie=0;
	functxt="";
	basenode=NULL;
	varcount=0;
	b_IsValid=false;
	for (long i=0;i<26;i++)variableindex[i]=-1;

}

//----------------------------------------------------------

FuncTree::~FuncTree()
{
	ClearTree();
}

//----------------------------------------------------------

FuncTreeNode *FuncTree::CreateSubnode(FuncTreeNode *parent)
{
	return new FuncTreeNode(this,parent);
}

//----------------------------------------------------------

FuncTreeNode *FuncTreeNode::Copy(FuncTree *root,FuncTreeNode *parent)
{
	FuncTreeNode *tgt=root->CreateSubnode(parent);
	tgt->neg		 =neg;
	tgt->otype		 =otype;
	tgt->specop		 =specop;
	tgt->IsVariable	 =IsVariable;
	tgt->varindex	 =varindex;
	tgt->value		 =value;
	tgt->lastt		 =lastt;
	tgt->lastval	 =lastval;

	if (child!=NULL)
	{
		((FuncTreeNode *)child)->Copy(root,tgt);	
		if (child->next!=NULL)
		{
			((FuncTreeNode *)child->next)->Copy(root,tgt);		
		}

	}
	return tgt;
}

//----------------------------------------------------------

void FuncTree::Copy(FuncTree *target,long cookie)
{
	if (IsValid())
	{
		long i;
		Lock.Lock();
		{
			if (IsVar)target->CreateV(variable);
			else // go recursively copy the tree, and then the variables
			{
				target->Lock.Lock();
				if (( cookie!=Cookie || cookie==-1) || target->varcount!=varcount)
				{
					target->ClearTree();
					target->basenode	=basenode->Copy(target,NULL);
					target->functxt		=functxt;
					target->b_IsValid	=b_IsValid;
					target->Cookie		=cookie;
					target->varcount	=varcount;
					target->variable	=variable;
					target->varlist.Resize(64);
					target->IsVar	    =false;

					for (i=0;i<varcount && i<target->varcount;i++)target->varlist[i].IsConstant=varlist[i].IsConstant,
																  target->varlist[i].value     =varlist[i].value;
					for (i=0;i<26;i++)target->variableindex[i]=variableindex[i];
				}
				target->Lock.Unlock(); 
			}
		}


		Lock.Unlock();
		Cookie=cookie;
	}
}

//----------------------------------------------------------

void FuncTree::ClearTree()
{
	if (IsValid())
	{
		if (basenode!=NULL)
		{
			basenode->delete_children();
			delete basenode;
		}
		basenode=NULL;
		// reset amazing index
		for (long i=0;i<26;i++)variableindex[i]=-1;
	}
}

//----------------------------------------------------------

bool FuncTree::CreateV(double v) // create function with only one value
{
	Lock.Lock();
	functxt=v; //this has to be done thoug.. for the eq editor thingies
	ClearTree();
	IsVar=true;
	variable=v;
	b_IsValid=true;
	Lock.Unlock();
	return true;
} 

//----------------------------------------------------------
 
bool FuncTree::Create(String &func,long xdata)
{
	//_control87( _EM_INVALID|_EM_DENORMAL|_EM_ZERODIVIDE|_EM_OVERFLOW|_EM_UNDERFLOW|_EM_INEXACT, _MCW_EM ); 
	Lock.Lock();

	// enge parsing shit
	char s[10000];
	short i,c,j,k,l,li,si,lsi;
	char t[10000];
	double v;
	long idx;

	// Clear old tree..
	ClearTree();
	b_IsValid=false;	
	varlist.Resize(64);
	functxt=func;

	for (i=0;i<26;i++)variableindex[i]=-1;

	// insert the 0
	varlist[0].IsConstant=true; // number
	varlist[0].value=0; // ZERO
	varcount=1;

	char checkstring[]=" =:;_,?$#|@()*+-%./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ^abcdefghijklmnopqrstuvwxyz~\0";
	#define operations 40
	char functiontable[]="sin|cos|tan|log|ln|abs|sqr|rnd|atan|asin|acos|int|deg|rad|fac|sum|saw|blk|ran|fsin|fcos|frac|two|fsqr|oosqr|oo|idf|in|it|dn|dt|vk|vl|vm|vg|vh|vi|zero|sign|flog\0";
	char operatortable[]="+-*/^~@%():;_?\0";
	
	s[0]=0;_copychar(s,func,-1);

	
	// Parse tree for weird characters, and convert to lowercase
	// except: ( ) * + - . / 01234567890 ABCDEFGHIJKLMNOPQRSTUVWXYZ ^ abcdefghijklmnopqrstuvxyz~

	for (i=0;s[i]!=0;i++)
	{
		bool found=false;
		for (j=0;checkstring[j]!=0;j++)
			if (s[i]==checkstring[j]){found=true;break;}
		if (found==false)goto error;
		s[i]=toLower(s[i]);
	}

	// Count brackets
	c=0;for (i=0;s[i]!=0;i++){ if (s[i]=='(')c++;if (s[i]==')')c--; if (c<0)goto error;}if (c!=0)goto error;
	// Count absolute marks..
	c=0;for (i=0;s[i]!=0;i++) if (s[i]=='|')c++;if ((c%2)!=0)goto error;
	
	// Replace absolute marks with abs()
	redomarks:;
	c=0;si=-1,li=0,lsi=-1,i=0;
	while (s[i]!=0)
	{
		if (s[i]=='(')li++;if (s[i]==')')li--; 
		if (s[i]=='|')
		{
			if (si==-1)
			{
				si=i;
				lsi=li;
			}
			else if (li==lsi)
			{ // found matching (level wize) |
				_deletechar(s,i);_deletechar(s,si);
				_insertchar(s,si,'(');_insertchar(s,si,'s');_insertchar(s,si,'b');_insertchar(s,si,'a');_insertchar(s,i+3,')');
				goto redomarks;
			}
		}
		i++;
	}
	// Look for leftover absolute marks.. ifso bailout
	for (i=0;s[i]!=0;i++)if (s[i]=='|')goto error;
	
	// Remove spaces
	for (i=0;s[i]!=0;i++)if (s[i]==' '){_deletechar(s,i);i--;}
	if (s[0]==0)goto error;
	
	// look for '='s
	for (i=0;s[i]!=0;i++)if (s[i]=='=')
	{
		if (i==0)goto error;// no beginning '='
		if (s[i+1]==0)goto error; // no ending '='
		
		_deletechar(s,i);_insertchar(s,i,'(');_insertchar(s,i,'-');
		_insertchar(s,i,')');_insertchar(s,0,'(');
		for (;s[i]!=0;i++);
		_insertchar(s,i,')');
		break;
	}

	// look leftover '=''s
	for (i=0;s[i]!=0;i++)if (s[i]=='=')goto error;

	// Replace numbers and variables with variable 'characters', and function strings with function characters
	// So, if a group of characters is NOT in the function list, parse them as variables

	for (i=0;s[i]!=0;i++)
	{
		// First check if you found a number
		if ( (s[i]>=48 && s[i]<=57) || s[i]=='.' || s[i]==',' || s[i]=='$' || s[i]=='#')
		{	//NUMBER!
			//Find the end of the number then..
			if(s[i]=='$') // hex number
			{
				char hextbl[]="0123456789abcdef\0";
				// scan forward until you find anything else then a number or a 'abcdef'
				for (j=i+1;s[j]!=0;j++)
				{
					if ((s[j]<48 || s[j]>57) && (s[j]<97 || s[j]>102) )break;
				}
								// j is now pointing at the end of the number..
				// Copy the number to the temp string for conversion..
				v=0;
				t[0]=0; _copychar(t,&s[i],j-i); 
				// now hand-convert the hex number..
				if (j-i-1>0) // it's not just a single '$'
				{
					double mul=1;
					for (k=j-1;k>i;k--)
					{
						for (l=0;l<16;l++)if (hextbl[l]==s[k])break;
						v+=mul*l;
						mul*=16;
					}
				}
			}
			else if (s[i]=='#') // Binary
			{
				for (j=i+1;s[j]!=0;j++)
				{
					if (s[j]!='0' && s[j]!='1')break;
				}
				// Copy the number to the temp string for conversion..
				v=0;
				t[0]=0; _copychar(t,&s[i],j-i); 
				// now hand-convert the binary number..
				if (j-i-1>0) // it's not just a single '#'
				{
					double mul=1;
					for (k=j-1;k>i;k--)
					{
						if (s[k]=='1')v+=mul;
						mul*=2;
					}
				}
			}
			else // Normal
			{
				bool isbreuk=false;
				for (j=i+1;s[j]!=0;j++)
				{
					if (s[j]==',')isbreuk=true;
					if ((s[j]<48 || s[j]>57) && s[j]!='.' && s[j]!=',')break;
				}
				if (isbreuk==false)
				{
					// j is now pointing at the end of the number..
					// Copy the number to the temp string for conversion..
					t[0]=0; // makes sure copychar appends a \0
					_copychar(t,&s[i],j-i);
					v=atof(t);
				}			
				else
				{
					// count..
					short c=0;
					if (s[i]==',')goto error;
					for (short k=i;k<j;k++)
					{
						if (s[k]==',')c++;
						if (s[k]==',' && s[k+1]=='`')goto error;
						if (c>2)goto error;
						if (s[k]=='.')goto error;
					}
					if (c==1)
					{
						short m;
						double a,b;	
						for (short k=i;k<j;k++)if (s[k]==','){m=k;break;}
						t[0]=0; _copychar(t,&s[i],m-i);   a=atof(t);
						t[0]=0; _copychar(t,&s[m+1],j-m); b=atof(t);
						if (b!=0)v=a/b;
						else goto error;
					}
					else if (c==2)
					{
						short m,n;
						double a,b,c;	
						bool first=false;
						for (short k=i;k<j;k++)
						{
							if (first==false)
							{
								if (s[k]==',')
								{
									first=true;
									m=k;
								}
							}
							else
							{
								if (s[k]==',')
								{
									n=k;
									break;
								}
							}
						}
						t[0]=0; _copychar(t,&s[i],m-i);   a=atof(t);
						t[0]=0; _copychar(t,&s[m+1],n-m); b=atof(t);
						t[0]=0; _copychar(t,&s[n+1],j-n); c=atof(t);
						if (c!=0)v=a+b/c;
						else goto error;
					}
					t[0]=0; _copychar(t,&s[i],j-i); 
				}
			} // end if normal number..

			// Create number in table.. 
			// prerequisites i=start, j=end,t=title.v=value
			// Look for a similar first..
			for (k=0;k<varcount;k++)
			{
				if (varlist[k].IsConstant && varlist[k].value==v)
				{
					_replacechar(s,i,j-i,(char)(128+k));
					goto skip;
				}
			}
			
			// Now allocate variable space for the number
			_replacechar(s,i,j-i,(char)(128+varcount));
			varlist[varcount].IsConstant=true; // number
			varlist[varcount].value=v; 

			varcount++;
		
			goto skip;
		}
		// Now check the normal operators
		for (j=0;operatortable[j]!=0;j++)if (s[i]==operatortable[j])goto skip;
		// Look for the textual operators..
		l=0;
		for (j=0;j<operations;j++)
		{
			short m=l;
			for (;functiontable[m]!='|';m++);
			if (_stringcmp(&functiontable[l], &s[i],m-l)==true)
			{
				// found operator .. now check for hidden variables
				if (j==_OPO_IDIF)
				{
					//log("Encountered Idif");
				}
				if (j==_OPO_DT || j==_OPO_IT) // hidden variable 't'
				{
					// see if it already existed
					idx='t'-97;
					if (variableindex[idx]==-1)
					{
						varlist[varcount].IsConstant=false; // variable
						varlist[varcount].value=0; // Not yet known
						variableindex[idx]=varcount;
						varcount++;
					}					
				}
				_replacechar(s,i,m-l,(char)(196+j));

				goto skip;
			}
			l=m+1;
		}

		// So, it was none of the above.. then it'l be classified as a variable..
		// Since variables are only 1 char long, just replace the char..
		
		// see if it already existed
		idx=s[i]-97;if ( idx<0 || idx >25)goto error;
		if (variableindex[idx]!=-1)
		{
			s[i]=(char)(128+variableindex[idx]);
			goto skip;
		}
		
		varlist[varcount].IsConstant=false; // variable
		varlist[varcount].value=0; // Not yet known
		variableindex[idx]=varcount;
		s[i]=(char)(128+varcount);
		
		varcount++;

		skip:;
		if (varcount>63)goto error;
	}

	// Now insert * things where needed..
	
	for (i=0;s[i]!=0;i++)
	{
		if (s[i+1]==0)break; // nothing to compare with..

		if ((s[i]>=128 && s[i]<196) || s[i]==')') // variable or )
		{
			if (s[i+1]>=128 && s[i]<196) // 2 adjecant variables.. insert *
				_insertchar(s,++i,'*');
			else
			if (s[i+1]=='(')
				_insertchar(s,++i,'*');
			else
			if (s[i+1]>=196)
				_insertchar(s,++i,'*');
		}

	}

	for (i=0;s[i]!=0;i++)
	{
		if (s[i+1]==0)break; // nothing to compare with..

	}

	// check the minusses, and do a math grammar check

	reloopsigns:;
	for (i=0;s[i]!=0;i++)
	{
		if (s[i]>=196)
			if (s[i+1]==0 || s[i+1]!='(')goto error;

			
		if (s[i+1]==0 || s[i+1]==')')
		{
			if (s[i]=='%')goto error;
			if (s[i]=='*')goto error;
			if (s[i]=='/')goto error;
			if (s[i]=='~')goto error;
			if (s[i]=='@')goto error;
			if (s[i]==';')goto error;
			if (s[i]=='_')goto error;
			if (s[i]=='^')goto error;
			if (s[i]=='+')goto error;
			if (s[i]=='-')goto error;
			if (s[i]=='?')goto error;
		}

		if (s[i+1]==0)break; // nothing to compare with..
				
		if (s[i]=='-' && s[i+1]=='-') // replace double -- by +
			{_replacechar(s,i,2,'+');goto reloopsigns;}
		if (s[i]=='-' && s[i+1]=='+') // replace double -+ by -
			{_replacechar(s,i,2,'-');goto reloopsigns;}
		if (s[i]=='+' && s[i+1]=='-') // replace double +- by -
			{_replacechar(s,i,2,'-');goto reloopsigns;}
		if (s[i]=='+' && s[i+1]=='+') // replace double ++ by +
			{_replacechar(s,i,2,'+');goto reloopsigns;}
		if (s[i]=='(' && s[i+1]== ')') // insert a 0
			_insertchar(s,i+1,(char)128);

		if (s[i+1]=='+')
		{
			if (s[i]=='%'){_deletechar(s,i+1);goto reloopsigns;}
			if (s[i]=='*'){_deletechar(s,i+1);goto reloopsigns;}
			if (s[i]=='/'){_deletechar(s,i+1);goto reloopsigns;}
			if (s[i]=='~'){_deletechar(s,i+1);goto reloopsigns;}
			if (s[i]==':'){_deletechar(s,i+1);goto reloopsigns;}
			if (s[i]==';'){_deletechar(s,i+1);goto reloopsigns;}
			if (s[i]=='_'){_deletechar(s,i+1);goto reloopsigns;}
			if (s[i]=='^'){_deletechar(s,i+1);goto reloopsigns;}
			if (s[i]=='@'){_deletechar(s,i+1);goto reloopsigns;}
			if (s[i]=='('){_deletechar(s,i+1);goto reloopsigns;}
			if (s[i]=='?'){_deletechar(s,i+1);goto reloopsigns;}
		}

		if (i==0 && s[i]=='+')
			_deletechar(s,i);

		// cannot have things like */ after eachother
		if (s[i]=='-' || s[i]=='+' || s[i]=='%' ||s[i]=='*' || s[i]=='/' || s[i]=='^' || s[i]=='~' || s[i]=='@'|| s[i]=='(')
		{
			if (s[i+1]=='%')goto error;
			if (s[i+1]=='/')goto error;
			if (s[i+1]=='^')goto error;
			if (s[i+1]=='~')goto error;
			if (s[i+1]=='@')goto error;
			if (s[i+1]==':')goto error;
			if (s[i+1]==';')goto error;
			if (s[i+1]=='?')goto error;
			if (s[i+1]=='_')goto error;
			if (s[i+1]=='*')goto error;
		}
		
		if (i==0) // can't do anything with something that's not there..
		{
			if (s[i]=='%')goto error;
			if (s[i]=='*')goto error;
			if (s[i]=='/')goto error;
			if (s[i]=='~')goto error;
			if (s[i]=='@')goto error;
			if (s[i]=='^')goto error;
			if (s[i]==':')goto error;
			if (s[i]==';')goto error;
			if (s[i]=='?')goto error;
			if (s[i]=='_')goto error;
		}
	}

	// NOW the string is prepared, and processed..
	//Now pass it to the recursive object splitter thing..
	for (i=0;s[i]!=0;i++); // find stringlength
	// create new basenode
	basenode=CreateSubnode();
	basenode->RecursiveSplit(s,0,i);

	// make sure e and p are fixed
	idx=(long)'e'-97;
	if (variableindex[idx]!=-1)varlist[variableindex[idx]].IsConstant=true,
							   varlist[variableindex[idx]].value=E;
	idx=(long)'p'-97;
	if (variableindex[idx]!=-1)varlist[variableindex[idx]].IsConstant=true,
							   varlist[variableindex[idx]].value=PI;

	// now go check our variable table for 'non-constants'
	for (i=0;i<varcount;i++)if (!varlist[i].IsConstant)break;
	if (i==varcount) // ok now we can make just one variable out of it
	{
		// check the tree for additive things
		FuncTreeNode *node=basenode;
		while(node!=NULL)
		{
			if (node->specop>=_OPO_IDIF && node->specop<=_OPO_VI)break;
			node=(FuncTreeNode *)node->walk(basenode);
		}
		if (node==NULL)
		{
			variable=basenode->Calculate();
			IsVar=true;
			ClearTree();
		}
		else IsVar=false;
	}
	else IsVar=false;
	b_IsValid=true;
	Lock.Unlock();
	return true;
error:;
	ClearTree();
	Lock.Unlock();
	return false;
}

//-------------------------------------------------------

bool FuncTree::SetVariable(char variable,double value)
{
	long idx=(long)toLower(variable)-97;
	if (idx>=0 && idx<=25)
	if (variableindex[idx]!=-1)
	{
		varlist[variableindex[idx]].value=value;
		return true;
	}
	return false;
}

//-------------------------------------------------------

bool FuncTree::SetVarMulti(char vr1,double vl1,char vr2,double vl2,char vr3,double vl3,char vr4,double vl4,char vr5,double vl5)
{
	long idx=0;
	idx=(long)toLower(vr1)-97;if(idx>=0 && idx<=25 && variableindex[idx]!=-1)varlist[variableindex[idx]].value=vl1;
	idx=(long)toLower(vr2)-97;if(idx>=0 && idx<=25 && variableindex[idx]!=-1)varlist[variableindex[idx]].value=vl2;
	idx=(long)toLower(vr3)-97;if(idx>=0 && idx<=25 && variableindex[idx]!=-1)varlist[variableindex[idx]].value=vl3;
	idx=(long)toLower(vr4)-97;if(idx>=0 && idx<=25 && variableindex[idx]!=-1)varlist[variableindex[idx]].value=vl4;
	idx=(long)toLower(vr5)-97;if(idx>=0 && idx<=25 && variableindex[idx]!=-1)varlist[variableindex[idx]].value=vl5;
	return(true);
}

//-------------------------------------------------------
//----------------------------------------------------------

double FuncTree::Calculate()
{
	double d=0;
	Lock.Lock();
	if (IsValid())
	{
		if (IsVar)d=variable;
		else d=basenode->Calculate();
	}
	Lock.Unlock();
	return(d);
}

//----------------------------------------------------------

FuncTreeNode::FuncTreeNode(FuncTree *ibase,_ctree *parent):_ctree(parent)
{
	base	 =ibase;
	neg		 =false;
	otype	 =0;
	specop	 =-1;
	IsVariable=false;
	value	 =0;
	lastt	 =0;
	lastval	 =0;
}

//----------------------------------------------------------

FuncTreeNode::~FuncTreeNode()
{
}

//-------------------------------------------------------


void FuncTreeNode::RecursiveSplit(char *c,short s,short e)
{
	short level,i;
	bool deop=false;
	restart:;
	
	// if it has a preceeding '-' set neg flag and restart
	// ONLY IF IT"S THE ONLY 'operator' THING IN the SCOPE.
  	
	level=0;
	if (c[s]=='-')
	{
		bool onlyone=true;
		for (i=s+1;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0 && c[i]!=')')
		{
			if (c[i]=='-' || c[i]=='+'){onlyone=false;break;}//normal operator.. so 
		}}
		if (onlyone==true)	
		{
			// Splice off the minus..
			otype=_OPT_ADD;
			//if (neg==true)neg=false;
			//else 
			neg=true;
			(base->CreateSubnode(this))->RecursiveSplit(c,s+1,e);
			//s++;
			goto pass;
		}
	}
	// take off () shells, ONLY when the level never becomes zero again.

	if (c[s]=='(' && c[e-1]==')')
	{
		level=1;
		bool waszero=false;
		for (i=s+1;i<e-1;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)waszero=true;}
		if (waszero==false)
		{
			s++,e--;
			goto restart;
		}
	}

	// Now look if this object maybe an operator with nothing
	if (c[s]>=196) 
	{
		bool alone=true;
		level=0;
		for (i=s;i<e;i++)
		{
			if (c[i]=='(')level++;if (c[i]==')')level--;
			if (level==0 && c[i]!=')')
				if (c[i]<128){alone=false;break;}
		}
		if (alone==true)
		{
			if (deop==false)
			{
				// set operation, and restart
				deop=true;
				specop=c[s]-196;
				s++;
				goto restart;
			}
			else //else re-split.. cause this is an op-in-op
			{
				otype=_OPT_ADD;
				(base->CreateSubnode(this))->RecursiveSplit(c,s,e);
			}
		}
	}
	if (c[s]>=128 && e==s+1) // is only 1 object left..
	{
		// were done here..
		IsVariable=true;
		varindex=c[s]-128;
		goto pass;
	}

	// Next look for ;'s  (select param b)
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]==';')
		{
			// this means, the thing coming before i must be an object if this is to split here..
			if (i>0 && (c[i-1]>=128 || c[i-1]==')')) // is an object
			{
				//split the thing here, using + or - tree types..
				otype=_OPT_NOP;
				(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
				(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
				goto pass;
				
			}
		}
	}}


	// Next look for +'s 
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]=='+')
		{
			// this means, the thing coming before i must be an object if this is to split here..
			if (i>0 && (c[i-1]>=128 || c[i-1]==')')) // is an object
			{
				//split the thing here, using + or - tree types..
				otype=_OPT_ADD;
				(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
				(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
				goto pass;
				
			}
		}
	}}

	// Next look for -'s .. Backwards!
	level=0;
	for (i=e-1;i>=s;i--){if (c[i]==')')level++;if (c[i]=='(')level--;if (level==0)
	{
		if (c[i]=='-')
		{
			// this means, the thing coming before i must be an object if this is to split here..
			if (i>0 && (c[i-1]>=128 || c[i-1]==')')) // is an object
			{
				//split the thing here, using + or - tree types..
				otype=_OPT_SUB;
				(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
				(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
				goto pass;
			}
		}
	}}
	

	// Then look /'s.
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]=='/')
		{
			// split here..
			otype=_OPT_DIV;
			(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
			(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
			goto pass;
		}
	}}

	// Then look *'s.
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]=='*')
		{
			// split here..
			otype=_OPT_MUL;
			(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
			(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
			goto pass;
		}
	}}

	// Then look @'s. // fix
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]=='@')
		{
			// split here..
			otype=_OPT_FIX;
			(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
			(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
			goto pass;
		}
	}}

	// First look at the modulo operator (%)
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]=='%')
		{
			// split here..
			otype=_OPT_MOD;
			(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
			(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
			goto pass;
		}
	}}


	// Then look :'s. // distance
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]==':')
		{
			// split here..
			otype=_OPT_DST;
			(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
			(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
			goto pass;
		}
	}}

	// Then look ?'s. // angle
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]=='?')
		{
			// split here..
			otype=_OPT_ANG;
			(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
			(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
			goto pass;
		}
	}}

	// Then look _'s. // alogb
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]=='_')
		{
			// split here..
			otype=_OPT_LOG;
			(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
			(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
			goto pass;
		}
	}}

	// Then look ^'s.
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]=='^')
		{ 
			// split here..
			otype=_OPT_POW;
			(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
			(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
			goto pass;
		}
	}}

	// Then look ~'s.
	level=0;
	for (i=s;i<e;i++){if (c[i]=='(')level++;if (c[i]==')')level--;if (level==0)
	{
		if (c[i]=='~')
		{
			// split here..
			otype=_OPT_NSQR;
			(base->CreateSubnode(this))->RecursiveSplit(c,s,i);
			(base->CreateSubnode(this))->RecursiveSplit(c,i+1,e);
			goto pass;
		}
	}}

pass:;
}

//-------------------------------------------------------

#define FLOATMAX 1.7E300 
#define FLOATMIN 1.7E-300

double FuncTreeNode::Calculate()
{
	double oldval=value;
	double t,d;
	long idx;
	if (IsVariable) // just return the variable stored..
	{
		value=base->varlist[varindex].value;
	}
	else
	{
		double a=0,b=0;
		if (child!=NULL)
		{
			b=((FuncTreeNode *)child)->Calculate();		
			if (child->next!=NULL)
			{
				a=((FuncTreeNode *)child->next)->Calculate();		
			}
		}
		// Now combine a and b using the specified operator
		switch(otype)
		{
			case _OPT_NOP: 
				value = b;
				break;
			case _OPT_MOD:
				if (b!=0)
					value=fmod(a,b);
				else value=0; 
				break;
			case _OPT_ADD:
				value=a+b;
				break;
			case _OPT_SUB:
				value=a-b;
				break;
			case _OPT_MUL:
				value=a*b;
				break;
			case _OPT_DIV:
				if (b!=0)
				value=a/b;
				else {
					if (a<0)value=-FLOATMAX; // about infinite ;)
					else if (a>0)value=FLOATMAX;
					else value=0;
				}
				break;
			case _OPT_POW:
				_control87( _EM_INVALID|_EM_DENORMAL|_EM_ZERODIVIDE|_EM_OVERFLOW|_EM_UNDERFLOW|_EM_INEXACT, _MCW_EM ); 
				//if (fabs(a)+fabs(b)>200)value=0;
				//else
				if (b==0)value=1;
				else
				if (a<0 && b<1E19)
				{
					b=floor(b+0.5);
					value=pow(a,b);
				}
				else
				if (a<1E19 && b<1E19 && a>0)
					value=pow(a,b);
				else 
				{
					if (a==1)value=1;				
					else
					if (a==0)value=0;
					else value=FLOATMAX;
				}
				break;
			case _OPT_NSQR:
				if (a>=0 && b>=0)
					value=pow(b,1.0/a);
				else value=0;
				break;
			case _OPT_FIX:
				if (b!=0)
				{	
					b=fabs(b);
					value=ceil((a/b)-0.5)*b;
				}
				else value=a;
				break;
			case _OPT_DST:
				value=sqrt(a*a+b*b);
				break;
			case _OPT_ANG:
				value=atan2(a,b);
				break;
			case _OPT_LOG:
				if (a<=0 || b<=0) // special treatment..
				{
					if (a==0 && b==0)value=0;
					else
					if (a==0)value=FLOATMAX;
					else
					if (b==0)value=-FLOATMAX; // -infinity
					else
					if(a<0 && b>0)
					{
						value=ceil(log(b)/log(-a))-0.5;
						// only even ones can be right..
						if (fmod(value,2)!=0)value=FLOATMAX;
					}
					else 
					if(a<0 && b<0)
					{
						value=ceil(log(b)/log(-a))-0.5;
						// only odd ones can be the right one..
						if (fmod(value,2)==0)value=FLOATMAX;
					}
					else value=FLOATMAX;
				}
				else 
				{
					a=log(a);
					b=log(b);
					if (a!=0)value=b/a;
					else value=FLOATMAX;
				}
				break;
		}
	}
	if (specop!=-1)
	{
		switch(specop)
		{
		case _OPO_SIN:
			value=sin(value);
			break;
		case _OPO_COS:
			value=cos(value);
			break;
		case _OPO_TAN:
			value=tan(value);
			break;
		case _OPO_LOG:
			if (value>0)value=log10(value);
			else value=0;
			break;
		case _OPO_LN:
			if (value>0)value=log(value);
			else value=0;
			break;
		case _OPO_ABS:
			value=fabs(value);
			break;
		case _OPO_SQR:
			if (value>=0)
				value=sqrt(value);
			else value=0;
			break;
		case _OPO_RND: // make random number.. with seed
			{
				float f=(float)value;
				long s= *((long *)(&f));
				s = s * 8192 ^ s;
				value = 1.0 - ((s*(s*s*15731+789221)+1376312589)& 0x7fffffff)/1073741824.0;
			}
			break;
		case _OPO_RAN: // make random number.. without seed, with limits
			{
				mfTab.RandomCounter++;
				if (mfTab.RandomCounter>=MF_RANDOM)mfTab.RandomCounter=0;
				value=mfTab.Random[mfTab.RandomCounter];
			}
			break;
		case _OPO_ATN:
			value=atan(value);
			break;
		case _OPO_ASN:
			if (value>1 || value<-1)value=0;
			else value=asin(value);
			break;
		case _OPO_ACS:
			if (value>1 || value<-1)value=0;
			else value=acos(value);
			break;
		case _OPO_INT:
			value=ceil(value-0.5);
			break;		
		case _OPO_RAD:
			value=value*TRAD;
			break;
		case _OPO_DEG:
			value=value*TDEG;
			break;		
		case _OPO_FAC:
			if (value<0 && value>-FACTABLESIZE)value=-ftbl.factable[-(long)value];
			else if (value>=0 && value<FACTABLESIZE)value=ftbl.factable[(long)value];
			else value=FLOATMAX;
			break;		
		case _OPO_SUM:
			if (value<0 && value>-SUMTABLESIZE) value=-ftbl.sumtable[-(long)value];
			else if (value>=0 && value<SUMTABLESIZE)value=ftbl.sumtable[(long)value];
			else if (value<0) value=-FLOATMAX;
			else value=FLOATMAX;
			break;		
		case _OPO_SAW: // saw wave
			if (value<0){d=-1;value=-value;}else d=1;
			t=fmod(value,PI2)/(PI12);
			if (t<1)
				value=t;
			else if (t<2)
				value=2-t;
			else if (t<3)
				value=2-t;
			else if (t<4)
				value=t-4;
			value*=d;
			break;		
		case _OPO_BLK: // block wave
			if (value<0){d=-1;value=-value;}else d=1;
			t=fmod(value,PI2);
			// Determine which quadrant..
			if ((t/PI)<1)value=1;
			else value=-1;
			value*=d;
			break;		
		case  _OPO_FSN:
			value=mfSin(value);
			break;
		case _OPO_FCS:
 			value=mfCos(value);
			break;
		case _OPO_FRAC:
			value=(value)-(floor(value));
			break;
		case _OPO_TWO:
			value=(value+1.0)/2.0;
			break;
		case _OPO_FSQR:
			value=mfFSqrt(value);
			break;
		case _OPO_OOSQR:
			value =mfOneOverSqrt(value);
			break;
		case _OPO_OO:
			value = mfOneOver(value);
			break;
		case _OPO_FLOG:
			value= log(1.2)/log((fabs(value)+1));
			break;
		case _OPO_IN: // normal additive thingy
			value=oldval+value;
			break;
		case _OPO_IT: // time dependant
			// get 't'
			idx='t'-97;t=0;if (base->variableindex[idx]!=-1)t=base->varlist[base->variableindex[idx]].value;
			value = oldval+value*fabs(t-lastt);
			break;
		case _OPO_DN:
			oldval = value-lastval;
			lastval = value;
			value   = oldval;
			break;
		case _OPO_DT:
			idx='t'-97;t=0;if (base->variableindex[idx]!=-1)t=base->varlist[base->variableindex[idx]].value;
			oldval  = (value-lastval)*fabs(t-lastt);
			lastval = value;
			value	= oldval;
			break;
		case _OPO_IZ:
			if (value!=0)value=1;
			break;
		case _OPO_SIGN:
			if (value>=0)value=1;
			else value=0;
			break;
		case _OPO_VK:idx='k'-97;if (base->variableindex[idx]!=-1)base->varlist[base->variableindex[idx]].value=value;break;
		case _OPO_VL:idx='l'-97;if (base->variableindex[idx]!=-1)base->varlist[base->variableindex[idx]].value=value;break;
		case _OPO_VM:idx='l'-97;if (base->variableindex[idx]!=-1)base->varlist[base->variableindex[idx]].value=value;break;
		case _OPO_VG:idx='g'-97;if (base->variableindex[idx]!=-1)base->varlist[base->variableindex[idx]].value=value;break;
		case _OPO_VH:idx='h'-97;if (base->variableindex[idx]!=-1)base->varlist[base->variableindex[idx]].value=value;break;
		case _OPO_VI:idx='i'-97;if (base->variableindex[idx]!=-1)base->varlist[base->variableindex[idx]].value=value;break;
		case _OPO_IDIF:
		if (lastval!=value)
		{
			lastval=value;
			value=oldval+value;
		}
		else value=oldval;
		break;
		}
	}
	if (neg==true)value*=-1;
	return(value);
}



//-------------------------------------------------------

