/*
---------------------------------------------------------------------------
|						Function tree objects
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
Requirements:

Changes:
	11-19-98	Initial file setup
*/

#ifndef _H_FUNCTIONTREE
#define _H_FUNCTIONTREE

//----------------------------------------------------------

// 2 object operators
enum
{
	 _OPT_ADD  =0,
	 _OPT_SUB  =1,
	 _OPT_MUL  =2,
	 _OPT_DIV  =3,
	 _OPT_POW  =4,
	 _OPT_NSQR =5,
	 _OPT_MOD  =6,
	 _OPT_FIX  =7,
	 _OPT_LOG  =8,
	 _OPT_DST  =9,  // may utilise fast sqrt
	 _OPT_ANG  =10,
	 _OPT_FDST =11,
	 _OPT_NOP  =12	// Nop separator
};

// 1 object operators
enum
{
	 _OPO_SIN  =0,
	 _OPO_COS  =1,
	 _OPO_TAN  =2,
	 _OPO_LOG  =3,
	 _OPO_LN   =4,
	 _OPO_ABS  =5,
	 _OPO_SQR  =6,
	 _OPO_RND  =7,	// random with table, invariable seed
	 _OPO_ATN  =8,
	 _OPO_ASN  =9,
	 _OPO_ACS  =10,
	 _OPO_INT  =11,
	 _OPO_DEG  =12,
	 _OPO_RAD  =13,
	 _OPO_FAC  =14,
	 _OPO_SUM  =15,
	 _OPO_SAW  =16,
	 _OPO_BLK  =17,
	 _OPO_RAN  =18,	// random with table, variable seed
	 _OPO_FSN  =19,
	 _OPO_FCS  =20,
	 _OPO_FRAC =21,
	 _OPO_TWO  =22, // +1/2
	 _OPO_FSQR =23,
	 _OPO_OOSQR=24, // fast on over sqr
	 _OPO_OO   =25,  // fast one over
	 _OPO_IDIF =26, // ( | | +1 )_1.2 fast log-transform
	 _OPO_IN   =27, // integrate element.. 
	 _OPO_IT   =28, // int * delta_t
	 _OPO_DN   =29,	// differentiate
	 _OPO_DT   =30,	// diff*delta_t
	 _OPO_VK   =31, // variable thingies
	 _OPO_VL   =32,
	 _OPO_VM   =33,
	 _OPO_VG   =34,
	 _OPO_VH   =35,
	 _OPO_VI   =36,
	 _OPO_IZ   =37,	// Is Zero.. return 0 if zero else 1
	 _OPO_SIGN =38,	// sign
	 _OPO_FLOG =39  // different integrating (dialing)
};

#define FACTABLESIZE 1024
#define SUMTABLESIZE 16384

struct ftable
{
	ftable();	
	double factable[FACTABLESIZE];
	double sumtable[SUMTABLESIZE];

	// constants
	double td_flog;
	double td_random;
	double td_arctableindex;
	double td_tableindex;
	double td_onehalf;
	double td_pi2;
	double td_pi12;
	double td_pi116;
	double td_trad;
	double td_tdeg;
	double td_two;
	double td_three;
	double td_four;
};
extern ftable ftbl;


//----------------------------------------------------------
class FuncTreeNode; // forward reference

// variable list
class VarElement
{
	public: 
	VarElement(){value=0,IsConstant=true;}
	~VarElement(){};

	bool   IsConstant;
	double value;
};

DECLARE_CLASS(FuncTree);
class  FuncTree // holder class
{
	public:
	FuncTree();
	FuncTree(String &eq);
	virtual ~FuncTree();

	rFuncTree operator=(rFuncTree   other)								{ other.Copy(this); return *this; }
	rFuncTree operator=(rString      other)								{ Create(other);return *this; }
	inline friend rArchive operator<<(rArchive arc, cFuncTree v)		{ arc <<v.b_IsValid<<v.IsVar<<v.variable<<v.functxt;return arc;}
	inline friend rArchive operator>>(rArchive arc, rFuncTree v)		{ arc >>v.b_IsValid>>v.IsVar>>v.variable>>v.functxt;if (!v.IsVar)v.Create(v.functxt);return arc;}

	String &GetString(){return functxt;}

	bool	b_IsValid;  // valid equation?

	bool	IsVar;		// wether it is just one variable
	double	variable;	// which variable it is

	String functxt; // the textual representation of the function

	virtual void Copy(FuncTree *target,long cookie=-1); // copys equation without reparsing the string (tree copy)

	virtual bool CreateV(double v); // create function with only one value
	virtual bool Create(String &func,long cookie=-1);
	
	bool IsValid(){return b_IsValid;}
	
	virtual void ClearTree();

	// subnode overloading functions:
	virtual FuncTreeNode *CreateSubnode(FuncTreeNode *parent=NULL);

	// variable shite
	long	varcount;
	Array<VarElement> varlist;
	long variableindex[26]; // variable index for a..z

	// Set variable, matching the name to value
	virtual bool  SetVariable(char vr1,double vl1);
	virtual bool  SetVarMulti(char vr1,double vl1,char vr2=0,double vl2=0,char vr3=0,double vl3=0,char vr4=0,double vl4=0,char vr5=0,double vl5=0);

	// calculating the things
	virtual double Calculate(char var1,double val1)							{SetVariable(var1,val1);return(Calculate());}
	virtual double Calculate(char var1,double val1,char var2,double val2)	{SetVariable(var1,val1);SetVariable(var2,val2);return(Calculate());}
	virtual double Calculate();

	FuncTreeNode *basenode;

	// update cookie
	long Cookie;

	// multithread lock
	MultithreadLock Lock;
};

//----------------------------------------------------------

class  FuncTreeNode : public _ctree
{
	public:

	FuncTreeNode(FuncTree *base,_ctree *parent=NULL);
	virtual ~FuncTreeNode();
	
	virtual FuncTreeNode *Copy(FuncTree *root,FuncTreeNode *parent);
	virtual void RecursiveSplit(char *c,short s,short e); // recursively subdivides nodes

	FuncTree *base;   // for finding variables

	// per node info
	bool   neg;		  // if it's a negative object or not..
	short  otype;	  // the way the 2 children are done together
	short  specop;    // the 'extra' 1 target op needed to be done on them
	bool   IsVariable;// combine node OR variable
	short  varindex;  // wich variable ID
	double value;	  // nodes value
	double lastt;	  // for 't' differencing
	double lastval;	  // for differencing
	// processes the tree
	virtual double Calculate();
};


#endif