/*
---------------------------------------------------------------------------
|						Function tree objects
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/

#include "Library.h"
#include "GridResolve.h"

//--------------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------------
//- Drawing/solving equations
//--------------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------------
 
void SolveFunctree::RefineSolve2D(long f,double x1,double y1,double x2,double y2,long depth)
{
	//flags is sign bitmap.. bit 0 tl tr bl br
	if (depth>=maxdepth)
	{
		//xp[points]=(x1+x2)/2.0,yp[points]=(y1+y2)/2.0;
		xp[points]=x1,yp[points]=y1;
		if (points<maxpoints)points++;
	}
	else
	{
		double xc,yc;
	
		if ((f&0xf000)!=0) // initial call..
		{
			f=0;
			f+=mfSignB(func->Calculate('x',x1,'y',y1));
			f+=mfSignB(func->Calculate('x',x2,'y',y1))<<1;
			f+=mfSignB(func->Calculate('x',x1,'y',y2))<<2;
			f+=mfSignB(func->Calculate('x',x2,'y',y2))<<3;
		}

		long flagq1=f&(1),flagq2=f&(1<<1),flagq3=f&(1<<2),flagq4=f&(1<<3),t;

		xc=(x1+x2)/2.0;yc=(y1+y2)/2.0;
		
		// mid left
		t=mfSignB(func->Calculate('x',x1,'y',yc));
		flagq1+=(t<<2);flagq3+=t;
		// mid right
		t=mfSignB(func->Calculate('x',x2));
		flagq2+=(t<<3);flagq4+=(t<<1);
		// centre
		t=mfSignB(func->Calculate('x',xc));
		flagq1+=(t<<3);flagq2+=(t<<2);flagq3+=(t<<1);flagq4+=t;
		// mid top
		t=mfSignB(func->Calculate('y',y1));
		flagq1+=(t<<1);flagq2+=t;
		// mid bottom
		t=mfSignB(func->Calculate('y',y2));
		flagq3+=(t<<3);flagq4+=(t<<2);
		if (depth<startlevel)
		{
			RefineSolve2D(flagq1,x1,y1,xc,yc,depth+1);
			RefineSolve2D(flagq2,xc,y1,x2,yc,depth+1);
			RefineSolve2D(flagq3,x1,yc,xc,y2,depth+1);
			RefineSolve2D(flagq4,xc,yc,x2,y2,depth+1);
		}
		else
		{
			if (flagq1!=15 && flagq1!=0)RefineSolve2D(flagq1,x1,y1,xc,yc,depth+1);
			if (flagq2!=15 && flagq2!=0)RefineSolve2D(flagq2,xc,y1,x2,yc,depth+1);
			if (flagq3!=15 && flagq3!=0)RefineSolve2D(flagq3,x1,yc,xc,y2,depth+1);
			if (flagq4!=15 && flagq4!=0)RefineSolve2D(flagq4,xc,yc,x2,y2,depth+1);
		}
		// really odd...
	}
}

//--------------------------------------------------------------------------------------------------------

void SolveFunctree::RefineDraw2D(long f,double x1,double y1,double x2,double y2,long depth)
{
	//flags is sign bitmap.. bit 0 tl tr bl br
	if (depth>=maxdepth)
	{
		((CDC *)dc)->SetPixelV((long)(ftow_xscale*x1+ftow_xoffset),
					  (long)(ftow_yscale*y1+ftow_yoffset),col);
	}
	else
	{
		double xc,yc;
	
		if ((f&0xf000)!=0) // initial call..
		{
			f=0;
			f+=mfSignB(func->Calculate('x',x1,'y',y1));
			f+=mfSignB(func->Calculate('x',x2,'y',y1))<<1;
			f+=mfSignB(func->Calculate('x',x1,'y',y2))<<2;
			f+=mfSignB(func->Calculate('x',x2,'y',y2))<<3;
		}


		long flagq1=f&(1),flagq2=f&(1<<1),flagq3=f&(1<<2),flagq4=f&(1<<3),t;

		xc=(x1+x2)/2.0;yc=(y1+y2)/2.0;
		
		// mid left
		t=mfSignB(func->Calculate('x',x1,'y',yc));
		flagq1+=(t<<2);flagq3+=t;
		// mid right
		t=mfSignB(func->Calculate('x',x2));
		flagq2+=(t<<3);flagq4+=(t<<1);
		// centre
		t=mfSignB(func->Calculate('x',xc));
		flagq1+=(t<<3);flagq2+=(t<<2);flagq3+=(t<<1);flagq4+=t;
		// mid top
		t=mfSignB(func->Calculate('y',y1));
		flagq1+=(t<<1);flagq2+=t;
		// mid bottom
		t=mfSignB(func->Calculate('y',y2));
		flagq3+=(t<<3);flagq4+=(t<<2);
		if (depth<startlevel)
		{
			RefineDraw2D(flagq1,x1,y1,xc,yc,depth+1);
			RefineDraw2D(flagq2,xc,y1,x2,yc,depth+1);
			RefineDraw2D(flagq3,x1,yc,xc,y2,depth+1);
			RefineDraw2D(flagq4,xc,yc,x2,y2,depth+1);
		}
		else
		{
			if (flagq1!=15 && flagq1!=0)RefineDraw2D(flagq1,x1,y1,xc,yc,depth+1);
			if (flagq2!=15 && flagq2!=0)RefineDraw2D(flagq2,xc,y1,x2,yc,depth+1);
			if (flagq3!=15 && flagq3!=0)RefineDraw2D(flagq3,x1,yc,xc,y2,depth+1);
			if (flagq4!=15 && flagq4!=0)RefineDraw2D(flagq4,xc,yc,x2,y2,depth+1);
		}
		// really odd...
	}
}

//--------------------------------------------------------------------------------------------------------
/*
void SolveFunctree::RefineSolve3D(long f,double x1,double y1,double z1,double x2,double y2,double z2,long depth)
{
	//flags is sign bitmap.. bit 0 tl tr bl br
	if (depth>=maxdepth)
	{
		//vec[points].n.x=(float)(x1+x2)*0.5f,vec[points].n.y=(y1+y2)*0.5f,vec[points].n.z=(z1+z2)*0.5f,vec[points].n.w=1;
		vec[points].x=(float)x1,vec[points].y=(float)y1,vec[points].z=(float)z1;
		if (points<maxpoints)points++;
	}
	else
	{
		double xc,yc,zc;
	
		if ((f&0xf000)!=0) // initial call..
		{
			f=0;
			// side 1
			func->SetVarByAddr(xaddr,x1);func->SetVarByAddr(yaddr,y1);func->SetVarByAddr(zaddr,z1);
			f+=mfSignB(func->Calculate());
			func->SetVarByAddr(xaddr,x2);func->SetVarByAddr(yaddr,y1);func->SetVarByAddr(zaddr,z1);	
			f+=mfSignB(func->Calculate())<<1;
			func->SetVarByAddr(xaddr,x2);func->SetVarByAddr(yaddr,y2);func->SetVarByAddr(zaddr,z1);
			f+=mfSignB(func->Calculate())<<2;
			func->SetVarByAddr(xaddr,x1);func->SetVarByAddr(yaddr,y2);func->SetVarByAddr(zaddr,z1);
			f+=mfSignB(func->Calculate())<<3;
			// side 2
			func->SetVarByAddr(xaddr,x1);func->SetVarByAddr(yaddr,y1);func->SetVarByAddr(zaddr,z2);
			f+=mfSignB(func->Calculate())<<4;
			func->SetVarByAddr(xaddr,x2);func->SetVarByAddr(yaddr,y1);func->SetVarByAddr(zaddr,z2);	
			f+=mfSignB(func->Calculate())<<5;
			func->SetVarByAddr(xaddr,x2);func->SetVarByAddr(yaddr,y2);func->SetVarByAddr(zaddr,z2);
			f+=mfSignB(func->Calculate())<<6;
			func->SetVarByAddr(xaddr,x1);func->SetVarByAddr(yaddr,y2);func->SetVarByAddr(zaddr,z2);
			f+=mfSignB(func->Calculate())<<7;

		}

		long t,flagq1=f&(1<<0),flagq2=f&(1<<1),flagq3=f&(1<<2),flagq4=f&(1<<3),
		       flagq5=f&(1<<4),flagq6=f&(1<<5),flagq7=f&(1<<6),flagq8=f&(1<<7);

		xc=(x1+x2)/2.0;yc=(y1+y2)/2.0;zc=(z1+z2)/2.0;
		
		// mid rear
		func->SetVarByAddr(xaddr,xc);func->SetVarByAddr(yaddr,yc);func->SetVarByAddr(zaddr,z1);
		t=mfSignB(func->Calculate());
		flagq1+=(t<<2);flagq2+=(t<<3);flagq3+=t;flagq4+=(t<<1);

		// mid front
		func->SetVarByAddr(zaddr,z2);t=mfSignB(func->Calculate());
		flagq5+=(t<<6);flagq6+=(t<<7);flagq7+=(t<<4);flagq8+=(t<<5);

		// centre
		func->SetVarByAddr(zaddr,zc);t=mfSignB(func->Calculate());
		flagq1+=(t<<6);flagq2+=(t<<7);flagq3+=(t<<4);flagq4+=(t<<5);
		flagq5+=(t<<2);flagq6+=(t<<3);flagq7+=(t<<0);flagq8+=(t<<1);

		//left 
		func->SetVarByAddr(xaddr,x1);t=mfSignB(func->Calculate());
		flagq1+=(t<<7);flagq4+=(t<<4);flagq5+=(t<<3);flagq8+=(t<<0);
	
		//Right
		func->SetVarByAddr(xaddr,x2);t=mfSignB(func->Calculate());	
		flagq2+=(t<<6);flagq3+=(t<<5);flagq6+=(t<<2);flagq7+=(t<<1);
		
		// Top
		func->SetVarByAddr(xaddr,xc);func->SetVarByAddr(yaddr,y1);
		t=mfSignB(func->Calculate());	
		flagq1+=(t<<5);flagq2+=(t<<4);flagq5+=(t<<1);flagq6+=(t<<0);
	
		// Bottom
		func->SetVarByAddr(xaddr,xc);func->SetVarByAddr(yaddr,y2);
		t=mfSignB(func->Calculate());	
		flagq3+=(t<<7);flagq4+=(t<<6);flagq7+=(t<<3);flagq8+=(t<<2);

		// bottom rear
		func->SetVarByAddr(zaddr,z1);t=mfSignB(func->Calculate());	
		flagq3+=(t<<3);flagq4+=(t<<2);

		// bottom front
		func->SetVarByAddr(zaddr,z2);t=mfSignB(func->Calculate());	
		flagq8+=(t<<6);flagq8+=(t<<7);

		// bottom left
		func->SetVarByAddr(zaddr,zc);func->SetVarByAddr(xaddr,x1);t=mfSignB(func->Calculate());	
		flagq8+=(t<<3);flagq4+=(t<<7);

		// bottom right
		func->SetVarByAddr(xaddr,x2);t=mfSignB(func->Calculate());	
		flagq3+=(t<<6);flagq7+=(t<<2);

		// top rear
		func->SetVarByAddr(xaddr,xc);func->SetVarByAddr(yaddr,y1);func->SetVarByAddr(zaddr,z1);t=mfSignB(func->Calculate());	
		flagq1+=(t<<1);flagq2+=(t<<0);

		// top front
		func->SetVarByAddr(zaddr,z2);t=mfSignB(func->Calculate());	
		flagq5+=(t<<5);flagq6+=(t<<4);

		// top left
		func->SetVarByAddr(zaddr,zc);func->SetVarByAddr(xaddr,x1);t=mfSignB(func->Calculate());	
		flagq1+=(t<<4);flagq5+=(t<<0);

		// top right
		func->SetVarByAddr(xaddr,x2);t=mfSignB(func->Calculate());	
		flagq2+=(t<<5);flagq6+=(t<<1);

		// middle left rear.
		func->SetVarByAddr(xaddr,x1);func->SetVarByAddr(yaddr,yc);func->SetVarByAddr(zaddr,z1);t=mfSignB(func->Calculate());	
		flagq1+=(t<<3);flagq4+=(t<<0);

		// middle right rear.
		func->SetVarByAddr(xaddr,x2);t=mfSignB(func->Calculate());	
		flagq2+=(t<<2);flagq3+=(t<<1);

		// middle left front.
		func->SetVarByAddr(xaddr,x1);func->SetVarByAddr(zaddr,z2);t=mfSignB(func->Calculate());	
		flagq5+=(t<<7);flagq8+=(t<<4);

		// middle right front.
		func->SetVarByAddr(xaddr,x2);t=mfSignB(func->Calculate());	
		flagq6+=(t<<6);flagq7+=(t<<5);


		if (depth<startlevel)
		{
			RefineSolve3D(flagq1,x1,y1,z1,xc,yc,zc,depth+1);
			RefineSolve3D(flagq2,xc,y1,z1,x2,yc,zc,depth+1);
			RefineSolve3D(flagq3,xc,yc,z1,x2,y2,zc,depth+1);
			RefineSolve3D(flagq4,x1,yc,z1,xc,y2,zc,depth+1);
			RefineSolve3D(flagq5,x1,y1,zc,xc,yc,z2,depth+1);
			RefineSolve3D(flagq6,xc,y1,zc,x2,yc,z2,depth+1);
			RefineSolve3D(flagq7,xc,yc,zc,x2,y2,z2,depth+1);
			RefineSolve3D(flagq8,x1,yc,zc,xc,y2,z2,depth+1);
		}
		else
		{
			if (flagq1!=255 && flagq1!=0)RefineSolve3D(flagq1,x1,y1,z1,xc,yc,zc,depth+1);
			if (flagq2!=255 && flagq2!=0)RefineSolve3D(flagq2,xc,y1,z1,x2,yc,zc,depth+1);
			if (flagq3!=255 && flagq3!=0)RefineSolve3D(flagq3,xc,yc,z1,x2,y2,zc,depth+1);
			if (flagq4!=255 && flagq4!=0)RefineSolve3D(flagq4,x1,yc,z1,xc,y2,zc,depth+1);
			if (flagq5!=255 && flagq5!=0)RefineSolve3D(flagq5,x1,y1,zc,xc,yc,z2,depth+1);
			if (flagq6!=255 && flagq6!=0)RefineSolve3D(flagq6,xc,y1,zc,x2,yc,z2,depth+1);
			if (flagq7!=255 && flagq7!=0)RefineSolve3D(flagq7,xc,yc,zc,x2,y2,z2,depth+1);
			if (flagq8!=255 && flagq8!=0)RefineSolve3D(flagq8,x1,yc,zc,xc,y2,z2,depth+1);
		}
		// really odd...
	}
}
*/
//--------------------------------------------------------------------------------------------------------
