/*
---------------------------------------------------------------------------
|						Image encapsulation class
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
					All Rights Reserved
---------------------------------------------------------------------------
*/

#include "Library.h"



enum
{
	IMF_POWEROF2	  =1<<8,
	IMF_256X256MAX	  =1<<9,	// max tex. reso
	IMF_1KX1KMAX	  =1<<10	// other max
};

// ---------------------------------------------------------------------------

Image::Image()
{
	data=NULL;
	cookie=-1;ValidObject=false;datasize=0;pixsize=0;pixblock=0;frame=0;
}

// ---------------------------------------------------------------------------

Image::Image(ImageDesc &iid)
{
	cookie=-1;ValidObject=false;datasize=0;pixsize=0;pixblock=0;frame=0;
	SetSize(iid);
}

// ---------------------------------------------------------------------------

Image::~Image()
{
	//ReleaseBuffer();
}

// ---------------------------------------------------------------------------

void *Image::LockData()	// gives access to data block.. multithread safe
{
	if(Lock.Lock())
	{
		if (ValidObject)
		{
			return data;
		}
		else {Lock.Unlock();return(NULL);}
	}else assert(false);
	return(NULL);
}
			

// ---------------------------------------------------------------------------

void Image::UnlockData()	// Unlocks data
{
	if (Lock.IsLocked())
	{
		Lock.Unlock();
	}
}

// ---------------------------------------------------------------------------

long Image::GetType(){return(0);}

void Image::ReleaseBuffer(bool lock)
{
	if(!lock || Lock.Lock())
	{
		if (IsValid())// delete old suface..
		{
			if (IsNonzero(imd.flags&IMF_CLIENTOWN))
			{
				ValidObject=false;
				data=NULL;
			}
			else
			{
				ValidObject=false;
				delete (byte *)data;
				data=NULL;
			}
		}
		if (lock)Lock.Unlock();
	}
}

// ---------------------------------------------------------------------------

bool Image::Copy(Image &srci,bool lock,bool equalformat,bool equalsize)
{

	if (srci.IsValid())
	{
		if (!CompareTypes(srci))
		{
			ImageDesc t=srci.imd;
			t.flags			=	imd.flags; // keep flags
			t.clientdata	=	imd.clientdata; // also keep potiential client owned surface
			t.clientsize	=	imd.clientsize; // do it
			if (!equalformat) // image resize needed
				t.pixelformat=imd.pixelformat;
			if (!equalsize)
				t.xs=imd.xs,t.ys=imd.ys;
			SetSize(t);	
		}
		if (srci.imd.pixelformat != imd.pixelformat ) // do slow pixelformat conversion
		{
			if (srci.LockData()!=NULL)
			{
				if (LockData()!=NULL)
				{
					float xfac=(float)srci.imd.xs/imd.xs,
						  yfac=(float)srci.imd.ys/imd.ys;
					long  x,y,sx,sy;
					for(x=0;x<imd.xs;x++)for (y=0;y<imd.ys;y++)
					{
						sx=(long)((float)x)*xfac,sy=(long)((float)y)*yfac;
						SetIPixel(x,y,srci.GetIPixel(sx,sy) );
					}
				UnlockData();
				}
				srci.UnlockData();
			}
		}
		else
		if (srci.imd.xs != imd.xs || srci.imd.ys!= imd.ys) // different sizes
		{	// resize copy
			// very slow float  copy
			void *src,*dst;
			src=srci.LockData();
			if (src!=NULL)
			{
				dst=LockData();
				if (dst!=NULL)
				{
					float xfac=(float)srci.imd.xs/imd.xs,
						  yfac=(float)srci.imd.ys/imd.ys;
					long  x,y,sx,sy;
					switch(pixblock)
					{
					case 1:
						for(x=0;x<imd.xs;x++)for (y=0;y<imd.ys;y++)
						{
							sx=(long)((float)x)*xfac,sy=(long)((float)y)*yfac;
							((byte *)dst)[y*imd.xs+x]=((byte *)dst)[clamp_l(sy*srci.imd.xs+sx,0,srci.pixsize)];
						}
					break;
					case 2:	
						for(x=0;x<imd.xs;x++)for (y=0;y<imd.ys;y++)
						{
							sx=(long)((float)x)*xfac,sy=(long)((float)y)*yfac;
							((word *)dst)[y*imd.xs+x]=((word *)dst)[clamp_l(sy*srci.imd.xs+sx,0,srci.pixsize)];
						}
					break;
					case 4:	
						for(x=0;x<imd.xs;x++)for (y=0;y<imd.ys;y++)
						{
							sx=(long)((float)x)*xfac,sy=(long)((float)y)*yfac;
							((word *)dst)[y*imd.xs+x]=((word *)dst)[clamp_l(sy*srci.imd.xs+sx,0,srci.pixsize)];
						}	
					break;
					}
					UnlockData();
				}
				srci.UnlockData();
				return(true);
			}
		}
		else // straight memcopy.. this should be used mostly
		{
			byte *src,*dst;
			src=(byte *)srci.LockData();
			if (src!=NULL)
			{
				dst=(byte *)LockData();
				if (dst!=NULL)
				{
					memcpy(dst,src,datasize);
					UnlockData();
				}
				srci.UnlockData();
			}
			return(true);
		}
	}
	return(false);
}

// ---------------------------------------------------------------------------
// clear image to color
void Image::Clear(RGBColor &c)
{
	long i;
	if (IsValid())
	{
		void *data=LockData();
		if (data!=NULL)
		{
			dword col=c.ToType(imd.pixelformat);
			switch(PF_Size(imd.pixelformat))
			{
				case 1:
					memset(data,(byte)col,datasize);
					break;
				case 2:
					for (i=pixsize-1;i>=0;i--)((word *)data)[i]=(word)col;
					break;
				case 3:
					break;
				case 4:
					for (i=pixsize-1;i>=0;i--)((dword *)data)[i]=col;
			}
		}
		
		UnlockData();
	}
}

// ---------------------------------------------------------------------------

bool Image::SetSize(ImageDesc &iimd)
{
	if(Lock.Lock())
	{
		if (IsNonzero(iimd.flags&IMF_POWEROF2))
		{
			iimd.xs=snappow2(iimd.xs,4,1024);
			iimd.ys=snappow2(iimd.ys,4,1024);
		};
		long imsize=iimd.xs*iimd.ys*PF_Size(iimd.pixelformat);
		if (imsize>0)
		{
			if (ValidObject && datasize!=imsize || data==NULL || !ValidObject)
			{
				ReleaseBuffer(false);
				
				if (IsNonzero(iimd.flags&IMF_CLIENTOWN)) // its a client owned surface
				{
					if (imsize!=iimd.clientsize)
					{
//						log("Image -- SetSize clientowned sizes don't match");
						Lock.Unlock();
						return false;
					};
					data=(void *)iimd.clientdata;
				}
				else data=new char[imsize];
				if (data==NULL)
				{
					Lock.Unlock();
					return false;
				}
			}
			imd=iimd;
			pixblock=PF_Size(imd.pixelformat);
			datasize=imsize;
			pixsize =imsize/pixblock;
			ValidObject=true;
			Lock.Unlock();
			return(true);
		}
		Lock.Unlock();
	}else assert(1==0);
	return(false);
}

// ---------------------------------------------------------------------------

void Image::SetPalette(ColorTable &p)
{
	palette=p;
}

// ---------------------------------------------------------------------------

void Image::SetSizeLike(Image &src)
{
	if (src.IsValid())
		SetSize(src.imd);
}

// ---------------------------------------------------------------------------

bool Image::CompareTypes(Image &src)
{
	if (src.IsValid() && IsValid() && src.imd==imd)return(true);
	return(false);
}

// ---------------------------------------------------------------------------

void Image::Save(rArchive arc)
{
	arc<<ValidObject;
	if (ValidObject)
	{
		arc << imd;
		if (!IsNonzero(imd.flags&IMF_NOSAVEBIN))
		{
			void *dta=LockData();
			if (dta!=NULL)arc.Write(dta,datasize);
			UnlockData();
		};
	}
}

// ---------------------------------------------------------------------------


void Image::Load(rArchive arc)
{
	bool valid;
	arc>>valid;
	if (valid)
	{
		ImageDesc iimd;
		arc>>iimd;
		if (!IsNonzero(imd.flags&IMF_NOSAVEBIN))
		{
			SetSize(iimd);
			void *dta=LockData();
			if (dta!=NULL)arc.Read(dta,datasize);
			UnlockData();
		}
	}
}


// ---------------------------------------------------------------------------

void Image::SetPixel(long x,long y,rRGBColor col)
{
	if (ValidObject && Lock.IsLocked())
	{
		if (x>=0 && y>=0 && x<imd.xs && y<imd.ys)
		{
			long offset=y*imd.xs+x;
			dword var=col.ToType(imd.pixelformat);
			// scan for the closest color
			switch(pixblock)
			{
				case 1:
				{
					// scan for closest color (palettized version)
					if (imd.pixelformat==PF_INDEX_8)
					{
						float sqdistance=10000000000;
						long  colindex=0;
						for (long i=0;i<256;i++)
						{
							float r=palette.col[i].r-col.r,
								  g=palette.col[i].g-col.g,
								  b=palette.col[i].b-col.b;
							float dst=r*r+g*g+b*b;
							if (dst==0){colindex=0;break;}
							if (dst<sqdistance)colindex=i,sqdistance=dst;
						};
						((byte *)data)[offset]=(byte)colindex;
					}
					else // alpha
					{
						((byte *)data)[offset]=(byte)var;
					}
				}
				break;

				case 2:
					((word *)data)[offset]=(word)var;
					break;
				case 3:
					break;
				case 4:
					((dword *)data)[offset]=var;
			}
		}
	}
}

// ---------------------------------------------------------------------------
 void Image::SetIPixel(long x,long y,rRGBIColor col)
{
	if (ValidObject && Lock.IsLocked())
	{
		if (x>=0 && y>=0 && x<imd.xs && y<imd.ys)
		{
			long offset=y*imd.xs+x;
			dword var=col.ToType(imd.pixelformat);
			// scan for the closest color
			switch(pixblock)
			{
				case 1:
				{
					// scan for closest color (palettized version)
					if (imd.pixelformat==PF_INDEX_8)
					{
						float sqdistance=10000000000;
						long  colindex=0;
						for (long i=0;i<256;i++)
						{
							float r=palette.col[i].r-col.r(),
								  g=palette.col[i].g-col.g(),
								  b=palette.col[i].b-col.b();
							float dst=r*r+g*g+b*b;
							if (dst==0){colindex=0;break;}
							if (dst<sqdistance)colindex=i,sqdistance=dst;
						};
						((byte *)data)[offset]=(byte)colindex;
					}
					else // alpha
					{
						((byte *)data)[offset]=(byte)var;
					}
				}
				break;

				case 2:
					((word *)data)[offset]=(word)var;
					break;
				case 3:
					break;
				case 4:
					((dword *)data)[offset]=var;
			}
		}
	}
}
// ---------------------------------------------------------------------------

RGBIColor Image::GetIPixel(long x,long y)
{
	RGBIColor d;
	if (ValidObject && Lock.IsLocked())
	{
		if (x>=0 && y>=0 && x<imd.xs && y<imd.ys)
		{
			dword col;
			long offset=y*imd.xs+x;
			switch(pixblock)
			{
				case 1:
					col=(long)((byte *)data)[offset];
					//if (imd.pixelformat==PF_INDEX_8)
					//	return palette.col[col];
					return RGBIColor(0,0,0,(byte)col);
					break;
				case 2:
					col=((word *)data)[offset];
					break;
				case 3:
					break;
				case 4:
					col=((dword *)data)[offset];
			}
			d.FromType(imd.pixelformat,col);
		}
	}
	return d;
}

// ---------------------------------------------------------------------------

RGBColor Image::GetPixel(long x,long y)
{
	RGBColor d;
	if (ValidObject && Lock.IsLocked())
	{
		if (x>=0 && y>=0 && x<imd.xs && y<imd.ys)
		{
			dword col;
			long offset=y*imd.xs+x;
			switch(pixblock)
			{
				case 1:
					col=(long)((byte *)data)[offset];
					if (imd.pixelformat==PF_INDEX_8)
						return palette.col[col];
					else return RGBColor(0,0,0,(byte)col/256.0f);
					break;
				case 2:
					col=((word *)data)[offset];
					break;
				case 3:
					break;
				case 4:
					col=((dword *)data)[offset];
			}
			d.FromType(imd.pixelformat,col);
		}
	}
	return d;
}

// ---------------------------------------------------------------------------

void Image::Line(long x1,long y1,long x2,long y2,rRGBColor col)
{
	// All these variables are part of the bresenham line algorythm
	if (ValidObject && Lock.IsLocked())
	{
		long i,deltax,deltay,numpixels,d,dinc1,dinc2,x,xinc1,xinc2,y,yinc1,yinc2;

		deltax=labs(x2-x1);deltay=labs(y2-y1);

		if (deltax>=deltay)
		{
			numpixels	= deltax+1;
			d			= (deltay<<1)-deltax;
			dinc1		= deltay<<1;
			dinc2		= (deltay-deltax)<<1;
			xinc1		= 1;
			xinc2		= 1;
			yinc1		= 0;
			yinc2		= 1;
		}
		else 
		{
			numpixels=deltay+1;
			d		 = (deltax<<1)-deltay;
			dinc1	 = deltax<<1;
			dinc2	 = (deltax-deltay)<<1;
			xinc1	 = 0;
			xinc2	 = 1;
			yinc1	 = 1;
			yinc2	 = 1;
		}
		
		if (x1>x2)xinc1=-xinc1,xinc2=-xinc2;
		if (y1>y2)yinc1=-yinc1,yinc2=-yinc2;
		
		x=x1,y=y1;

		dword cl=col.ToType(imd.pixelformat);
		switch(pixblock)
		{
			case 1:
				for (i=1;i<numpixels;i++)
				{   
					if (x>=0 && y>=0 && x<=imd.xs && y<=imd.ys)
						((byte *)data)[x+y*imd.xs]=(byte)cl;
					if (d<0)d+=dinc1,x+=xinc1,y+=yinc1;else d+=dinc2,x+=xinc2,y+=yinc2;
				} 
			break;
			case 2:
				for (i=1;i<numpixels;i++)
				{   
					if (x>=0 && y>=0 && x<=imd.xs && y<=imd.ys)
						((word *)data)[x+y*imd.xs]=(word)cl;
					if (d<0)d+=dinc1,x+=xinc1,y+=yinc1;else d+=dinc2,x+=xinc2,y+=yinc2;
				} 
			break;
			case 3:
			break;
			case 4:
				for (i=1;i<numpixels;i++)
				{   
					if (x>=0 && y>=0 && x<=imd.xs && y<=imd.ys)
						((dword *)data)[x+y*imd.xs]=cl;
					if (d<0)d+=dinc1,x+=xinc1,y+=yinc1;else d+=dinc2,x+=xinc2,y+=yinc2;
				} 
			break;
		}
	}
}

// ---------------------------------------------------------------------------

void Image::LineInt(long x1,long y1,long x2,long y2,rRGBColor col1,rRGBColor col2)
{
	if (ValidObject && Lock.IsLocked())
	{

		// All these variables are part of the bresenham line algorythm
		long i,deltax,deltay,numpixels,d,dinc1,dinc2,x,xinc1,xinc2,y,yinc1,yinc2;

		deltax=labs(x2-x1);deltay=labs(y2-y1);

		if (deltax>=deltay){
			numpixels=deltax+1;
			d=(deltay<<1)-deltax;
			dinc1=deltay<<1;
			dinc2=(deltay-deltax)<<1;
			xinc1=1;
			xinc2=1;
			yinc1=0;
			yinc2=1;
		}
		else {
			numpixels=deltay+1;
			d=(deltax<<1)-deltay;
			dinc1=deltax<<1;
			dinc2=(deltax-deltay)<<1;
			xinc1=0;
			xinc2=1;
			yinc1=1;
			yinc2=1;
		}
		
		if (x1>x2)xinc1=-xinc1,xinc2=-xinc2;
		if (y1>y2)yinc1=-yinc1,yinc2=-yinc2;
		
		x=x1,y=y1;
		if (numpixels>0)
		{
			float frac=0;float fd=1/(float)(numpixels);
			switch (pixblock)
			{
				case 1:
				{
					for (i=1;i<numpixels;i++)
					{   
						if (x>=0 && y>=0 && x<=imd.xs && y<=imd.ys)
						((byte *)data)[x+y*imd.xs]=(byte)col1.Interpolate(col2,frac).ToType(imd.pixelformat);frac+=fd;
						if (d<0)d+=dinc1,x+=xinc1,y+=yinc1;
						else d+=dinc2,x+=xinc2,y+=yinc2;
					} 
				}
				break;
				case 2:
				{
					for (i=1;i<numpixels;i++)
					{   
						if (x>=0 && y>=0 && x<=imd.xs && y<=imd.ys)
						((word *)data)[x+y*imd.xs]=(word)col1.Interpolate(col2,frac).ToType(imd.pixelformat);frac+=fd;
						if (d<0)d+=dinc1,x+=xinc1,y+=yinc1;else d+=dinc2,x+=xinc2,y+=yinc2;
					} 
				}
				break;
				case 4:
					
					for (i=1;i<numpixels;i++)
					{   
						if (x>=0 && y>=0 && x<=imd.xs && y<=imd.ys)
						((dword *)data)[x+y*imd.xs]=col1.Interpolate(col2,frac).ToType(imd.pixelformat);frac+=fd;
						if (d<0)d+=dinc1,x+=xinc1,y+=yinc1;else d+=dinc2,x+=xinc2,y+=yinc2;
					} 
				break;
			}	
		}	
	}
}

// ---------------------------------------------------------------------------

void Image::LineH(long y,rRGBColor col)
{
	if (ValidObject && Lock.IsLocked())
	{
		if (y>=0 && y<imd.ys)
		{
			long offset=y*imd.xs,x;
			switch (pixblock)
			{
				case 1:
				{
					byte clr=(byte)col.ToType(imd.pixelformat);
					for (x=0;x<imd.xs;x++,offset++)
						((byte *)data)[offset]=clr;
				}
				break;
				case 2:
				{
					word clr=(word)col.ToType(imd.pixelformat);
					for (x=0;x<imd.xs;x++,offset++)
						((word *)data)[offset]=clr;
				}
				break;
				case 4:// HACK!!
				{
					dword clr=col.ToType(imd.pixelformat);
					for (x=0;x<imd.xs;x++,offset++)
						((dword *)data)[offset]=clr;
				}
				break;
			}
		}
	}
}
