/*
---------------------------------------------------------------------------
|			Image base class, deriviates implement storage
|		Base class implements 'generic' storage (new/delete buffer)
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/

#ifndef _H_IMAGE
#define _H_IMAGE

// ---------------------------------------------------------------------------
DECLARE_CLASS(Image);
DECLARE_CLASS(ImageDesc);

enum
{
	IMF_NOSAVEBIN =1<<0,	// doesn't save binary
	IMF_CLIENTOWN =1<<1,	// client owned data buffer
	IMF_CLIENTSET =1<<2		// used when setting the new data pointer (keeps certain structures)
};

class  ImageDesc
{
	public:
	ImageDesc(long ixs=-1,long iys=-1,long ipixelformat=-1,long iflags=0){clientdata=0,clientsize=0,xs=ixs,ys=iys,pixelformat=ipixelformat,flags=iflags;}
	cImageDesc operator=(cImageDesc o)								{ xs=o.xs,ys=o.ys,pixelformat=o.pixelformat,flags=o.flags,clientdata=o.clientdata,clientsize=o.clientsize;return *this;}

	 friend	rArchive operator<<(rArchive arc, rImageDesc b){arc<<b.xs<<b.ys<<b.pixelformat<<b.flags;return arc;} // archiving 
	 friend	rArchive operator>>(rArchive arc, rImageDesc b){arc>>b.xs>>b.ys>>b.pixelformat>>b.flags;return arc;} // plain raw image saving

	inline bool operator==(cImageDesc o)							{ return ((xs==o.xs) && (ys==o.ys) && (pixelformat==o.pixelformat)); }
	inline bool operator!=(cImageDesc o)							{ return (!((*this)==o));}
	long xs,ys,pixelformat,flags;
	long clientdata;	// pointer to surface data (when creating a CLIENTOWN image)
	long clientsize;	// size of client data (bugtracking)
};

class  Image
{
	public:
		Image();
		Image(ImageDesc &id);
		virtual ~Image();
		
		virtual long GetType(); // returns derived class type

		// doesn't obey LOCK since this is quite impossible in the base
		virtual bool Copy(Image &src,bool lock=true,bool equalsize=true,bool equalformat=true);
				
		virtual bool CompareTypes(Image &src);
		// returns true if equal

		// Data access
		virtual void *LockData();		// gives access to data block.. multithread safe
		virtual void UnlockData();		// Unlocks data

		// data storage functions
		virtual bool SetSize(ImageDesc &iid);
		virtual void SetSizeLike(Image &src);
		virtual void SetPalette(ColorTable &p);  // updates palette
				
		virtual bool IsValid()	{return(ValidObject);}	// has it got a valid image?

		// causes memory inside image to be released
		virtual void ReleaseBuffer(bool lock=true);
		
		virtual void UpdateContents(){};	// updates mipmap chains for mipmap textures

		// Image functions
		virtual void Clear(RGBColor &r=RGBColor(cBLACK));			// sets image to 0

		// Require you to 'LockData' first!!
		// Primitive drawing functions
		virtual void	 SetPixel(long x,long y,rRGBColor col);
		virtual RGBColor GetPixel(long x,long y);
		virtual void	 SetIPixel(long x,long y,rRGBIColor col);
		virtual RGBIColor GetIPixel(long x,long y);

		// normal line
		virtual void Line(long x1,long y1,long x2,long y2,rRGBColor col);
		virtual void Line(rVector2 p1,rVector2 p2,rRGBColor col)				   {Line(p1.x,p1.y,p2.x,p2.y,col);}

		// line with interpolated color 
		virtual void LineInt(long x1,long y1,long x2,long y2,rRGBColor col1,rRGBColor col2);
		virtual void LineInt(rVector2 p1,rVector2 p2,rRGBColor col1,rRGBColor col2){LineInt(p1.x,p1.y,p2.x,p2.y,col1,col2);}

		// fast horizontal line
		virtual void LineH(long y,rRGBColor col);

		 friend	rArchive operator<<(rArchive arc, rImage b){b.Save(arc);return arc;} // archiving 
		 friend	rArchive operator>>(rArchive arc, rImage b){b.Load(arc);return arc;} // plain raw image saving
		
		// overloaded loading and saving
		virtual void Load(rArchive arc);
		virtual void Save(rArchive arc);

		// Data
		long	  frame;		// image frame for video streams
		long	  cookie;		// some kind of data
		ImageDesc imd;			// image descriptor

		long	   pixblock;	// size of pixel block 1 2 3 4 bytes
		long	   datasize;	// size in bytes of data
		long	   pixsize;		// size in pixels of data

		bool ValidObject;		// wether the object contains a valid image

		ColorTable palette;		// palette for 8 bit images

		void *data;				// Linear data block.. only valid when locked (to outside world)
		
		MultithreadLock Lock;	// wether it's locked or not..(no access)
};

#endif