/*
---------------------------------------------------------------------------
|						Image encapsulation class
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
					All Rights Reserved
---------------------------------------------------------------------------
*/
#include "Library.h"

// ---------------------------------------------------------------------------

ImageGDI::ImageGDI()
{
	hBitmap=NULL;
}

// ---------------------------------------------------------------------------

ImageGDI::~ImageGDI()
{
	ReleaseBuffer();
}

// ---------------------------------------------------------------------------

void ImageGDI::ReleaseBuffer(bool lock)
{
	if(!lock || Lock.Lock())
	{
		if (IsValid())// delete old suface..
		{
			DeleteObject((HBITMAP)hBitmap);
			delete BitmapInfo;
			data=NULL;
			ValidObject=false;
			if (lock)Lock.Unlock();
		}
	}
}

// ---------------------------------------------------------------------------

bool ImageGDI::SetSize(ImageDesc &iimd)
{
	if(Lock.Lock())
	{
		if (IsNonzero(iimd.flags&IMF_CLIENTOWN))
		{
			Lock.Unlock();
//			log("ImageGDI::SetSize - no support for client owned surfaces!");
			return false;
		}
		long imsize=iimd.xs*iimd.ys*PF_Size(iimd.pixelformat);
		if (imsize>0)
		{
			ReleaseBuffer(false);

			BitmapInfo=new char[sizeof(BITMAPV4HEADER)];
			BITMAPV4HEADER &bh=*(BITMAPV4HEADER *)BitmapInfo;
			long am,rm,gm,bm;PF_MaskFromType(iimd.pixelformat,rm,gm,bm,am);

			bh.bV4Size			 = sizeof(BITMAPV4HEADER);
			bh.bV4Width			 = iimd.xs;
			bh.bV4Height		 = -iimd.ys;
			bh.bV4Planes		 = 1;
			bh.bV4BitCount		 = PF_BPPByte(PF_BppOfType(iimd.pixelformat))*8;
			bh.bV4V4Compression	 = BI_BITFIELDS;
			bh.bV4SizeImage		 = iimd.xs*iimd.ys*PF_Size(iimd.pixelformat);
			bh.bV4XPelsPerMeter	 = 0; 
			bh.bV4YPelsPerMeter	 = 0;
			bh.bV4ClrUsed		 = 0;
			bh.bV4ClrImportant	 = 0;
			bh.bV4RedMask	  	 = rm;
			bh.bV4GreenMask		 = gm;
			bh.bV4BlueMask		 = bm;
			bh.bV4AlphaMask		 = am;
			bh.bV4CSType		 = 0;
			HDC hDC=GetDC(NULL); 
			hBitmap= CreateDIBSection (hDC, (BITMAPINFO *)&bh, DIB_RGB_COLORS,
								  (void **)&data, NULL, 0);
			ReleaseDC(NULL,hDC);
			if (hBitmap==NULL || data==NULL)
			{
				//log("ImageGDI::SetSize() error creating DIBSection");
				data=NULL;
				Lock.Unlock();
				return(false);
			}
			imd=iimd;
			pixblock=PF_Size(imd.pixelformat);
			datasize=imd.xs*imd.ys*pixblock;
			pixsize =imd.xs*imd.ys;
			
			ValidObject=true;
			Lock.Unlock();
			return(true);
		}
		Lock.Unlock();
	}else assert(1==0);
	return(false);
};

// ---------------------------------------------------------------------------

void ImageGDI::SetPalette(ColorTable &p)
{
	palette=p;
}

// ---------------------------------------------------------------------------
// don't use MFC here now then
void ImageGDI::DrawGDI(void *tdc,void *trect,bool scale)
{
	if(Lock.Lock())
	{
		if (IsValid() && tdc!=NULL && trect!=NULL)
		{
			HDC  &odc  =*(HDC *)tdc;
			RECT &rect =*(RECT *)trect;
			HDC  imgdc;
			imgdc=CreateCompatibleDC(odc);
			HGDIOBJ prevobj;
			prevobj=SelectObject(imgdc,(HGDIOBJ)hBitmap);
						
			if (scale)
			{
				SetStretchBltMode(odc,COLORONCOLOR);
				StretchBlt(odc,rect.left,rect.top,rect.right-rect.left,rect.bottom-rect.top,imgdc,0,0,imd.xs,imd.ys,SRCCOPY);
			}
			else
			{
				BitBlt(odc,rect.left,rect.top,rect.right-rect.left,rect.bottom-rect.top,imgdc,0,0,SRCCOPY);
			}
			SelectObject(imgdc,(HGDIOBJ)prevobj);
			DeleteDC(imgdc);
		}
		Lock.Unlock();
	}
}

// ---------------------------------------------------------------------------

void ImageGDI::DrawGDI(CDC &dc,CRect &rect,bool scale,bool flipv)
{
	if(Lock.Lock())
	{
		if (IsValid())
		{
			HDC imgdc;
			imgdc = CreateCompatibleDC(dc.m_hDC);
			HGDIOBJ prevobj;
			prevobj=SelectObject(imgdc,(HGDIOBJ)hBitmap);
						
			if (scale)
			{
				SetStretchBltMode(dc.m_hDC,COLORONCOLOR);
				if (!flipv)
					StretchBlt(dc.m_hDC,rect.left,rect.top,rect.right-rect.left,rect.bottom-rect.top,imgdc,0,0,imd.xs,imd.ys,SRCCOPY);
				else
					StretchBlt(dc.m_hDC,rect.left,rect.top,rect.right-rect.left,rect.bottom-rect.top,imgdc,0,0,imd.xs,imd.ys,SRCCOPY);
			}
			else
			{
				BitBlt(dc.m_hDC,rect.left,rect.top,rect.right-rect.left,rect.bottom-rect.top,imgdc,0,0,SRCCOPY);
			}
			SelectObject(imgdc,(HGDIOBJ)prevobj);
			DeleteDC(imgdc);
		}
		Lock.Unlock();
	}
}