/*
---------------------------------------------------------------------------
|					All math utils you could ever imagine
|				  Copyright (C) 1998 VF Technologies and Inifite Reality Inc
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/

#include "Library.h"


// Build the tables
//----------------------------------------------------------
mfTables::mfTables()
{
	// Create the SqrtFrac table
	dword b,i;float f,g;
	for (i=0;i<MF_SQRT_ENTRIES;i++)
	{
		b=((i<<MF_SQRT_GET_FRACT)&MF_FRACTION_MASK)+((MF_EXPONENT_BIAS)<<(MF_GET_EXPONENT_SHIFT));
		f=(float)sqrt(mfFloat(b));
		if ( i&(MF_SQRT_ENTRIES>>1) ) 
			SqrtFrac[i]=(mfBin(f)&MF_FRACTION_MASK)+(1<<MF_GET_EXPONENT_SHIFT)+MF_SQRT_BIASMASK;
		else {
			f*=(float)sqrt(2.0f);
			SqrtFrac[i]=(mfBin(f)&MF_FRACTION_MASK)+MF_SQRT_BIASMASK;
		}
	};
	// create a fairly good float random table
	RandomCounter=0;
	for (i=0;i<MF_RANDOM;i++)
	{
		dword n=(rand()<<16)+(rand()<<1)+rand();
		Random[i]=((float)n/(float)0x3fffffff)-1.0; // make it a float between -1 and 1
		/*
		long s = 71 * ((rand()<<15)+rand());s = s * 8192 ^ s;
		Random[i]= 1.0 - ((s*(s*s*15731+789221)+1376312589)& 0x7fffffff)/1073741824.0;*/
	}
	// Create the OneOverSqrt and OneOver tables
	for (i=0;i<MF_FRACTION_ENTRIES;i++)
	{
		// Create a float off of the binary fraction
		g=(float)((i&((1<<9)-1))-MF_EXPONENT_BIAS);
		b=((i<<MF_FRACTION_APPROX_SHIFT)&MF_FRACTION_MASK)+((MF_EXPONENT_BIAS)<<(MF_GET_EXPONENT_SHIFT));
		f=mfFloat(b);
		OoSqrtFrac[i]=(float)pow((double)f,-0.5);		
		OoFrac[i]    =(float)pow((double)f,-1.0);		
	};
	for (i=0;i<MF_EXPONENT_ENTRIES;i++) 
	{
		g=(float)((((signed long)i)&((1<<8)-1))-MF_EXPONENT_BIAS);
		
		OoSqrtExp[i]=(float)pow(pow(2.0f,g),-0.5f);
		OoExp[i]    =(float)pow(pow(2.0f,g),-1.0f);
		// take sign flag in account..
		if ( (i&(1<<8))!=0 )OoExp[i]*=-1.0f;
	}
	for (i=0;i<MF_ANGLES;i++)  // 1024=2*Pi
	{
		float f=((float)(PI2)/1024.0f)*(float)i;
		Sine[i]=(float)sin(f);
		if (i<(MF_ANGLES/4))
			Saw[i]=(float)i/(float)(MF_ANGLES/4);
		else if (i<3*(MF_ANGLES/4))
			Saw[i]=1.0-(i-(MF_ANGLES/4))/(float)(MF_ANGLES/4);
		else
			Saw[i]=(i-3*(MF_ANGLES/4))/(float)(MF_ANGLES/4)-1.0f;
		ASine[i]=asin(1.0-(2.0/(float)MF_ANGLES)*(float)i);		
		ACosine[i]=acos(1.0-(2.0/(float)MF_ANGLES)*(float)i);		
	}
};
 mfTables			mfTab;										// Tables defined in gfc_math.cpp

//--------------------------------------------------------

//void BoundingBox2::ConvertToMFCRect(void *d){ CRect *c=(CRect *)d;c->top=Min.y,c->left=Min.x,c->bottom=Max.y,c->right=Max.x;}			
//void BoundingBox2::ConvertFromMFCRect(void *d){ CRect *c=(CRect *)d;Min.y=c->top,Min.x=c->left,Max.y=c->bottom,Max.x=c->right;}
		
//--------------------------------------------------------

float Matrix4x4::MatIdentity[4*4]={1,0,0,0,
						           0,1,0,0,
								   0,0,1,0,
								   0,0,0,1};

//--------------------------------------------------------

void Matrix4x4::SetRT(Vector3 *r,Vector3 *t)
{
	float cx,sx, cy,sy, cz,sz;
	mfSinCos(r->x*TRAD,sx,cx);mfSinCos(r->y*TRAD,sy,cy);mfSinCos(r->z*TRAD,sz,cz);
	val[0+0]=cy*cz;				val[0+1]=-sz*cy;			val[0+2]=sy;		val[0+3]=t->x;
	val[4+0]=(cz*sx*sy+sz*cx);	val[4+1]=(-sx*sy*sz+cx*cz);	val[4+2]=-sx*cy;	val[4+3]=t->y;
	val[8+0]=(-cx*sy*cz+sx*sz);	val[8+1]=(cx*sy*sz+cz*sx);	val[8+2]=cx*cy;		val[8+3]=t->z;
	val[12+0]=0;				val[12+1]=0;				val[12+2]=cx*cy;	val[12+3]=1;
}

//--------------------------------------------------------

void Matrix4x4::SetTR(Vector3 *r,Vector3 *t)
{
	float cx,sx, cy,sy, cz,sz;
	mfSinCos(r->x*TRAD,sx,cx);mfSinCos(r->y*TRAD,sy,cy);mfSinCos(r->z*TRAD,sz,cz);
	val[0+0]=cy*cz;				val[0+1]=-sz*cy;			val[0+2]=sy;		val[0+3]=val[0+0]*t->x+val[0+1]*t->y+val[0+2]*t->z;
	val[4+0]=(cz*sx*sy+sz*cx);	val[4+1]=(-sx*sy*sz+cx*cz);	val[4+2]=-sx*cy;	val[4+3]=val[4+0]*t->x+val[4+1]*t->y+val[4+2]*t->z;
	val[8+0]=(-cx*sy*cz+sx*sz);	val[8+1]=(cx*sy*sz+cz*sx);	val[8+2]=cx*cy;		val[8+3]=val[8+0]*t->x+val[8+1]*t->y+val[8+2]*t->z;
	val[12+0]=0;				val[12+1]=0;				val[12+2]=cx*cy;	val[12+3]=1;
}

//--------------------------------------------------------

void Matrix4x4::SetRZXY(Vector3 *r)
{
	/*
	Y*X*Z
	cycz+sxsysz		-szcy+czsxsy	sycx	0
	szcx			cxcz			-sx		0
	-sycz+cysxsz	sysz+cysxcz		cycx	0
	0				0				0		1	
	*/	
	float cx,sx, cy,sy, cz,sz;
	mfSinCos(r->x*TRAD,sx,cx);mfSinCos(r->y*TRAD,sy,cy);mfSinCos(r->z*TRAD,sz,cz);

	val[0+0]=cy*cz+sx*sy*sz	,val[0+1]=-sz*cy+cz*sx*sy,val[0+2]=sy*cx ,val[0+3] =0,
	val[4+0]=sz*cx			,val[4+1]=cx*cz			 ,val[4+2]=-sx	 ,val[4+3] =0,
	val[8+0]=-sy*cz+cy*sx*sz,val[8+1]=sy*sz+cy*sx*cz ,val[8+2]=cy*cx ,val[8+3] =0,
	val[12+0]=0				,val[12+1]=0			 ,val[12+2]=0	 ,val[12+3]=1;	
}

//--------------------------------------------------
void Matrix4x4::SetRTZXY(Vector3 *r,Vector3 *t)
{
	/*
	Y*X*Z
	cycz+sxsysz		-szcy+czsxsy	sycx	0
	szcx			cxcz			-sx		0
	-sycz+cysxsz	sysz+cysxcz		cycx	0
	0				0				0		1	
	*/	
	float cx,sx, cy,sy, cz,sz;
	mfSinCos(r->x*TRAD,sx,cx);mfSinCos(r->y*TRAD,sy,cy);mfSinCos(r->z*TRAD,sz,cz);

	val[0+0]=cy*cz+sx*sy*sz	,val[0+1]=-sz*cy+cz*sx*sy,val[0+2]=sy*cx ,val[0+3] =t->x,
	val[4+0]=sz*cx			,val[4+1]=cx*cz			 ,val[4+2]=-sx	 ,val[4+3] =t->y,
	val[8+0]=-sy*cz+cy*sx*sz,val[8+1]=sy*sz+cy*sx*cz ,val[8+2]=cy*cx ,val[8+3] =t->z,
	val[12+0]=0				,val[12+1]=0			 ,val[12+2]=0	 ,val[12+3]=1;	
}
//--------------------------------------------------
void Matrix4x4::SetSRTZXY(Vector3 *s,Vector3 *r,Vector3 *t)
{
	/*
	Y*X*Z
	cycz+sxsysz		-szcy+czsxsy	sycx	0
	szcx			cxcz			-sx		0
	-sycz+cysxsz	sysz+cysxcz		cycx	0
	0				0				0		1	
	*/	
	float cx,sx, cy,sy, cz,sz;
	mfSinCos(r->x*TRAD,sx,cx);mfSinCos(r->y*TRAD,sy,cy);mfSinCos(r->z*TRAD,sz,cz);

	val[0+0]=cy*cz+sx*sy*sz	,val[0+1]=-sz*cy+cz*sx*sy,val[0+2]=sy*cx ,val[0+3] =t->x,
	val[4+0]=sz*cx			,val[4+1]=cx*cz			 ,val[4+2]=-sx	 ,val[4+3] =t->y,
	val[8+0]=-sy*cz+cy*sx*sz,val[8+1]=sy*sz+cy*sx*cz ,val[8+2]=cy*cx ,val[8+3] =t->z,
	val[12+0]=0				,val[12+1]=0			 ,val[12+2]=0	 ,val[12+3]=1;	
}


//--------------------------------------------------------

void Matrix4x4::SetT(Vector3 *t)
{
	Identity();
	SetColumn(3,*t);
}

//--------------------------------------------------------

void Matrix4x4::SetR(Vector3 *r)
{
	/*
	cy*cz			-sz*cy			sy			0
	cz*sx*sy+sz*cx	-sx*sy*sz+cx*cz	-sx*cy		0
	-cx*sy*cz+sx*sz	cx*sy*sz+cz*sx	cx*cy		0
	0				0				0			1
	*/
	// Fill up matrix
	float cx,sx, cy,sy, cz,sz;
	mfSinCos(r->x*TRAD,sx,cx);mfSinCos(r->y*TRAD,sy,cy);mfSinCos(r->z*TRAD,sz,cz);
	val[0+0 ]=cy*cz;				val[0+1 ]=-sz*cy;				val[0+2 ]=sy;		val[0+3] =0;
	val[4+0 ]=(cz*sx*sy+sz*cx);		val[4+1 ]=(-sx*sy*sz+cx*cz);	val[4+2 ]=-sx*cy;	val[4+3] =0;
	val[8+0 ]=(-cx*sy*cz+sx*sz);	val[8+1 ]=(cx*sy*sz+cz*sx);		val[8+2 ]=cx*cy;	val[8+3] =0;
	val[12+0]=0;					val[12+1]=0;					val[12+2]=cx*cy;	val[12+3]=1;
}

//--------------------------------------------------------

void Matrix4x4::SetProjectFOV(Vector2 &fov,float znear,float zfar)
{
	// clear matrix
	Zero();

	float q=zfar/(zfar-znear);
	float w=1.0f/mfTan(0.5*fov.x*TRAD);
	float h=1.0f/mfTan(0.5*fov.y*TRAD);
	val[0+0]=w,// 0               0                0				
/*  0		*/	val[4+1]=h//    0                0
/*	0            0 	*/			,val[8+2]=q		 ,val[8+3] =-q*znear
/*	0  			 0 */           ,val[12+2]=1;//   0
}

//--------------------------------------------------------

void Matrix4x4::SetProjectVPR(Vector2 &vpr,float znear,float zfar)
{
	// clear matrix
	Zero();

	float q=zfar/(zfar-znear);
	float w=2*znear/vpr.x;
	float h=2*znear/vpr.y;
	val[0+0]=w,// 0               0                0				
/*  0		*/	val[4+1]=h//    0                0
/*	0            0 	*/			,val[8+2]=q		 ,val[8+3] =-q*znear
/*	0  			 0 */           ,val[12+2]=1;//   0
}

//--------------------------------------------------------

void Matrix4x4::SetS(Vector3 *s)
{
	Zero();
	val[0+0]=s->x	  /*0*/				/*0*/					 /*0*/					 
	/*0*/			 ,val[4+1]=s->y		/*0*/					 /*0*/					 
	/*0*/			 /*0*/ 				,val[8+2] =s->z			 /*0*/					 
	/*0*/  		 	 /*0*/				/*0*/					 ,val[12+3]=1;

}

//--------------------------------------------------------

void Matrix4x4::SetClip(Vector2 &t,Vector2 &w,Vector2 &d)
{
	// clear matrix
	Zero();

	val[0+0]=0.5*w.x  /*0*/				/*0*/					 ,val[0+3]=-1-2*(t.x/w.x)
	/*0*/			 ,val[4+1]=0.5*w.y  /*0*/					 ,val[4+3]= 1-2*(t.y/w.y) 
	/*0*/			 /*0*/ 				,val[8+2] =1.0f/(d.y-d.x),val[8+3]=-d.x/(d.y-d.x)
	/*0*/  		 	 /*0*/				/*0*/					 ,val[12+3]=1;
}

//--------------------------------------------------------

Matrix4x4	Matrix4x4::Transpose()
{
	Matrix4x4 m;
	m.val[0+0 ]=val[0 +0 ];m.val[0+1 ]=val[4 +0 ];m.val[0+2 ]=val[8 +0 ];m.val[0+3 ]=val[12+0 ];
	m.val[4+0 ]=val[0 +1 ];m.val[4+1 ]=val[4 +1 ];m.val[4+2 ]=val[8 +1 ];m.val[4+3 ]=val[12+1 ];
	m.val[8+0 ]=val[0 +2 ];m.val[8+1 ]=val[4 +2 ];m.val[8+2 ]=val[8 +2 ];m.val[8+3 ]=val[12+2 ];
	m.val[12+0]=val[0 +3 ];m.val[12+1]=val[4 +3 ];m.val[12+2]=val[8 +3 ];m.val[12+3]=val[12+3 ];
	return m;
}

//--------------------------------------------------------

Matrix4x4	Matrix4x4::TransposeMinor()
{	
	Matrix4x4 m;
	m.val[0+0 ]=val[0 +0 ];m.val[0+1 ]=val[4 +0 ];m.val[0+2 ]=val[0 +2 ];m.val[0+3 ]=val[8 +0 ];
	m.val[4+0 ]=val[0 +1 ];m.val[4+1 ]=val[4 +1 ];m.val[4+2 ]=val[8 +1 ];m.val[4+3 ]=val[4 +3 ];
	m.val[8+0 ]=val[0 +3 ];m.val[8+1 ]=val[4 +2 ];m.val[8+2 ]=val[8 +2 ];m.val[8+3 ]=val[8 +3 ];
	m.val[12+0]=val[12+0 ];m.val[12+1]=val[12+1 ];m.val[12+2]=val[12+2 ];m.val[12+3]=val[12+3 ];
	return m;
}

//--------------------------------------------------------
				
Matrix4x4	Matrix4x4::Inverse()
{
	Matrix4x4 m=TransposeMinor();
	Vector3   v=-1*m.GetColumn(3);
	v=m.MulMinor(v);m.SetColumn(3,v);
	return(m);
}

//--------------------------------------------------------

Matrix4x4  &Matrix4x4::Normalize()
{
	float dif=0;long i=0;
	for (i=0;i<16;i++)dif+=val[i]*val[i];
	dif=1.0f/sqrt(dif);
	for (i=0;i<16;i++)val[i]*=dif;
	return *this;
}

//--------------------------------------------------------

Vector2	Matrix4x4::TransformAndProject(rVector3 src,rVector2 Aspect,rVector2 centre)
{
	Vector3 c;
	c=(*this) * src;
	if (c.z<=0)return( Vector2(0,0) );
	return (Vector2 (c.x*Aspect.x/c.z+centre.x, centre.y-c.y*Aspect.y/c.z) );
}

//--------------------------------------------------------

Orient3& Orient3::FromVector(rVector3 tgt,float up)       // gets rotation from targetting vector
{
	// roll = up..
	Set(mfArcTan2(tgt.y,tgt.z)*TDEG,mfArcTan2(-tgt.x,tgt.z)*TDEG,up);
	return(*this);
}

//--------------------------------------------------------

Matrix4x4 Orient3::GetMatrix(Vector3 &Pos)								// gets matrix for transform
{
	Matrix4x4 m;
	m.SetRZXY(Vector());
	m.SetColumn(3,Pos);
	return m;
}

//--------------------------------------------------------

Matrix4x4 Orient3::GetRotMatrix()								// gets matrix for rotation
{
	Matrix4x4 m;
	m.SetRZXY(Vector());
	return m;
}

//--------------------------------------------------------

void  Vector2::ClampPow2(float maxx,float maxy)
{
	long xm=maxx,ym=maxy;
		
	while (xm/x>=1.5 && xm>4)xm=xm>>1; // shift left one
	while (ym/y>=1.5 && ym>4)ym=ym>>1; // shift left one
				
	x=xm;
	y=ym;
}

//--------------------------------------------------------

Vector3 &Vector3::GetNormal(Vector3 &a,Vector3 &b, Vector3 &c)
{
	Vector3  v = b-a, w = b-c;
	Set(v.y*w.z - v.z*w.y, v.z*w.x - v.x*w.z, v.x*w.y - v.y*w.x);
	return *this;
}

//--------------------------------------------------------

Quaternion::Quaternion()
{
  W = 1.0;
  X = 0.0;
  Y = 0.0;
  Z = 0.0;
}

//--------------------------------------------------------

Quaternion::Quaternion(const float w, const float x, const float y, const float z)
{
  W = w;
  X = x;
  Y = y;
  Z = z;
}
//--------------------------------------------------------

Quaternion::Quaternion(Vector4 &vec)
{
  W = vec.w;
  X = vec.x;
  Y = vec.y;
  Z = vec.z;
}

//--------------------------------------------------------

Quaternion operator * (Quaternion &a, Quaternion &b)
{
  float w,x,y,z;

  w = a.W*b.W - (a.X*b.X + a.Y*b.Y + a.Z*b.Z);
  
  x = a.W*b.X + b.W*a.X + a.Y*b.Z - a.Z*b.Y;
  y = a.W*b.Y + b.W*a.Y + a.Z*b.X - a.X*b.Z;
  z = a.W*b.Z + b.W*a.Z + a.X*b.Y - a.Y*b.X;

  return Quaternion(w,x,y,z); 
}

//--------------------------------------------------------

Quaternion& Quaternion::operator *= (Quaternion &q)
{
  float w = W*q.W - (X*q.X + Y*q.Y + Z*q.Z);

  float x = W*q.X + q.W*X + Y*q.Z - Z*q.Y;
  float y = W*q.Y + q.W*Y + Z*q.X - X*q.Z;
  float z = W*q.Z + q.W*Z + X*q.Y - Y*q.X;

  W = w;
  X = x;
  Y = y;
  Z = z;
  return *this;
}

//--------------------------------------------------------

Quaternion& Quaternion::operator ~ ()
{
  X = -X;
  Y = -Y;
  Z = -Z;
  return *this;
}
//--------------------------------------------------------

Quaternion& Quaternion::operator - ()
{
  float norme = sqrt(W*W + X*X + Y*Y + Z*Z);
  if (norme == 0.0)
    norme = 1.0;

  float recip = 1.0 / norme;

  W =  W * recip;
  X = -X * recip;
  Y = -Y * recip;
  Z = -Z * recip;

  return *this;
}

//--------------------------------------------------------

Quaternion& Quaternion::Normalize()
{
  float norme = sqrt(W*W + X*X + Y*Y + Z*Z);
  if (norme == 0.0)
    {
	  W = 1.0; 
	  X = Y = Z = 0.0;
	}
  else
    {
	  float recip = 1.0/norme;

	  W *= recip;
	  X *= recip;
	  Y *= recip;
	  Z *= recip;
	}
  return *this;
}

//--------------------------------------------------------

Quaternion& Quaternion::FromAxis(float Angle, Vector3 &vec)
{
  float omega, s, c,x=vec.x,y=vec.y,z=vec.z;
   
  s = sqrt(x*x + y*y + z*z);
  
  if (fabs(s) > FLT_EPSILON)
    {
	  c = 1.0/s;
	  
	  x *= c;
	  y *= c;
	  z *= c;

	  omega = -0.5f * Angle*TRAD;
	  s = (float)sin(omega);

	  X = s*x;
	  Y = s*y;
	  Z = s*z;
	  W = (float)cos(omega);
	}
  else
    {
	  X = Y = 0.0f;
	  Z = 0.0f;
	  W = 1.0f;
	}
  Normalize();
  return *this;
}


//--------------------------------------------------------

void Quaternion::ToMatrix(Matrix4x4 &m)
{
	
 m.val[0+0 ]=1.0 - 2*Y*Y - 2*Z*Z; m.val[0+1 ]=2*X*Y - 2*W*Z;		m.val[0+2 ]=2*X*Z + 2*W*Y;		 m.val[0+3 ]=0;
 m.val[4+0 ]=2*X*Y + 2*W*Z;       m.val[4+1 ]=1.0 - 2*X*X - 2*Z*Z;	m.val[4+2 ]=2*Y*Z - 2*W*X;		 m.val[4+3 ]=0;
 m.val[8+0 ]=2*X*Z - 2*W*Y;       m.val[8+1 ]=2*Y*Z + 2*W*X;		m.val[8+2 ]=1.0 - 2*X*X - 2*Y*Y; m.val[8+3 ]=0;
 m.val[12+0]=0;					  m.val[12+1]=0;					m.val[12+2]=0;					 m.val[12+3]=1;
/*
 m.val[0+0 ]=1.0 - 2*Y*Y - 2*Z*Z; m.val[0+1 ]=2*X*Y + 2*W*Z;		m.val[0+2 ]=2*X*Z - 2*W*Y;		 m.val[0+3 ]=0;
 m.val[4+0 ]=2*X*Y - 2*W*Z;       m.val[4+1 ]=1.0 - 2*X*X - 2*Z*Z;	m.val[4+2 ]=2*Y*Z + 2*W*X;		 m.val[4+3 ]=0;
 m.val[8+0 ]=2*X*Z + 2*W*Y;       m.val[8+1 ]=2*Y*Z - 2*W*X;		m.val[8+2 ]=1.0 - 2*X*X - 2*Y*Y; m.val[8+3 ]=0;
 m.val[12+0]=0;					  m.val[12+1]=0;					m.val[12+2]=0;					 m.val[12+3]=1;

  matrix[0][0] = 1.0 - 2*Y*Y - 2*Z*Z;
  matrix[1][0] = 2*X*Y + 2*W*Z;
  matrix[2][0] = 2*X*Z - 2*W*Y;

  matrix[0][1] = 2*X*Y - 2*W*Z;
  matrix[1][1] = 1.0 - 2*X*X - 2*Z*Z;
  matrix[2][1] = 2*Y*Z + 2*W*X;

  matrix[0][2] = 2*X*Z + 2*W*Y;
  matrix[1][2] = 2*Y*Z - 2*W*X;
  matrix[2][2] = 1.0 - 2*X*X - 2*Y*Y;
  */
}

//--------------------------------------------------------


void Quaternion::Slerp(Quaternion &a,Quaternion &b, float t)
{
  float omega, cosom, sinom, sclp, sclq;


  cosom = a.X*b.X + a.Y*b.Y + a.Z*b.Z + a.W*b.W;


  if ((1.0f+cosom) > FLT_EPSILON)
	{
	  if ((1.0f-cosom) > FLT_EPSILON)
		{
		  omega = acos(cosom);
		  sinom = sin(omega);
		  sclp = sin((1.0f-t)*omega) / sinom;
		  sclq = sin(t*omega) / sinom;
		}
	  else
		{
		  sclp = 1.0f - t;
		  sclq = t;
		}

      X = sclp*a.X + sclq*b.X;
	  Y = sclp*a.Y + sclq*b.Y;
	  Z = sclp*a.Z + sclq*b.Z;
	  W = sclp*a.W + sclq*b.W;

	}
  else
	{
	  X =-a.Y;
	  Y = a.X;
	  Z =-a.W;
	  W = a.Z;

	  sclp = sin((1.0f-t) * PI * 0.5);
	  sclq = sin(t * PI * 0.5);

	  X = sclp*a.X + sclq*b.X;
	  Y = sclp*a.Y + sclq*b.Y;
	  Z = sclp*a.Z + sclq*b.Z;

	}
}

//--------------------------------------------------------

Quaternion& Quaternion::exp()
{                               
  float Mul;
  float Length = sqrt(X*X + Y*Y + Z*Z);

  if (Length > 1.0e-4)
    Mul = sin(Length)/Length;
  else
    Mul = 1.0;

  W = cos(Length);

  X *= Mul;
  Y *= Mul;
  Z *= Mul; 

  return *this;
}

//--------------------------------------------------------

Quaternion& Quaternion::log()
{
  float Length;

  Length = sqrt(X*X + Y*Y + Z*Z);
  Length = atan(Length/W);

  W = 0.0;

  X *= Length;
  Y *= Length;
  Z *= Length;

  return *this;
}


//--------------------------------------------------------
