
/*
---------------------------------------------------------------------------
|					All math utils you could ever imagine
|					  Copyright (C) 1998 VF Technologies
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/

// Generic utils
// Include this as the base thing..

// Mathematical constants
#ifndef _H_MATH
#define _H_MATH
#define PI	3.1415926535897932384626433
#define PI2	2.0*PI
#define PI12	0.5*PI

#define TRAD 0.0174532925199433
#define TDEG 57.295779513082323
#define TGRAD 0.159154943091895

#define	E		2.7182818284590452353602874		
#define DELTA	(0.01f)
#ifndef FLT_MAX
#define FLT_MAX	1000000000.0f
#endif
enum {	
	MF_FRACTION_MASK			= 0x007FFFFF,	MF_GET_FRACTION_SHIFT	= 0,	MF_FRACTION_BITS = 23,
	MF_EXPONENT_MASK			= 0x7F800000,	MF_GET_EXPONENT_SHIFT	= 23,	MF_EXPONENT_BITS = 8,
	MF_SIGN_MASK				= 0x80000000,	MF_GET_SIGN_SHIFT		= 31,	MF_SIGN_BITS = 1,
	MF_EXPONENT_BIAS			= 127,

	// Constants used for sine approximation
	MF_ANGLES					= 1024,										// 1024 angles
	MF_RANDOM					= 16384,
	// Constants used for power-table approximation
	MF_FRACTION_APPROX_BITS		= 10,										// Fraction bits in approximation
	MF_FRACTION_ENTRIES			= 1 << MF_FRACTION_APPROX_BITS,				// Entries needed for fraction approximation
	MF_FRACTION_APPROX_MASK		= MF_FRACTION_ENTRIES-1,					// Mask for fraction table
	MF_FRACTION_APPROX_SHIFT	= MF_FRACTION_BITS-MF_FRACTION_APPROX_BITS,	// Shifter to get approximation fraction
	MF_EXPONENT_APPROX_BITS		= MF_EXPONENT_BITS+MF_SIGN_BITS,			// Bits in exponent table to approximate
	MF_EXPONENT_ENTRIES			= 1 << MF_EXPONENT_APPROX_BITS,				// Entries needed for exponent approximation (full range)
	MF_EXPONENT_SIGN_MASK		= MF_EXPONENT_ENTRIES-1,					// Mask for table lookup

	// Constants for sqrt(f) approximation
	MF_SQRT_APPROX_BITS			= 10,										// Fraction bits to approximate
	MF_SQRT_TABLE_BITS			= MF_SQRT_APPROX_BITS+1,					// Table is twice the size for odd/even exponent
	MF_SQRT_ENTRIES				= 1 << MF_SQRT_TABLE_BITS,					// Entries in sqrt table
	MF_SQRT_GET_FRACT			= MF_FRACTION_BITS - MF_SQRT_APPROX_BITS,	// Shifter to get fraction for table index
	MF_SQRT_MASK				= MF_SQRT_ENTRIES - 1,						// Mask for index in sqrt table
	MF_SQRT_MASK_EXP			= MF_FRACTION_MASK & (MF_FRACTION_MASK<<1),	// Mask for getting exponent/2
	MF_SQRT_EXPBIAS			    = (MF_EXPONENT_BIAS-1)<<(MF_GET_EXPONENT_SHIFT), // Masks added by Arc
	MF_SQRT_BIASMASK			= MF_SQRT_EXPBIAS-(MF_SQRT_EXPBIAS>>1)
};


// Fast float approximation data (for table-based math approximations)
struct mfTables {
	mfTables();														// Tables are inited on construction
	dword SqrtFrac[MF_SQRT_ENTRIES];								// For fast Sqrt(f)
	float OoSqrtFrac[MF_FRACTION_ENTRIES],	OoSqrtExp[MF_EXPONENT_ENTRIES];// For fast 1/Sqrt(f)
	float OoFrac[MF_FRACTION_ENTRIES],		OoExp[MF_EXPONENT_ENTRIES];	// For fast 1/f
	float Sine[MF_ANGLES];											// For fast Sin(i) and Cos(i)
	float Saw[MF_ANGLES];										// For fast Sin(i) and Cos(i)
	float ASine[MF_ANGLES];										
	float ACosine[MF_ANGLES];										
	float Random[MF_RANDOM];
	long  RandomCounter;
};
extern  mfTables			mfTab;										// Tables defined in gfc_math.cpp

#define FLOAT_TO_20_0	(1 << 20)								// Round float to fixed-point 20.0  format
#define FLOAT_TO_12_8	(1 << 12)								// Round float to fixed-point 12.8  format
#define FLOAT_TO_4_16	(1 <<  4)								// Round float to fixed-point  4.16 format

// IEEE data access
inline dword		mfBin(float f)								{ return *(dword *)&f;							}
inline float		mfFloat(dword bin)							{ return *(float *)&bin;						}

// Explicit binary operations
inline float		mfOR(float f,  int om)						{ return mfFloat(mfBin(f) | om);				}
inline float		mfXOR(float f, int xm)						{ return mfFloat(mfBin(f) ^ xm);				}
inline float		mfAND(float f, int am)						{ return mfFloat(mfBin(f) & am);				}

// Binary operations
inline int		mfFraction(float f)							{ return mfBin(f) & MF_FRACTION_MASK;			}
inline byte		mfExponent(float f)							{ return (byte)(mfBin(f) >> MF_GET_EXPONENT_SHIFT);	}
inline dword	mfSignMask(float f)							{ return mfBin(f) & MF_SIGN_MASK;				}
inline int		mfSignBit(float f)							{ return mfBin(f) >> MF_GET_SIGN_SHIFT;			}
inline int		mfSign(float f)								{ return ((((int)mfSignMask(f) >> 30)+1) & (((50<<MF_GET_EXPONENT_SHIFT) - (int)(mfBin(f) & MF_EXPONENT_MASK)) >> 31)); }
inline int		mfSign(float f, float delta)				{ if (f < -delta) return -1; else if (f > delta) return 1; else return 0; }
inline short	mfSignB(double f)							{ if (f<0)return(1);else return(0);}
inline bool		mfIsPositive(float f)						{ return mfSignMask(f)==0;						}
inline bool		mfIsNegative(float f)						{ return mfSignMask(f)!=0;						}

// Data conversion
inline uint		mfFixed20_0(float f)						{ double res = (f + (0x7000/(    1.0 * 65536.0)) + FLOAT_TO_20_0);	return *(((int *)&res)+1) & 0xffff; }
inline uint		mfFixed12_8(float f)						{ double res = (f + (0x7000/(  256.0 * 65536.0)) + FLOAT_TO_12_8);	return *(((int *)&res)+1) & 0xffff; }
inline uint		mfFixed4_16(float f)						{ double res = (f + (0x7000/(65536.0 * 65536.0)) + FLOAT_TO_4_16);	return *(((int *)&res)+1) & 0xffff; }
inline int		mfIntRound(float f)							{ double res = (f + (0x7000/(    1.0 * 65536.0)) + FLOAT_TO_20_0);	return *(((int *)&res)+1) & 0xffff; }
inline int		mfInt(float f)								{ double res = (f + FLOAT_TO_20_0);	return *(((int *)&res)+1) & 0xfffff; }

// Numerical operations
inline float	mfAbs(float f)								{ return mfAND(f, ~MF_SIGN_MASK);				}
inline float	mfNeg(float f)								{ return mfXOR(f, MF_SIGN_MASK);				}
inline float	mfMakeNeg(float f)							{ return mfOR(f, MF_SIGN_MASK);					}
inline bool		mfIsInfinite(float f)						{ return ((mfBin(f) & (MF_EXPONENT_MASK | MF_FRACTION_MASK)) == MF_EXPONENT_MASK); } 
inline bool		mfIsFinite(float f)							{ return !mfIsInfinite(f);						}
inline bool		mfIsANumber(float f)						{ return ((mfBin(f) & MF_EXPONENT_MASK) != MF_EXPONENT_MASK); }
inline bool		mfIsNAN(float f)							{ return ((mfBin(f) & MF_EXPONENT_MASK) == MF_EXPONENT_MASK) && (mfBin(f) & MF_FRACTION_MASK); }
inline bool		mfIsNearZero(float f, int exp = -80)		{ return (mfExponent(f)<=(exp+MF_EXPONENT_BIAS));	}
inline bool		mfIsNear(float f, float ref, int exp = -80)	{ return mfIsNearZero(f-ref, exp);				}
inline bool		mfIsEqual(float& f1, float &f2)				{ return (*(dword *)&f1) == (*(dword *)&f2);	}
inline bool		mfIsNotEqual(float& f1, float &f2)			{ return (*(dword *)&f1) != (*(dword *)&f2);	}

// Complex math
inline float	mfPow(float f, float p)						{ return (float)pow(f, p);						}
inline float	mfMod(float f, float mod)					{ return (float)fmod(f, mod);					}
inline float	mfSqrt(float f)								{ return (float)sqrt(f);						}
inline float	mfFSqrt(float f)							{ return mfFloat(mfTab.SqrtFrac[(mfBin(f) >> MF_SQRT_GET_FRACT) & MF_SQRT_MASK] + ((mfBin(f) >> 1) & MF_EXPONENT_MASK)); }
inline float	mfOneOverSqrt(float f)						{ return (mfTab.OoSqrtFrac [(mfBin(f) >> MF_FRACTION_APPROX_SHIFT) & MF_FRACTION_APPROX_MASK] * mfTab.OoSqrtExp [(mfBin(f) >> MF_GET_EXPONENT_SHIFT) & MF_EXPONENT_SIGN_MASK]); }
inline float	mfOneOver(float f)							{ return (mfTab.OoFrac     [(mfBin(f) >> MF_FRACTION_APPROX_SHIFT) & MF_FRACTION_APPROX_MASK] * mfTab.OoExp     [(mfBin(f) >> MF_GET_EXPONENT_SHIFT) & MF_EXPONENT_SIGN_MASK]); }

// Trigonometric math
inline float	mfSin(float angle)							{ return (float)sin(angle);						}
inline float	mfCos(float angle)							{ return (float)cos(angle);						}
inline float	mfFSin(int angle)							{ return mfTab.Sine[angle & (MF_ANGLES-1)];		}
inline float	mfFCos(int angle)							{ return mfTab.Sine[(angle + MF_ANGLES/4) & (MF_ANGLES-1)];	}
inline void		mfFSinCos(int angle, float& s, float& c)	{ s=mfTab.Sine[angle & (MF_ANGLES-1)];	c=mfTab.Sine[(angle + MF_ANGLES/4) & (MF_ANGLES-1)]; }
inline float	mfTan(float angle)							{ return (float)tan(angle);							}
inline float	mfArcSin(float angle)						{ return (float)asin(angle);						}
inline float	mfArcCos(float angle)						{ return (float)acos(angle);						}
inline float	mfArcTan(float angle)						{ return (float)atan(angle);						}
inline float	mfArcTan2(float x, float y)					{ return (float)atan2(x, y);						}
inline void		mfSinCos(float angle, float& s, float& c)	{ __asm push	ecx __asm fld [angle] __asm fsincos __asm mov ecx, dword ptr[c] __asm fstp dword ptr [ecx] __asm mov ecx, dword ptr[s] __asm fstp dword ptr [ecx] __asm pop ecx }

// Overloads
//inline float		abs(float f)								{ return mfAbs(f);}

// 2d vector..
DECLARE_CLASS(Vector2);
class  Vector2
{
public:

//	virtual long   ObjectID(){return(IDVECTOR2);}
//	virtual String ObjectName(){return(String("BASEOBJ"));}

	// Initialization
	Vector2()														{ assert((&v) == (&u+1)); }
	Vector2(float u)												: u(u), v(u) { }
	Vector2(float u, float v)										: u(u), v(v) { }
	Vector2(float v[2])												: u(v[0]), v(v[1]) { }
	Vector2(cVector2 other)											: u(other.u), v(other.v) { }
	cVector2 operator=(cVector2 other)								{ u=other.u; v=other.v; return *this; }
	cVector2 operator=(float    other)								{ u=other; v=other; return *this; }
	inline void Set(float xx, float yy)								{ u=xx; v=yy; }

	// Squared and normal length of vector
	inline float	SquareLength() const							{ return (u*u + v*v); }
	inline float	Length() const									{ return mfSqrt(SquareLength()); }
	
	// Normalize vector, |v| = 1.0
	inline Vector2& Normalize()										{ float iLen = 1.0f/Length(); u*=iLen; v*=iLen; return *this; }
	inline Vector2& FNormalize()									{ float iLen = mfOneOverSqrt(SquareLength()); u*=iLen; v*=iLen; return *this; }
	void   ClampPow2(float maxx,float maxy);
	// Access operators
	inline float&	operator[](int row)								{ assert(row<2); return ((float *)&u)[row]; }
	inline operator	float*()										{ return &u; }
	inline operator	const float*() const							{ return &u; }

	// Comparison
	inline bool operator==(cVector2 rhs)							{ return ((u==rhs.u) && (v==rhs.v)); }

	// Vector addition/substraction
	inline Vector2&	Scale(cVector2 rhs)								{ u*=rhs.u; v*=rhs.v; return *this; }
	inline Vector2  operator-() const								{ return Vector2(mfNeg(u), mfNeg(v)); }
	inline cVector2 operator+=(cVector2 rhs)						{ u+=rhs.u; v+=rhs.v; return *this; }
	inline cVector2 operator-=(cVector2 rhs)						{ u-=rhs.u; v-=rhs.v; return *this; }
	inline friend Vector2 operator+(cVector2 lhs, cVector2 rhs)		{ return Vector2(lhs.u+rhs.u, lhs.v+rhs.v); }
	inline friend Vector2 operator-(cVector2 lhs, cVector2 rhs)		{ return Vector2(lhs.u-rhs.u, lhs.v-rhs.v); }
	
	// Vector*scalar and vector/scalar
	inline cVector2 operator*=(float m)								{ u*=m; v*=m; return *this; }
	inline cVector2 operator*=(cVector2 dhs)						{ u*=dhs.u; v*=dhs.v; return *this; }
	inline friend Vector2 operator*(cVector2 lhs,cVector2 dhs)		{ return Vector2(lhs.u*dhs.u, lhs.v*dhs.v); }
	inline friend Vector2 operator*(cVector2 lhs, float m)			{ return Vector2(lhs.u*m, lhs.v*m); }
	inline friend Vector2 operator*(float m, cVector2 rhs)			{ return Vector2(m*rhs.u, m*rhs.v); }
	inline cVector2 operator/=(float d)								{ float	 iMul=1.0f/d; u*=iMul; v*=iMul; return *this; }
	inline cVector2 operator/=(cVector2 dhs)						{ u/=dhs.u; v/=dhs.v; return *this; }
	inline friend Vector2 operator/(cVector2 lhs, float d)			{ float  iMul=1.0f/d; return Vector2(lhs.u*iMul, lhs.v*iMul); }
	inline friend Vector2 operator/(float d, cVector2 rhs)			{ return Vector2(d/rhs.u, d/rhs.v); }
	inline friend Vector2 operator/(cVector2 rhs,cVector2 dhs)		{ return Vector2(rhs.u/dhs.u,rhs.v/dhs.v); }
	inline friend bool    operator>(cVector2 rhs,cVector2 dhs)      { if (rhs.x>dhs.x)return(true);if(rhs.y>dhs.y)return(true);return(false);}
	inline friend bool    operator<(cVector2 rhs,cVector2 dhs)      { return (dhs>rhs);}
	inline friend bool    operator<=(cVector2 rhs,cVector2 dhs)     { if (rhs.x<=dhs.x)return(true);if(rhs.y<=dhs.y)return(true);return(false);}
	inline friend bool    operator>=(cVector2 rhs,cVector2 dhs)     { return (dhs>=rhs);}
	// transformations
	inline Vector2 Transform_st (cVector2 s,cVector2 t) const	    { return Vector2((x+t.x)*s.x,(y+t.y)*s.y); }
	inline Vector2 Transform_stt(cVector2 s,cVector2 t,cVector2 t2)const
																	{ return Vector2((x+t.x)*s.x+t2.x*s.x,(y+t.y)*s.y+t2.y*s.y); }
	inline	Vector2 Rotate(float r)									{ float cr,sr;mfSinCos(r,sr,cr);return Vector2(cr*x-sr*y,cr*y+sr*x);}
	inline	Vector2 RotateAround(Vector2 &c,float r)				{ float cr,sr;mfSinCos(r,sr,cr);return Vector2(cr*(x-c.x)-sr*(y-c.y)+c.x,cr*(y-c.y)+sr*(x-c.x)+c.y);}
	// Serializing.
	inline friend rArchive operator<<(rArchive arc, cVector2 v)		{ arc << v.u << v.v; return arc; }
	inline friend rArchive operator>>(rArchive arc, rVector2 v)		{ arc >> v.u >> v.v; return arc; }

	// Vector2 -> String
	inline String ToString() const									{ String t; t.Format("(x: %8.3f, y: %8.3f)", x, y); return t; }

	// Data members
	union { float x; float u; };									// x or u coordinate
	union { float y; float v; };									// y or v coordinate
};

//
// Vector in 3-space
//
DECLARE_CLASS(Vector3);
class  Vector3
{
public:
	// Initialization
	Vector3()														{  }
	Vector3(float x, float y, float z)								: x(x), y(y), z(z) { }
	Vector3(float v[3])												: x(v[0]), y(v[1]), z(v[2]) { }
	Vector3(cVector3 other)											: x(other.x), y(other.y), z(other.z) { }
	cVector3 operator=(cVector3 other)								{ x=other.x; y=other.y; z=other.z; return *this; }
	inline void Set(float xx, float yy, float zz)					{ x=xx; y=yy; z=zz; }

	// Dot product (.): cos phi = (v.w)/(|v||w|)
	//					cos phi =  v.w              when |v|=|w|=1
	inline float	Dot(cVector3 rhs) const							{ return (x*rhs.x + y*rhs.y + z*rhs.z); }

	// Cross product (x)
	inline Vector3	Cross(cVector3 rhs) const						{ return Vector3(y*rhs.z - z*rhs.y, z*rhs.x - x*rhs.z, x*rhs.y - y*rhs.x); }
	
	// Squared and normal length of vector
	inline float	SquareLength() const							{ return (x*x + y*y + z*z); }
	inline float	Length() const									{ return mfSqrt(SquareLength()); }

	// Normalize vector, |v| = 1.0
	inline Vector3&	Normalize()										{ float iLen = 1.0f/Length(); x*=iLen; y*=iLen; z*=iLen; return *this; }
	inline Vector3&	FNormalize()									{ float iLen = mfOneOverSqrt(SquareLength()); x*=iLen; y*=iLen; z*=iLen; return *this; }
	inline Vector3&	FSNormalize(float scale)						{ float iLen = mfOneOverSqrt(SquareLength())/scale; x*=iLen; y*=iLen; z*=iLen; return *this; }
	inline Vector3	Normalized() const								{ float iLen = 1.0f/Length(); return Vector3(x*iLen, y*iLen, z*iLen); }
	inline Vector3	FNormalized() const								{ float iLen = mfOneOverSqrt(SquareLength()); return Vector3(x*iLen, y*iLen, z*iLen); }
	inline bool		IsNormalized() const							{ return (Length() == 1.0f); }

	// Access operators
	inline float&	operator[](int row)								{ assert(row<3); return ((float*)&x)[row]; }
	inline operator float*()										{ return (float *)&x; }
	inline operator const float*() const							{ return (const float *)&x; }

	// Comparison
	inline bool		operator==(cVector3 rhs)						{ return ((x==rhs.x) && (y==rhs.y) && (z==rhs.z)); }
	inline bool		operator!=(cVector3 rhs)						{ return ((x!=rhs.x) || (y!=rhs.y) || (z!=rhs.z)); }
	inline bool		IsClose(cVector3 other, float distsq = 0.9f) const { float xx = (x-other.x); xx*=xx; if (xx > distsq) return false; float yy = (y-other.y); xx+=yy*yy;	if (xx > distsq) return false; float zz = (z-other.z); xx+=zz*zz; return (xx < distsq); }
	inline bool		IsZero(float distsq = 0.9f)						{ Vector3 dimf = *this; return (dimf.Dot(dimf) < distsq); }

	// Vector addition/substraction
	inline Vector3&	Scale(cVector3 rhs)								{ x*=rhs.x; y*=rhs.y; z*=rhs.z; return *this; }
	inline Vector3&	Scale(float factor)								{ x*=factor; y*=factor; z*=factor; return *this; }
	inline Vector3  operator-() const								{ return Vector3(mfNeg(x), mfNeg(y), mfNeg(z)); }
	inline cVector3 operator+=(cVector3 rhs)						{ x+=rhs.x; y+=rhs.y; z+=rhs.z; return *this; }
	inline cVector3 operator-=(cVector3 rhs)						{ x-=rhs.x; y-=rhs.y; z-=rhs.z; return *this; }
	inline friend Vector3 operator+(cVector3 lhs, cVector3 rhs)		{ return Vector3(lhs.x+rhs.x, lhs.y+rhs.y, lhs.z+rhs.z); }
	inline friend Vector3 operator-(cVector3 lhs, cVector3 rhs)		{ return Vector3(lhs.x-rhs.x, lhs.y-rhs.y, lhs.z-rhs.z); }

	// Vector*scalar and vector/scalar
	inline cVector3	operator*=(float m)								{ x*=m; y*=m; z*=m; return *this; }
	inline cVector3	operator/=(float d)								{ float	 iMul=1.0f/d; x*=iMul; y*=iMul; z*=iMul; return *this; }
	inline friend Vector3 operator*(cVector3 lhs, float m)			{ return Vector3(lhs.x*m, lhs.y*m, lhs.z*m); }
	inline friend Vector3 operator*(float m, cVector3 rhs)			{ return Vector3(m*rhs.x, m*rhs.y, m*rhs.z); }
	inline friend Vector3 operator*(cVector3 a, cVector3 b)			{ return Vector3(a.x*b.x, a.y*b.y, a.z*b.z); }
	inline friend Vector3 operator/(cVector3 lhs, float d)			{ float  iMul=1.0f/d; return Vector3(lhs.x*iMul, lhs.y*iMul, lhs.z*iMul); }
	inline friend Vector3 operator/(float d, cVector3 rhs)			{ return Vector3(d/rhs.x, d/rhs.y, d/rhs.z); }

	// Serializing.
	inline friend rArchive operator<<(rArchive arc, cVector3 v)		{ arc << v.x << v.y << v.z; return arc; }
	inline friend rArchive operator>>(rArchive arc, rVector3 v)		{ arc >> v.x >> v.y >> v.z; return arc; }
	Vector3 &GetNormal(Vector3 &a,Vector3 &b, Vector3 &c);

	// Vector3 -> String
	inline String ToString() const									{ String t; t.Format("(x: %8.3f, y: %8.3f, z: %8.3f)", x, y, z); return t; }

	// Data members
	float x, y, z;
};

//
// Vector in 3-space with homogenous-w
//
DECLARE_CLASS(Vector4); 
class  Vector4
{
public:
	// Initialization
	Vector4()														{w=1.0f;}
	Vector4(float x, float y, float z,float w=1.0f)					: x(x), y(y), z(z),w(w) { }
	//Vector4(float v[4])												: x(v[0]), y(v[1]), z(v[2]),w(v[3]) { }
	Vector4(cVector4 other)											: x(other.x), y(other.y), z(other.z),w(other.w) { }
	cVector4 operator=(cVector4 other)								{ x=other.x; y=other.y; z=other.z;w=other.w; return *this; }
	cVector4 operator=(cVector3 other)								{ x=other.x; y=other.y; z=other.z;w=1.0f; return *this; }
	inline void Set(float xx, float yy, float zz,float ww=1.0f)		{ x=xx; y=yy; z=zz;w=ww; }
	
	// vector 3 conversions
	inline Vector3  V3() const										{ return Vector3(x,y,z); }
	inline void  V3(Vector3 &v) const								{ v.x=x,v.y=y,v.z=z;}

	// Dot product (.): cos phi = (v.w)/(|v||w|)
	//					cos phi =  v.w              when |v|=|w|=1
	inline float	Dot(cVector4 rhs) const							{ return (x*rhs.x + y*rhs.y + z*rhs.z+w*rhs.w); }

	// Cross product (x)
	//inline Vector4	Cross(cVector4 rhs) const						{ return Vector3(y*rhs.z - z*rhs.y, z*rhs.x - x*rhs.z, x*rhs.y - y*rhs.x); }

	// Squared and normal length of vector
	inline float	SquareLength() const							{ return (x*x + y*y + z*z); }
	inline float	Length() const									{ return mfSqrt(SquareLength()); }

	// Normalize vector, |v| = 1.0
	inline Vector4&	Normalize()										{ float iLen = 1.0f/Length(); x*=iLen; y*=iLen; z*=iLen; return *this; }
	inline Vector4&	FNormalize()									{ float iLen = mfOneOverSqrt(SquareLength()); x*=iLen; y*=iLen; z*=iLen; return *this; }
	//inline Vector4	Normalized() const								{ float iLen = 1.0f/Length(); return Vector3(x*iLen, y*iLen, z*iLen); }
	//inline Vector4	FNormalized() const								{ float iLen = mfOneOverSqrt(SquareLength()); return Vector3(x*iLen, y*iLen, z*iLen); }
	inline bool		IsNormalized() const							{ return (Length() == 1.0f); }
	
	// homogenous specific 
	inline Vector4 &DeHomogenize()									{ float d=1.0f/w;x*=d,y*=d,z*=d;return (*this);}
	inline char		HomogenousClip()								{ if (x<=-w)return(1);if (x>  w)return(2);if (y<=-w)return(3);if (y>  w)return(4);if (z<=-w)return(5);if (z>  w)return(6);return(0);	}
	inline Vector2  ToViewport2(Vector2 &Scale,Vector2 &Centre)		{ if (w!=0){return Vector2((x/w)*Scale.x+Centre.x,(y/w)*Scale.y+Centre.y);};return Vector2(0,0);}
	inline Vector3  ToViewport3(Vector2 &Scale,Vector2 &Centre)		{ float oow=1.0f/w;return Vector3((x*oow)*Scale.x+Centre.x,(y*oow)*Scale.y+Centre.y,z*oow);}

	// Access operators
	inline float&	operator[](int row)								{ assert(row<3); return ((float*)&x)[row]; }
	inline operator float*()										{ return (float *)&x; }
	inline operator const float*() const							{ return (const float *)&x; }

	// Comparison
	inline bool		operator==(cVector4 rhs)						{ return ((x==rhs.x) && (y==rhs.y) && (z==rhs.z) && (w==rhs.w)); }
	inline bool		operator!=(cVector4 rhs)						{ return ((x!=rhs.x) || (y!=rhs.y) || (z!=rhs.z) || (z!=rhs.w)); }
	
	// Vector addition/substraction
	inline Vector4&	Scale(cVector4 rhs)								{ x*=rhs.x; y*=rhs.y; z*=rhs.z; return *this; }
	inline Vector4&	Scale(float factor)								{ x*=factor; y*=factor; z*=factor; return *this; }
	inline Vector4  operator-() const								{ return Vector4(mfNeg(x), mfNeg(y), mfNeg(z)); }
	inline cVector4 operator+=(cVector4 rhs)						{ x+=rhs.x; y+=rhs.y; z+=rhs.z; return *this; }
	inline cVector4 operator-=(cVector4 rhs)						{ x-=rhs.x; y-=rhs.y; z-=rhs.z; return *this; }
	inline friend Vector4 operator+(cVector4 lhs, cVector4 rhs)		{ return Vector4(lhs.x+rhs.x, lhs.y+rhs.y, lhs.z+rhs.z); }
	inline friend Vector4 operator-(cVector4 lhs, cVector4 rhs)		{ return Vector4(lhs.x-rhs.x, lhs.y-rhs.y, lhs.z-rhs.z); }

	// Vector*scalar and vector/scalar
	inline cVector4	operator*=(float m)								{ x*=m; y*=m; z*=m; return *this; }
	inline cVector4	operator/=(float d)								{ float	 iMul=1.0f/d; x*=iMul; y*=iMul; z*=iMul; return *this; }
	inline friend Vector4 operator*(cVector4 lhs, float m)			{ return Vector4(lhs.x*m, lhs.y*m, lhs.z*m); }
	inline friend Vector4 operator*(float m, cVector4 rhs)			{ return Vector4(m*rhs.x, m*rhs.y, m*rhs.z); }
	inline friend Vector4 operator*(cVector4 a, cVector4 b)			{ return Vector4(a.x*b.x, a.y*b.y, a.z*b.z); }
	inline friend Vector4 operator/(cVector4 lhs, float d)			{ float  iMul=1.0f/d; return Vector4(lhs.x*iMul, lhs.y*iMul, lhs.z*iMul); }
	inline friend Vector4 operator/(float d, cVector4 rhs)			{ return Vector4(d/rhs.x, d/rhs.y, d/rhs.z); }

	// Serializing.
	inline friend rArchive operator<<(rArchive arc, cVector4 v)		{ arc << v.x << v.y << v.z << v.w; return arc; }
	inline friend rArchive operator>>(rArchive arc, rVector4 v)		{ arc >> v.x >> v.y >> v.z >> v.w; return arc; }

	// Data members
	float x, y, z, w;
};

//
// Plane equation for infinite plane in 3-space
//
DECLARE_CLASS(PlaneEq);
class  PlaneEq
{
public:
	// Initialization
	PlaneEq()														: Normal(0.0, 1.0, 0.0), Constant(0.0) { }
	PlaneEq(float a, float b, float c, float d)						: Normal(a, b, c), Constant(d) { }
	PlaneEq(cVector3 n, float c)									: Normal(n), Constant(c) { }
	cPlaneEq operator=(cPlaneEq other)								{ Normal=other.Normal; Constant=other.Constant; return *this; }

	// Orientation and intersections of plane with other vectors, etc.
	inline int		Orientation(cVector3 v, float delta=0.02) const	{ return mfSign(Normal.Dot(v) + Constant, delta); }

	//
	// Finds the fraction for interpolation. Also very convenient for calculating
	// intersections with planes for other purposes than interpolation.
	//

	inline float InterpolateFraction(cVector3 p, cVector3 dir) const
	{
		float c   = -Normal.Dot(p) - Constant;
		float den = Normal.Dot(dir);
		
		if (mfAbs(c-den) < 0.0000001) return 1.0;								// Nearly x/x -> 1.0

		assert(den != 0.0);

		return c/den;
	}
	// Finds the fraction for interpolation. Also very convenient for calculating
	// intersections with planes for other purposes than interpolation.
	inline float InterpolateFractionOmfset(cVector3 p, cVector3 dir, float omfs) const
	{
		float c   = -Normal.Dot(p+Normal*omfs) - Constant;
		float den = Normal.Dot(dir);
		
		if (mfAbs(c-den) < 0.0000001) return 1.0;							// Nearly x/x -> 1.0

		assert(den != 0.0);

		return c/den;
	}

	//
	// Return a point on the line p1-p2 that intersects the plane
	//
	inline Vector3 Intersect(cVector3 p1, cVector3 p2) const
	{
		Vector3 dir = p2 - p1;
		return p1 + dir*InterpolateFraction(p1, dir);
	}

	//
	// Return a point on the line p1-p2 that intersects the plane
	//
	inline Vector3 IntersectOmfset(cVector3 p1, cVector3 p2, float omfs) const
	{
		Vector3 dir = p2 - p1;
		return p1 + dir*InterpolateFractionOmfset(p1, dir, omfs);
	}


	inline int		OrientationOmfset(cVector3 v, float omfs) const	{ return mfSign(Normal.Dot(v-Normal*omfs) + Constant, 0.02f); }
	inline Vector3	Project(cVector3 p) const						{ return p - Normal * (Normal.Dot(p)+Constant); }



	// Serializing.
	inline friend	rArchive operator<<(rArchive arc, cPlaneEq p)	{ arc << p.Normal << p.Constant; return arc; }
	inline friend	rArchive operator>>(rArchive arc, rPlaneEq p)	{ arc >> p.Normal >> p.Constant; return arc; }

	// Data members
	Vector3	Normal;
	float	Constant;
};



//
// Bounding box in 2-space
//
DECLARE_CLASS(BoundingBox2);
class  BoundingBox2
{
public:
	// Construction/Initialization
	BoundingBox2()													: Min(FLT_MAX, FLT_MAX), Max(-FLT_MAX, -FLT_MAX) { }
	BoundingBox2(cBoundingBox2 other)								: Min(other.Min), Max(other.Max) { }
	BoundingBox2(cVector2 min, cVector2 max)						: Min(min), Max(max) { }
	BoundingBox2(float a,float b,float c,float d)					: Min(a, b), Max(c, d) { }
	cBoundingBox2 operator=(cBoundingBox2 other)					{ Min=other.Min; Max=other.Max; return *this; }
	inline void		Init(cVector2 center, float radius)				{ Min.Set(center.u-radius, center.v-radius); Max.Set(center.u+radius, center.v+radius); }
	inline void		Init(cVector2 min=Vector2(FLT_MAX, FLT_MAX), cVector2 max=Vector2(-FLT_MAX, -FLT_MAX))	{ Min=min; Max=max; }
	inline void		Init(float a,float b,float c,float d)			{Min=Vector2(a,b);Max=Vector2(c,d); }

	// Update the bounding box to totally include a vector

	inline void		Encapsulate(cVector2 v)							{ Min.u = min(Min.u, v.u); Min.v = min(Min.v, v.v); Max.u = max(Max.u, v.u); Max.v = max(Max.v, v.v); }
	inline void		Encapsulate(cBoundingBox2 other)				{ Encapsulate(other.Min); Encapsulate(other.Max); }
	inline void		Widen(float delta = DELTA)						{ Min.u-=delta; Min.v-=delta; Max.u+=delta; Max.v+=delta; }
	inline void		Translate(cVector2 v)							{ Min+=v; Max+=v; }
	
	inline Vector2  Centre() const									{ return Vector2((Min.x+Max.x)/2.0f,(Min.y+Max.y)/2.0f); }
	inline Vector2  Size() const 									{ return Vector2((Max.x-Min.x),(Max.y-Min.y)); }
	inline BoundingBox2 Transform_st (cVector2 s,cVector2 t) const	{ return BoundingBox2(Vector2((Min.x+t.x)*s.x,(Min.y+t.y)*s.y),Vector2((Max.x+t.x)*s.x,(Max.y+t.y)*s.y) ); }
	inline BoundingBox2 Transform_stt(cVector2 s,cVector2 t,cVector2 t2)const
																	{ return BoundingBox2(Vector2((Min.x+t.x)*s.x+t2.x*s.x,(Min.y+t.y)*s.y+t2.y*s.y),Vector2((Max.x+t.x)*s.x+t2.x*s.x,(Max.y+t.y)*s.y+t2.y*s.y) ); }
	inline void     ClipVector(Vector2 *s)							{ s->x=clamp_f(s->x,Min.x,Max.x);s->y=clamp_f(s->y,Min.y,Max.y); }
	inline BoundingBox2 ClipBox(BoundingBox2 &src)					{ Vector2 min=src.Min,max=src.Max;ClipVector(&min);ClipVector(&max);return BoundingBox2(min,max);}
	// Check if a bounding box contains a vector or another boundingbox
	inline bool		Contains(cVector2 v) const						{ return (inrange(v.u, Min.u, Max.u) && inrange(v.v, Min.v, Max.v)); }
	inline bool		Contains(cBoundingBox2 b) const					{ return (Contains(b.Min) && Contains(b.Max)); }
	inline bool		Intersects(cBoundingBox2 other) const			{ return !(Min.u > other.Max.u || Max.u < other.Min.u || Min.v > other.Max.v || Max.v < other.Min.v); }
	inline bool		IsVisible(cBoundingBox2 i/*innerbox*/) const	{return !( (i.Min.x<Min.x && i.Max.x<Min.x) || (i.Min.x>Max.x && i.Max.x>Max.x) || (i.Min.y<Min.y && i.Max.y<Min.y) || (i.Min.y>Max.y && i.Max.y>Max.y) );}
	inline BoundingBox2 SnapToGrid(Vector2 g)						{ return BoundingBox2(fix(Min.x,g.x),fix(Min.y,g.y),fix(Max.x,g.x),fix(Max.y,g.y));}
	inline String	ToString() const								{ String t; t.Format("%s-%s", (const char *)Min.ToString(), (const char *)Max.ToString()); return t; }
	

	// arithmetic
	inline cBoundingBox2	    operator+=(cVector2 m)				{ Min+=m,Max+=m; return *this; }
	inline cBoundingBox2	    operator-=(cVector2 m)				{ Min-=m,Max-=m; return *this; }
	inline friend BoundingBox2 operator+(cBoundingBox2 b, cVector2 d){ return BoundingBox2(b.Min+d,b.Max+d); }
	//inline friend BoundingBox2 operator+(cBoundingBox2 b, cVector2 d){ return b+d; }
	inline friend BoundingBox2 operator-(cBoundingBox2 b, cVector2 d){ return BoundingBox2(b.Min-d,b.Max+d); }
	//inline friend BoundingBox2 operator-(cBoundingBox2 b, cVector2 d){ return b-d; }
	inline cBoundingBox2	    operator*=(cVector2 m)				{ Min*=m,Max*=m; return *this; }
	inline cBoundingBox2	    operator/=(cVector2 m)				{ Min/=m,Max/=m; return *this; }
	inline cBoundingBox2	    operator*=(cBoundingBox2 m)			{ Min*=m.Min,Max*=m.Max; return *this; }
	inline cBoundingBox2	    operator/=(cBoundingBox2 m)			{ Min/=m.Min,Max/=m.Max; return *this; }
	inline friend BoundingBox2 operator*(cBoundingBox2 b, cVector2 d){ return BoundingBox2(b.Min*d,b.Max*d); }
	inline friend BoundingBox2 operator*(cVector2 d, cBoundingBox2 b){ return b*d; }
	inline friend BoundingBox2 operator/(cBoundingBox2 b, cVector2 d){ return BoundingBox2(b.Min/d,b.Max/d); }
	inline friend BoundingBox2 operator/(cVector2 d, cBoundingBox2 b){ return b/d; }

	// conversions
	//void			ConvertToMFCRect(void *c);			
	//void			ConvertFromMFCRect(void *c);

	// Serializing.
	inline friend	rArchive operator<<(rArchive arc, cBoundingBox2 b){ arc << b.Min << b.Max; return arc; }
	inline friend	rArchive operator>>(rArchive arc, rBoundingBox2 b){ arc >> b.Min >> b.Max; return arc; }

	// Data members
	Vector2 Min, Max;
};



//
// Bounding box in 3-space
//
DECLARE_CLASS(BoundingBox3);
class  BoundingBox3
{
public:
	// Construction/Initialization
	BoundingBox3()													: Min(FLT_MAX, FLT_MAX, FLT_MAX), Max(-FLT_MAX, -FLT_MAX, -FLT_MAX) { }
	BoundingBox3(cBoundingBox3 other)								: Min(other.Min), Max(other.Max) { }
	BoundingBox3(cVector3 min, cVector3 max)						: Min(min), Max(max) { }
	cBoundingBox3 operator=(cBoundingBox3 other)					{ Min=other.Min; Max=other.Max; return *this; }
	inline void		Init(cVector3 center, float radius)				{ Min.Set(center.x-radius, center.y-radius, center.z-radius); Max.Set(center.x+radius, center.y+radius, center.z+radius); }
	inline void		Init(cVector3 min = Vector3(FLT_MAX, FLT_MAX, FLT_MAX), cVector3 max = Vector3(-FLT_MAX, -FLT_MAX, -FLT_MAX))				{ Min=min; Max=max; }

	// Update the bounding box to totally include a vector
	inline void		Encapsulate(cVector3 v)							{ Min.x = min(Min.x, v.x); Min.y = min(Min.y, v.y); Min.z = min(Min.z, v.z); Max.x = max(Max.x, v.x); Max.y = max(Max.y, v.y); Max.z = max(Max.z, v.z); }
	inline void		Encapsulate(cBoundingBox3 other)				{ Encapsulate(other.Min); Encapsulate(other.Max); }
	inline void		Widen(float delta = DELTA)						{ Min.x-=delta; Min.y-=delta; Min.z-=delta; Max.x+=delta; Max.y+=delta; Max.z+=delta; }
	inline void		Translate(cVector3 v)							{ Min+=v; Max+=v; }
	inline void		Maximize()										{ Min = Vector3(-FLT_MAX, -FLT_MAX, -FLT_MAX);	Max = Vector3(FLT_MAX, FLT_MAX, FLT_MAX); }
	inline void		MakeBox(cVector3 Pos,float delta)				{ Min = Pos-Vector3(delta, delta, delta);		Max = Pos+Vector3(delta, delta, delta); }

	// Check if a bounding box contains a vector or another boundingbox
	inline bool		Contains(cVector3 v) const						{ return (inrange(v.x, Min.x, Max.x) && inrange(v.z, Min.z, Max.z) && inrange(v.y, Min.y, Max.y)); }
	inline bool		Contains(cVector4 v) const						{ return (inrange(v.x, Min.x, Max.x) && inrange(v.z, Min.z, Max.z) && inrange(v.y, Min.y, Max.y)); }
	inline bool		Contains(cBoundingBox3 b) const					{ return (Contains(b.Min) && Contains(b.Max)); }
	inline bool		Intersects(cBoundingBox3 other) const			{ return !(Min.x > other.Max.x || Max.x < other.Min.x || Min.y > other.Max.y || Max.y < other.Min.y || Min.z > other.Max.z || Max.z < other.Min.z); }
	bool			Intersects(cVector3 origin, cVector3 direction, float *tnear = NULL, float *tfar =NULL) const;

	inline String	ToString() const								{ String t; t.Format("%s-%s", Min.ToString().GetReadPtr(), Max.ToString().GetReadPtr()); return t; }

	// Serializing.
	inline friend	rArchive operator<<(rArchive arc, cBoundingBox3 b){ arc << b.Min << b.Max; return arc; }
	inline friend	rArchive operator>>(rArchive arc, rBoundingBox3 b){ arc >> b.Min >> b.Max; return arc; }

	// Data members
	Vector3 Min, Max;
};


//
// Bounding sphere
//
DECLARE_CLASS(BoundingSphere);
class  BoundingSphere
{
public:
	BoundingSphere()	{}
	BoundingSphere(cVector3 pos,float radius)	: Pos(pos), Radius(radius), RadiusSq(radius*radius)	{ }
	BoundingSphere(cVector3 pos,float radius,float radiussq)	: Pos(pos), Radius(radius), RadiusSq(radiussq)	{ }

	cBoundingSphere operator=(cBoundingSphere other)					{ Pos=other.Pos; Radius=other.Radius; return *this; }

//	inline void		Encapsulate(cVector3 v)		{ Vector3 dimf = Pos - v; float dist=dimf.Dot(dimf); if (dist>RadiusSq) {RadiusSq=dist; Radius=mfFSqrt(dist);}}
	inline bool		Contains(cVector3 v)	{	Vector3 dimf = Pos - v; float dist=dimf.Dot(dimf); return (dist<=RadiusSq);	}

	inline bool		Contains(cBoundingBox3 b);

	inline String	ToString() const								{ String t; t.Format("%s-%f", (const char *)Pos.ToString(), Radius); return t; }

	// Serializing.
	inline friend	rArchive operator<<(rArchive arc, cBoundingSphere b){ arc << b.Pos << b.Radius; return arc; }
	inline friend	rArchive operator>>(rArchive arc, rBoundingSphere b){ arc >> b.Pos >> b.Radius; b.RadiusSq=b.Radius*b.Radius; return arc; }

	Vector3	Pos;
	float Radius;
	float RadiusSq; // squared radius... for checking against squared distances
};


// TODO.. redesign this matrix class to do some usefull stuff..
//
// 4x4 Matrix, stored in a b c d	order						
//						 e f g h
//						 i j k l 
//						 m n o p 
//
DECLARE_CLASS(Matrix4x4);
class  Matrix4x4
{
public:
	// Initialization
	Matrix4x4()														{ }
	Matrix4x4(cMatrix4x4 other)										{ memcpy(val, other.val, 4*4*sizeof(float)); }
	Matrix4x4(float *m)												{ memcpy(val, &m[0], 4*4*sizeof(float)); }
	inline Matrix4x4(cVector3 v1, cVector3 v2, cVector3 v3)			{ val[0] = v1.x;	val[4] = v2.x;	val[8]  = v3.x;	val[12] = 0.0; val[1] = v1.y;	val[5] = v2.y;	val[9]  = v3.y;	val[13] = 0.0; val[2] = v1.z;	val[6] = v2.z;	val[10] = v3.z;	val[14] = 0.0; val[3] = 0.0f;	val[7] = 0.0f;	val[11] = 0.0f;	val[15] = 1.0f; }
	inline Matrix4x4(cVector3 v1, cVector3 v2, cVector3 v3, cVector3 v4) { val[0] = v1.x;	val[4] = v2.x;	val[8]  = v3.x;	val[12] = v4.x; val[1] = v1.y;	val[5] = v2.y;	val[9]  = v3.y;	val[13] = v4.y; val[2] = v1.z;	val[6] = v2.z;	val[10] = v3.z;	val[14] = v4.z; val[3] = 0.0f;	val[7] = 0.0f;	val[11] = 0.0f;	val[15] = 1.0f; }
	inline cMatrix4x4 operator=(cMatrix4x4 other)					{ memcpy(val, other.val, 4*4*sizeof(float)); return *this; }

	// Standard matrices
	inline void			Copy(void *target)							{ memcpy(target,val,4*4*sizeof(float));}
	inline Matrix4x4&	Zero()										{ memset(val, 0, 4*4*sizeof(float)); return *this; }
	inline Matrix4x4&	Identity()									{ memcpy(val, MatIdentity,4*4*sizeof(float)); return *this; }

	// Properties
	inline bool			IsIdentity() const							{ for (int x=0; x<4; x++) for (int y=0; y<4; y++) if (val[x*4+y] != ((x==y) ? 1.0f : 0.0f)) return false; return true; }

	// Data access
	inline float&		Elem(int row, int col)						{ bounds(row, 0, 3); bounds(col, 0, 3); return val[col*4+row]; }
	inline float		ConstElem(int row, int col) const			{ bounds(row, 0, 3); bounds(col, 0, 3); return val[col*4+row]; }
	inline void			SetElem(int row, int col, float val)		{ Elem(row, col) = val; }
	inline float		GetElem(int row, int col) const				{ return ConstElem(row, col); }
	inline Vector3		GetColumn(int col) const					{ bounds(col, 0, 3); return Vector3(val[col+0], val[col+4], val[col+8]); }
	inline void			SetColumn(int col, cVector3 v)				{ bounds(col, 0, 3); val[col+0]=v.x; val[col+4]=v.y; val[col+8]=v.z; }
	inline float&		operator[](int col)				 			{ assert(col<4); return val[col]; }

	// Initialization functions
	void				SetRT(Vector3 *r,Vector3 *t);
	void				SetRT(Vector3 &r,Vector3 &t){SetRT(&r,&t);}
	void				SetTR(Vector3 *t,Vector3 *r);
	void				SetTR(Vector3 &t,Vector3 &r){SetTR(&r,&t);}
	void				SetT(Vector3 *r);
	void				SetT(Vector3 &r){SetT(&r);}
	void				SetR(Vector3 *t);
	void				SetR(Vector3 &t){SetR(&t);}
	void				SetS(Vector3 *s);
	void				SetS(Vector3 &s){SetS(&s);}
	void				SetRZXY(Vector3 *r);
	void				SetRZXY(Vector3 &r){SetRZXY(&r);}
	void				SetRTZXY(Vector3 *r,Vector3 *t);
	void				SetRTZXY(Vector3 &r,Vector3 &t){SetRTZXY(&r,&t);}
	void				SetSRTZXY(Vector3 *s,Vector3 *r,Vector3 *t);
	void				SetSRTZXY(Vector3 &s,Vector3 &r,Vector3 &t){SetSRTZXY(&s,&r,&t);}
	void				SetProjectFOV(Vector2 &fov,float znear,float zfar=10000000);
	void				SetProjectVPR(Vector2 &vpr,float znear,float zfar=10000000);
	void				SetClip   (Vector2 &Top,Vector2 &Width,Vector2 &Depth);

	// Simple mathematical operators
	inline cMatrix4x4	operator+=(cMatrix4x4 rhs)					{ for (int i=0; i<4*4; i++) val[i]+=rhs.val[i]; return *this; }
	inline cMatrix4x4	operator-=(cMatrix4x4 rhs)					{ for (int i=0; i<4*4; i++) val[i]-=rhs.val[i]; return *this; }
	inline friend Matrix4x4 operator+(cMatrix4x4 lhs, cMatrix4x4 rhs) { Matrix4x4 n; for (int i=0; i<4*4; i++) n.val[i] = lhs.val[i]+rhs.val[i]; return n; }
	inline friend Matrix4x4 operator-(cMatrix4x4 lhs, cMatrix4x4 rhs) { Matrix4x4 n; for (int i=0; i<4*4; i++) n.val[i] = lhs.val[i]-rhs.val[i]; return n; }

	// Expensive mathematical operators
	inline cMatrix4x4	operator*=(float mul)						{for (long j=0;j<16;j+=1)val[j]*=mul;return *this; }
	inline cMatrix4x4	operator*=(cMatrix4x4 rhs)					{float	t[4*4];const float *s=rhs.val;for (long j=0;j<16;j+=4)t[j  ]=val[j  ]*s[0+0]+val[j+1]*s[4+0]+val[j+2]*s[8+0]+val[j+2]*s[12+0],t[j+1]=val[j+1]*s[0+1]+val[j+1]*s[4+1]+val[j+2]*s[8+1]+val[j+2]*s[12+1],t[j+2]=val[j+2]*s[0+2]+val[j+1]*s[4+2]+val[j+2]*s[8+2]+val[j+2]*s[12+2],t[j+3]=val[j+3]*s[0+3]+val[j+1]*s[4+3]+val[j+2]*s[8+3]+val[j+2]*s[12+3];memcpy(val, t,4*4*sizeof(float)); return *this;}		
	inline Vector3		operator*(cVector3 rhs) const				{return Vector3(val[0+0]*rhs.x+val[0+1]*rhs.y+val[0+2]*rhs.z+val[0+3],val[4+0]*rhs.x+val[4+1]*rhs.y+val[4+2]*rhs.z+val[4+3],val[8+0]*rhs.x+val[8+1]*rhs.y+val[8+2]*rhs.z+val[8+3]);}
	inline Vector4		operator*(cVector4 rhs) const				{return Vector4(val[0+0] *rhs.x+val[0+1] *rhs.y+val[0+2] *rhs.z+val[0+3] *rhs.w,val[4+0] *rhs.x+val[4+1] *rhs.y+val[4+2] *rhs.z+val[4+3] *rhs.w,val[8+0] *rhs.x+val[8+1] *rhs.y+val[8+2] *rhs.z+val[8+3] *rhs.w,val[12+0]*rhs.x+val[12+1]*rhs.y+val[12+2]*rhs.z+val[12+3]*rhs.w);}
	inline Vector4		MulV4(Vector3 rhs){return Vector4(val[0+0] *rhs.x+val[0+1] *rhs.y+val[0+2] *rhs.z+val[0+3],val[4+0] *rhs.x+val[4+1] *rhs.y+val[4+2] *rhs.z+val[4+3],val[8+0] *rhs.x+val[8+1] *rhs.y+val[8+2] *rhs.z+val[8+3],val[12+0]*rhs.x+val[12+1]*rhs.y+val[12+2]*rhs.z+val[12+3]);}
	inline Vector3		MulMinor(cVector3 rhs) const				{return Vector3(val[0+0]*rhs.x+val[0+1]*rhs.y+val[0+2]*rhs.z,val[4+0]*rhs.x+val[4+1]*rhs.y+val[4+2]*rhs.z,val[8+0]*rhs.x+val[8+1]*rhs.y+val[8+2]*rhs.z);}
	inline Matrix4x4	operator*(float mul)						{float t[4*4];for (long j=0;j<16;j+=1)t[j]=val[j]*mul;return(Matrix4x4(t));}
	inline Matrix4x4	operator*(cMatrix4x4 rhs) const				{float	t[4*4];const float *s=rhs.val;for (long j=0;j<16;j+=4)t[j+0]=val[j+0]*s[0+0]+val[j+1]*s[4+0]+val[j+2]*s[8+0]+val[j+3]*s[12+0],t[j+1]=val[j+0]*s[0+1]+val[j+1]*s[4+1]+val[j+2]*s[8+1]+val[j+3]*s[12+1],t[j+2]=val[j+0]*s[0+2]+val[j+1]*s[4+2]+val[j+2]*s[8+2]+val[j+3]*s[12+2],t[j+3]=val[j+0]*s[0+3]+val[j+1]*s[4+3]+val[j+2]*s[8+3]+val[j+3]*s[12+3];return(Matrix4x4(t));}		
	inline friend	rArchive operator<<(rArchive arc, rMatrix4x4 b){ return arc;}
	inline friend	rArchive operator>>(rArchive arc, rMatrix4x4 b){ return arc;}	
	int operator==(cMatrix4x4 rhs) const			{return (memcmp(val, rhs.val, 4*4*sizeof(float))==0); }

	// Matrix operations & derived matrices
	Matrix4x4			Transpose();
	Matrix4x4			TransposeMinor();					
	Matrix4x4			Inverse();
	Matrix4x4		   &Normalize();
	// 3D helper function
	Vector2				TransformAndProject(rVector3 src,rVector2 Aspect,rVector2 centre);

	// Casting/conversion
	inline operator		float*()									{ return (float *)val; }
	inline operator		const float*() const						{ return (float *)val; }

	// Matrix4x4 -> String
	inline String ToString() const									{ String t; t.Format("%8.3f %8.3f %8.3f %8.3f\n%8.3f %8.3f %8.3f %8.3f\n%8.3f %8.3f %8.3f %8.3f\n%8.3f %8.3f %8.3f %8.3f",val[0],val[1],val[2],val[3],val[4],val[5],val[6],val[7],val[8],val[9],val[10],val[11],val[12],val[13],val[14],val[15]);return(t);}

	// Matrix data
	float				val[4*4];									// Matrix values index = row*4 + col
	static float		MatIdentity[4*4];							// Identity matrix

	//inline friend	rArchive operator<<(rArchive arc, rMatrix4x4 b){ arc.Write(b.val,sizeof(float)*4*4); return arc; }
	//inline friend	rArchive operator>>(rArchive arc, rMatrix4x4 b){ arc.Read(b.val,sizeof(float)*4*4); return arc; }

};


DECLARE_CLASS(Orient3);  // 3D position/orientation holder, using Yaw/Tilt/Roll system
class  Orient3
{
	public:
	// Initialization
	Orient3()														{ }
	Orient3(float Tilt,float Yaw,float Roll)						{ rx=Tilt,ry=Yaw,rz=Roll;}
	cOrient3 operator=(cOrient3 other)								{ rx=other.rx,ry=other.ry,rz=other.rz; return (*this);}
	Orient3(Vector3 r)												{ rx=r.x,ry=r.y,rz=r.z;}
	inline Orient3&	Reset()										    { rx=ry=rz=0;return (*this);}

	// transformations
	inline Orient3&	RotateDelta(float Tilt=0,float Yaw=0,float Roll=0){rx+=Tilt,ry+=Yaw,rz+=Roll;return (*this);}  // rotates position by dx,dy,dr

	inline Orient3&	Set(rVector3 r)								   	  {rx=r.x,ry=r.y,rz=r.z;return (*this);}// Sets rotation to r
	inline Orient3&	Set(float a,float b,float c)				   	  {rx=a,ry=b,rz=c;return (*this);}// Sets rotation to r
	inline Orient3&	Set(rOrient3 r)								   	  {(*this)=r;return (*this);}// Sets rotation to r
	
	inline Vector3  Vector()										  {return Vector3(rx,ry,rz);}
	Orient3&			FromVector(rVector3 tgt,float up);           // gets rotation from targetting vector

	Matrix4x4			GetMatrix(Vector3 &pos);					// gets matrix for transform
	Matrix4x4			GetRotMatrix();								// gets matrix for rotation
	
	inline bool			HasEffect()									{return ( rx!=0 || ry!=0 || rz!=0 ); }
	// Data
	float				rx,ry,rz;

	// archiving
	inline friend	rArchive operator<<(rArchive arc, rOrient3 b){ arc<<b.rx<<b.ry<<b.rz; return arc;}
	inline friend	rArchive operator>>(rArchive arc, rOrient3 b){ arc>>b.rx>>b.ry>>b.rz; return arc;}
};


DECLARE_CLASS(Quaternion);  // 3D position/orientation holder, using Yaw/Tilt/Roll system
class  Quaternion
{
  private:
    float W, X, Y, Z;      // components of a quaternion

  public:
    // default constructor
    Quaternion();
    Quaternion(float,  float = 0.0,  float = 0.0,  float = 0.0);
	Quaternion(Vector4 &vec);
	
    // quaternion multiplication
    friend Quaternion operator * (Quaternion& a, Quaternion&b);
    Quaternion& operator *= (Quaternion& o);

    // konjugiertes Quaternion
    Quaternion& operator ~ ();

    // this function inverts the quaternion
    Quaternion& operator - ();

    // this normalizes a quaternion 
    Quaternion& Normalize();
    Quaternion& FromAxis(float rot,Vector3 &axis);
    void ToMatrix(Matrix4x4 &mat);

    void Slerp(Quaternion &a,Quaternion &b, float fac);

    // some additional Quaternion functions
    // getting the exponent of a quaternion
    Quaternion& exp();
    Quaternion& log();
};
#endif
