/*
---------------------------------------------------------------------------
|					Function drawing mode for VFCalc
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/
#include "Library.h"
#include "vfcalcglobals.h" 

void CalcMode_ft::Draw()
{
	CVfcalcDlg::functions &func=vc->func;
	CVfcalcDlg::variables &var =vc->var;	
	CVfcalcDlg::switches  &sw  =vc->sw;
	CVfcalcDlg::colors	  &col =vc->col;
	CVfcalcDlg::DrawData  &dd  =vc->dd;

	double	x,y,u,v,w,r,ou,ov,ow;
	double	step=((dd.w_end-dd.w_start)*var.stepsize)/dd.ftow_xscale;
	long	ptcount;

	if (sw.nor_a==true)
	{
		step=(var.dt2-var.dt1)*var.stepsize;
		CPoint *pointlst=new CPoint[(long)((var.dt2-var.dt1)/step)+10];

		// Draw the normal f(t)(xy)
		ptcount=0;
		for (u=var.dt1;u<=var.dt2;u+=step)
		{
			x=(dd.ftow_xscale*vc->clipf(func.fa.Calculate('t',u,'x',u))+dd.ftow_xoffset);
			y=(dd.ftow_yscale*vc->clipf(func.fb.Calculate('t',u,'x',y))+dd.ftow_yoffset);
			if (y<dd.w_min)y=dd.w_min-1;if (y>dd.w_max)y=dd.w_max+1;
			if (x<dd.w_start)x=dd.w_start-1;if (x>dd.w_end)x=dd.w_end+1;
			pointlst[ptcount].x=(long)x,pointlst[ptcount].y=(long)y;

			ptcount++;
		}
		dd.dc->SelectObject(&dd.p_line1);	dd.dc->Polyline(pointlst,ptcount);
		delete pointlst;
	}
	if (sw.der_a==true)
	{
		// Draw the normal f(t)(xy) deriviate
		step=(var.dt2-var.dt1)*var.stepsize;
		long size=(long)((var.dt2-var.dt1)/step)+10;
		CPoint *pointlst=new CPoint[size];ptcount=0;
		bool first=true;
		for (u=var.dt1;u<=var.dt2;u+=step)
		{
			v=func.fb.Calculate('t',u,'x',u);r=func.fa.Calculate('t',u,'x',u);		
			if (u==var.dt1)
			{
				ov=v;ou=r;
				v=func.fb.Calculate('t',u,'x',u);r=func.fa.Calculate('t',u,'x',u);		
			}
			else
			{
				x=(dd.ftow_xscale*u+dd.ftow_xoffset);
				y=(dd.ftow_yscale*((v-ov)/(r-ou))+dd.ftow_yoffset);
				if (y<dd.w_min)y=dd.w_min-1;if (y>dd.w_max)y=dd.w_max+1;
				if (x<dd.w_start)x=dd.w_start-1;if (x>dd.w_end)x=dd.w_end+1;
				if (first==true && u!=var.dt1)pointlst[ptcount].x=(long)x,pointlst[ptcount].y=(long)y,first=false,ptcount++;
				pointlst[ptcount].x=(long)x,pointlst[ptcount].y=(long)y;ptcount++;
				ov=v;ou=r;
			}

		}
		dd.dc->SelectObject(&dd.p_line1);	dd.dc->Polyline(pointlst,ptcount);
		delete pointlst;

	}
	if (sw.der2_a==true)
	{
		// Draw the double f(t)(xy) deriviate
		step=(var.dt2-var.dt1)*var.stepsize;
		CPoint *pointlst=new CPoint[(long)((var.dt2-var.dt1)/step)+10];ptcount=0;
		bool first=true;
		for (u=var.dt1;u<=var.dt2;u+=step)
		{
			v=func.fb.Calculate('x',u,'t',u);r=func.fa.Calculate('x',u,'t',u);
			if (u<=var.dt1+step)
			{
				ow=((v-ov)/(r-ou));
			}
			else
			{
				w=((v-ov)/(r-ou));
				x=(dd.ftow_xscale*u+dd.ftow_xoffset);
				y=(dd.ftow_yscale*(w-ow)/step+dd.ftow_yoffset);
				if (y<dd.w_min)y=dd.w_min-1;if (y>dd.w_max)y=dd.w_max+1;
				if (x<dd.w_start)x=dd.w_start-1;if (x>dd.w_end)x=dd.w_end+1;
				if (first==true && u!=var.dt1)pointlst[ptcount].x=(long)x,pointlst[ptcount].y=(long)y,first=false,ptcount++;
				pointlst[ptcount].x=(long)x,pointlst[ptcount].y=(long)y;ptcount++;
				ow=w;
			}
			ov=v;ou=r;
		}
		dd.dc->SelectObject(&dd.p_line1);	dd.dc->Polyline(pointlst,ptcount);
		delete pointlst;
	}
	if (sw.len_a==true)
	{
		step=(var.dt2-var.dt1)*var.stepsize;
		CPoint *pointlst=new CPoint[(long)((var.dt2-var.dt1)/step)+10];
		
		// Draw the normal f(t)(xy)
		ptcount=0;ou=0;
		for (u=var.dt1;u<=var.dt2;u+=step)
		{
			v=func.fa.Calculate('x',u,'t',u);
			w=func.fb.Calculate('x',u,'t',u);
			if (u==var.dt1)
			{
				ov=v,ow=w;
			}
			else
			{
				ou+=sqrt((v-ov)*(v-ov)+(w-ow)*(w-ow));
				x=(dd.ftow_xscale*u+dd.ftow_xoffset);
				y=(dd.ftow_yscale*ou+dd.ftow_yoffset);
				if (y<dd.w_min)y=dd.w_min-1;if (y>dd.w_max)y=dd.w_max+1;
				if (x<dd.w_start)x=dd.w_start-1;if (x>dd.w_end)x=dd.w_end+1;
				pointlst[ptcount].x=(long)x,pointlst[ptcount].y=(long)y;ptcount++;
			}
			ov=v,ow=w;
		}
		dd.dc->SelectObject(&dd.p_line1);	dd.dc->Polyline(pointlst,ptcount);
		delete pointlst;
	}
}

//-----------------------------------------------------------------------------

void CalcMode_ft::SetControls()
{
	vc->m_ViewSubType.ResetContent();
	show(&vc->m_DomT1);				show(&vc->m_DomT2);				show(&vc->m_SldD1);			show(&vc->m_SldD2);
	hide(&vc->m_DomT4);				hide(&vc->m_DomT3);				hide(&vc->m_SldD3);			hide(&vc->m_SldD4);
	txt(&vc->m_FPreTypeA,"x(t)=");txt(&vc->m_FPreTypeB,"y(t)=");hide(&vc->m_FPreTypeC);
	txt(&vc->m_DomainTitle,_T_FT_DT);
	show(&vc->m_FuncCA);			show(&vc->m_FuncCB);		hide(&vc->m_FuncCC);
	show(&vc->m_AddA);				show(&vc->m_AddB);			hide(&vc->m_AddC);
	show(&vc->m_FOPreTypeA);	show(&vc->m_FOPreTypeB);show(&vc->m_FOPreTypeC);		show(&vc->m_FOPx);		show(&vc->m_FOPy);
	show(&vc->m_Derive_A);		hide(&vc->m_Derive_B);	hide(&vc->m_Derive_C);
	hide(&vc->m_Int_A);			hide(&vc->m_Int_B);		hide(&vc->m_Int_C);
	show(&vc->m_Length_A);		hide(&vc->m_Length_B);	hide(&vc->m_Length_C);
	show(&vc->m_Derive2_A);		hide(&vc->m_Derive2_B);	hide(&vc->m_Derive2_C);
	show(&vc->m_Normal_A);		hide(&vc->m_Normal_B);	hide(&vc->m_Normal_C);
	show(&vc->m_ViewX);			show(&vc->m_ViewY);		show(&vc->m_ViewA);			show(&vc->m_ViewB);		show(&vc->m_ViewC);
	txt(&vc->m_FOPreTypeA,"fx(t)=");txt(&vc->m_FOPreTypeB,"fy(t)=");txt(&vc->m_FOPreTypeC,"f'(t)=");txt(&vc->m_FOPx,"x/t=");txt(&vc->m_FOPy,"y=");
	txt(&vc->m_FixParamX,"t:");
}

//-----------------------------------------------------------------------------

void CalcMode_ft::ShowNumbers(double x,double y)
{
	CVfcalcDlg::functions &func=vc->func;
	CVfcalcDlg::variables &var =vc->var;	
	CVfcalcDlg::switches  &sw  =vc->sw;
	CVfcalcDlg::colors	  &col =vc->col;
	CVfcalcDlg::DrawData  &dd  =vc->dd;

	double v,r;

	String t;
	t.FormatSpec(x,NDIGITS,var.outputmode,var.acc);t.SetWnd(&vc->m_ViewX);
	t.FormatSpec(y,NDIGITS,var.outputmode,var.acc);t.SetWnd(&vc->m_ViewY);

	if (x>=var.dt1 && x<=var.dt2)
	{

		r=func.fa.Calculate('x',x,'t',x);v=func.fb.Calculate('x',x,'t',x);		
		t.FormatSpec(v,NDIGITS,var.outputmode,var.acc);t.SetWnd(&vc->m_ViewA);
		t.FormatSpec(r,NDIGITS,var.outputmode,var.acc);t.SetWnd(&vc->m_ViewB);

		double dx=x+0.0000001;
		double n=func.fa.Calculate('x',dx,'t',dx),m=func.fb.Calculate('x',dx,'t',dx);

		if ((n-r)==0)
			t.Format("100000000000").SetWnd(&vc->m_ViewC);
		else
			t.FormatSpec(((m-v)/(n-r)),NDIGITS,var.outputmode,var.acc).SetWnd(&vc->m_ViewC);
	}

}
