/*
---------------------------------------------------------------------------
|					Function drawing mode for VFCalc
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/
 #include "Library.h"
#include "vfcalcglobals.h"

void CalcMode_3dxy::Draw()
{
	CVfcalcDlg::functions &func=vc->func;
	CVfcalcDlg::variables &var =vc->var;	
	CVfcalcDlg::switches  &sw  =vc->sw;
	CVfcalcDlg::colors	  &col =vc->col;
	CVfcalcDlg::DrawData  &dd  =vc->dd;
	
	float x,y;
	double step;
	long offset=0,ptcount=0;
	 
	// 3d surface drawing..
	// draw x from var.dt1 to var.dt2 and y from var.dt3 to var.dt4..
	// stepping by 1/(stepsize*100);
	dd.dc->FillRect(dd.DrawFrame,&dd.b_back);
	if (var.gridfac>0)
	{
		if (var.gridfac>125)
			step=1.0/125.0+0.0001;
		else
			step=1.0/var.gridfac+0.0001;

		long xsteps=(long)(1/step)+1;
		long ysteps=(long)(1/step)+1,xt,yt;
		double xstep=(var.dt2-var.dt1)*step;
		double ystep=(var.dt4-var.dt3)*step;

		CPoint  *points=new CPoint[xsteps*ysteps];


		Matrix4x4 m,mn,mt;
		Vector3 translate,out,pretranslate,in;
		
		pretranslate.Set(-(var.dt1+var.dt2)/2.0f,-(var.dt3+var.dt4)/2.0f,0               );
		   translate.Set(                      0,                      0,var.viewdistance);

		mt.SetRT(&var.viewrotate,&translate);mn.SetT(&pretranslate);
		m=mt*mn;
		dd.dc->SelectObject(&dd.t_texts);	
		float AspectX=768,AspectY=768;
		float  ScreenYCentre=(float)((dd.DrawFrame.bottom-dd.DrawFrame.top)/2+dd.DrawFrame.top),
	  		   ScreenXCentre=(float)((dd.DrawFrame.right-dd.DrawFrame.left)/2+dd.DrawFrame.left);
		float Clip_MaxX=(float)dd.DrawFrame.right+1000,
			  Clip_MinX=(float)dd.DrawFrame.left-1000,
			  Clip_MaxY=(float)dd.DrawFrame.bottom+1000,
			  Clip_MinY=(float)dd.DrawFrame.top-1000;

		for (x=var.dt1;x<var.dt2;x+=xstep)
		{
			func.fa.SetVariable('x',x);
			for(y=var.dt3;y<var.dt4;y+=ystep)
			{	
				in.z=(float)vc->clipf(func.fa.Calculate('y',y))*(float)var.fscale;
				in.x=(float)x;
				in.y=(float)y;

				// Transform and project
				out=m*in;
				if (out.z<=0.1)out.z=0.1f;
				points[offset].x=(long)clamp_f((out.x*AspectX/out.z)+ScreenXCentre,Clip_MinX,Clip_MaxX);
				points[offset].y=(long)clamp_f((out.y*AspectY/out.z)+ScreenYCentre,Clip_MinY,Clip_MaxY);

				if (!sw.ShowNum)
				{
					String a; a.FormatSpec(in.x,3,var.outputmode,var.acc);
					String b; b.FormatSpec(in.y,3,var.outputmode,var.acc);
					String c; c.FormatSpec(in.z,3,var.outputmode,var.acc);
					String s;s.Format("(%s,%s,%s)",(char *)a,(char *)b,(char *)c);
					dd.dc->TextOut(points[offset].x,points[offset].y,s,s.SLength());
				}


				offset++;		
			}
		}

		dd.dc->SelectObject(&dd.p_line1);	
		// Now draw vertical lines
		CPoint *pointlst=new CPoint[xsteps];
		for (xt=0;xt<xsteps;xt++)
			dd.dc->Polyline(&points[xt*ysteps],xsteps);
		// Now draw Horizontal lines..
		for (yt=0;yt<ysteps;yt++)
		{
			for (xt=0;xt<xsteps;xt++)
			{
				pointlst[xt]=points[xt*xsteps+yt];
			}
			dd.dc->Polyline(pointlst,xsteps);
		}

		delete pointlst;
		delete points;
	}
}

//-----------------------------------------------------------------------------------

void CalcMode_3dxy::SetControls()
{
	vc->m_ViewSubType.ResetContent();
	txt(&vc->m_DomainTitle,_T_FXY_DT);
	show(&vc->m_DomT1);				show(&vc->m_DomT2);			show(&vc->m_SldD1);			show(&vc->m_SldD2);
	show(&vc->m_DomT4);				show(&vc->m_DomT3);			show(&vc->m_SldD3);			show(&vc->m_SldD4);
	txt(&vc->m_FPreTypeA,"fz(xy)=");hide(&vc->m_FPreTypeB);		hide(&vc->m_FPreTypeC);
	show(&vc->m_FuncCA);			hide(&vc->m_FuncCB);		hide(&vc->m_FuncCC);
	show(&vc->m_AddA);				hide(&vc->m_AddB);			hide(&vc->m_AddC);
	show(&vc->m_FOPreTypeA);		hide(&vc->m_FOPreTypeB);	hide(&vc->m_FOPreTypeC);	show(&vc->m_FOPx);		show(&vc->m_FOPy);
	hide(&vc->m_Derive_A);			hide(&vc->m_Derive_B);		hide(&vc->m_Derive_C);
	hide(&vc->m_Length_A);			hide(&vc->m_Length_B);		hide(&vc->m_Length_C);
	hide(&vc->m_Derive2_A);			hide(&vc->m_Derive2_B);		hide(&vc->m_Derive2_C);
	hide(&vc->m_Normal_A);			hide(&vc->m_Normal_B);		hide(&vc->m_Normal_C);
	hide(&vc->m_Int_A);				hide(&vc->m_Int_B);			hide(&vc->m_Int_C);
	hide(&vc->m_ViewX);				hide(&vc->m_ViewY);			show(&vc->m_ViewA);			hide(&vc->m_ViewB);		hide(&vc->m_ViewC);
	txt(&vc->m_FOPreTypeA,"z=");
	txt(&vc->m_FixParamX,"x:");
}

//-----------------------------------------------------------------------------------
	
void CalcMode_3dxy::ShowNumbers(double x,double y)
{
	CVfcalcDlg::functions &func=vc->func;
	CVfcalcDlg::variables &var=vc->var;

	func.fa.SetVariable('x',var.parx);	func.fa.SetVariable('y',var.pary);
	String t;
	t.FormatSpec(func.fa.Calculate(),NDIGITS,var.outputmode,var.acc);t.SetWnd(&vc->m_ViewA);
}	
