/*
---------------------------------------------------------------------------
|	Popup menu encapsulation/simplification class
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/

#include "Library.h"
#include "PopupMenu.h"

//---------------------------------------------------------------------------

PopupTree::PopupTree(_ctree *iparent,bool inserttop):_ctree(iparent,inserttop)
{
	//_ctree::_ctree(iparent,inserttop);
	name	=" ";
	type	=POT_STRING;
	pmenu	=NULL;
	btarget	=NULL;
	ltarget	=NULL;
	smenu	=NULL;
	TreeVal	=-1;
	ValueChanged=false;
	for (long i=0;i<MAXLTDEPTH;i++)LevelTree[i]=-1;
};

//---------------------------------------------------------------------------

void PopupTree::Set(String &tn,String &n,long itype,void *data,long val)
{
	name=n;
	treename=tn;
	type=itype;
	switch(type)
	{
		case POT_BOOL:	 btarget=(bool *)data; break;
		case POT_SWITCH: ltarget=(long *)data; break;
		case POT_SUBTREE:smenu  =(CMenu *)data;break;
	};
	TreeVal=val;
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

PopupMenu::PopupMenu()
{
	IsSwitch=false;
	ltarget=NULL;
	SubDir="\\";
	IDOffset=0;
	for (long i=0;i<MAXLTDEPTH;i++)IDTree[i]=-1;
	// create a new CMenu in the root
	root.Set(String("\\"),String(""),POT_SUBTREE,AddMenu());
};

//---------------------------------------------------------------------------

PopupMenu::~PopupMenu()
{
	root.delete_children();
	menulist.Clear(true); // delete all CMenu things in menulist
}

//---------------------------------------------------------------------------

void PopupMenu::Clear()
{
	root.delete_children();
	menulist.Clear(true); // delete all CMenu things in menulist
	IsSwitch=false;
	ltarget=NULL;
	SubDir="\\";
	IDOffset=0;
	// create a new CMenu in the root
	root.Set(String("\\"),String(""),POT_SUBTREE,AddMenu());
}

//---------------------------------------------------------------------------

long PopupMenu::Show(CWnd* pParent,CPoint *origin) // shows popup menu, return result
{
	short ret;
	OpenPopupMenu(pParent,&ret,root.smenu,origin);
	if (ret>=0)
	{
		// do autobool/autoswitching
		PopupTree *node=GetNode(ret);
		if (node!=NULL)
		{
			if (node->type==POT_SWITCH) // modify switch value
			{
				if (node->MayChange)
				{
					*node->ltarget=node->TreeVal-node->loffset;
					node->ValueChanged=true;
				}
			}else
			if (node->type==POT_BOOL) // flip some bool
			{
				if (node->MayChange)
				{
					flipbool(node->btarget); 
					node->ValueChanged=true;
				}
			}
		}		
	}
	return(ret);
}

//---------------------------------------------------------------------------

String PopupMenu::GetNameTree(long val) // gets the menu tree string by return value
{
	PopupTree *tree=&root;
	while (tree!=NULL)
	{
		if (tree->TreeVal==val)return(tree->treename);
		tree=(PopupTree *)tree->walk(&root);
	}
	return("");
}

//---------------------------------------------------------------------------

PopupTree *PopupMenu::GetNode(long val) // returns node by retval
{
	PopupTree *tree=&root;
	while (tree!=NULL)
	{
		if (tree->TreeVal==val)return(tree);
		tree=(PopupTree *)tree->walk(&root);
	}
	return(NULL);
}
//---------------------------------------------------------------------------

long PopupMenu::AddItem(const char *string,void *boold,bool separator,bool valmaychange,bool IsTree,long ID,long IDlevel)
{
	// decompose the name and add subtree's if needed
	PopupTree *parent=&root;
	long     i=0,r;
	FileName f=string;
	String level;
	if (!f.IsFullPath())f=SubDir+f;
	while ((r=f.GetPathLevel(&level,i))!=GP_IS_PASTEND)
	{
		if (r==GP_IS_END || !IsTree) // is end.. so add string
		{
			if (!IsTree)level=string;
			PopupTree *newnode=new PopupTree(parent,false);
			// handle switching
			if (separator)
			{
				newnode->Set(f,level,POT_SEPARATOR); // bool menu item
				parent->smenu->AppendMenu(MF_SEPARATOR);
			}
			else
			if (boold!=NULL)
			{
				newnode->Set(f,level,POT_BOOL,boold); // bool menu item
				newnode->MayChange=valmaychange;
				parent->smenu->AppendMenu(MF_STRING|BOOLPOPUPMENUSWITCH(*((bool *)boold)),2500+IDOffset,level);
			}
			else // handle switching
			if (IsSwitch)
			{
				newnode->Set(f,level,POT_SWITCH,ltarget); // switch menu item
				parent->smenu->AppendMenu(MF_STRING,2500+IDOffset,level);
				newnode->MayChange=ValMayChange;
				switchnode=newnode;
				if (switchstart==-1)switchstart=IDOffset;
				newnode->loffset=switchstart;
			}
			else // normal
			{
				newnode->Set(f,level,POT_STRING); // just a string
				parent->smenu->AppendMenu(MF_STRING,2500+IDOffset,level);
			}
			// set ID's
			for (i=0;i<MAXLTDEPTH;i++)newnode->LevelTree[i]=IDTree[i];
			if (ID!=-1)newnode->LevelTree[IDlevel]=ID;
			newnode->pmenu=parent->smenu;
			newnode->TreeVal=IDOffset;
			IDOffset++;
			return(IDOffset);
		}
		else // find subtree with string'name'
		{
			PopupTree *chld=(PopupTree *)parent->child,*found=NULL;
			while(chld!=NULL)
			{
				// string compare
				if (chld->name.IsEqualNoCase(level)){found=chld;break;}
				chld=(PopupTree *)chld->next;
			}
			if (found!=NULL)
			{
				parent=found;
			}
			else // needs new subnode
			{
				PopupTree *newnode=new PopupTree(parent,false);
				newnode->Set(f,level,POT_SUBTREE,AddMenu());
				newnode->pmenu=parent->smenu;
				// add in parent menu tree
				parent->smenu->AppendMenu(MF_STRING|MF_POPUP,(int)newnode->smenu->m_hMenu,level);
				parent=newnode;
			}
		}
		i++;
	}
	return(-1);
}

//---------------------------------------------------------------------------

void PopupMenu::StartSwitch(long &l,bool maychange)	// indicates a switch start
{

	assert(!IsSwitch);
	IsSwitch	=true;
	ValMayChange=maychange;
	ltarget		=&l;
	switchnode	=NULL;
	switchstart	=-1;
}

//---------------------------------------------------------------------------

void PopupMenu::StopSwitch()
{
	if (IsSwitch && switchnode!=NULL) // call a set switch thing
	{
		switchnode->pmenu->CheckMenuRadioItem(switchstart+2500,switchnode->TreeVal+2500,2500+*ltarget+switchstart,MF_BYCOMMAND);
		IsSwitch=false;
	}
}

//---------------------------------------------------------------------------

void PopupMenu::SetLevelID(long ID,long level)
{
	assert(level<MAXLTDEPTH);
	IDTree[level]=ID;
}

//---------------------------------------------------------------------------

long PopupMenu::GetType(long Val,long level)
{
	assert(level<MAXLTDEPTH);
	PopupTree *node=GetNode(Val);
	if (node!=NULL)
	{
		return(node->LevelTree[level]);
	}
	return(0);
}

//---------------------------------------------------------------------------

long PopupMenu::GetType(PopupTree *node,long level)
{
	assert(level<MAXLTDEPTH);
	if (node!=NULL)
	{
		return(node->LevelTree[level]);
	}
	return(0);
}

//---------------------------------------------------------------------------

bool PopupMenu::IsOfType(long Val,long ID,long level)
{
	assert(level<MAXLTDEPTH);
	PopupTree *node=GetNode(Val);
	if (node!=NULL)
	{
		return(node->LevelTree[level]==ID);
	}
	return(false);
}

//---------------------------------------------------------------------------

bool PopupMenu::IsOfType(PopupTree *node,long ID,long level)
{
	assert(level<MAXLTDEPTH);
	if (node!=NULL)
	{
		return(node->LevelTree[level]==ID);
	}
	return(false);
}

//---------------------------------------------------------------------------
// variable helpers
bool PopupMenu::BoolChanged(bool &b)
{
	PopupTree *tree=&root;
	while (tree!=NULL)
	{
		if (tree->type==POT_BOOL)
		{
			if (tree->btarget==&b)return(tree->ValueChanged);
		}
		tree=(PopupTree *)tree->walk(&root);
	}
	return(false);
}

//---------------------------------------------------------------------------

bool PopupMenu::SwitchChanged(long &s)
{
	PopupTree *tree=&root;
	bool changed=false;
	while (tree!=NULL)
	{
		if (tree->type==POT_BOOL)
		{
			if (tree->ltarget==&s && tree->ValueChanged)changed=true;
		}
		tree=(PopupTree *)tree->walk(&root);
	}
	return(changed);
}
//---------------------------------------------------------------------------

