/*
---------------------------------------------------------------------------
|				Object Templates.. some structures derived from Pigs (Digital Dimensions BV)
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/
#ifndef _H_TEMPLATES
#define _H_TEMPLATES


//---------------------------------------------------------------------------------------------
template <class T>
class Array
{
	// Sort function and standard sort function
	typedef long		(__cdecl *CmpFunc)(const T& lhs, const T& rhs);
	static	long		__cdecl StdCmp(const T& lhs, const T& rhs)		{ if (lhs < rhs) return -1; else if (lhs == rhs) return 0; else return 1; }

public:
	// Initialization
	inline			Array()											: Len(0), Max(0), Data(NULL) { }
	inline			Array(T *e, long num)							: Len(0), Max(0), Data(NULL) { AllocBuffer(num); for (long i=0; i<num; i++) Data[i] = e[i]; }
	inline			Array(long init_size)							: Len(0), Max(0), Data(NULL) { AllocElements(init_size, init_size, 0); }
	inline			Array(const Array& other)						: Len(0), Max(0), Data(NULL) { AllocBuffer(other.Length()); for (long i=0; i<Len; i++) Data[i] = other.Data[i]; }
	inline			~Array()										{ delete[] Data; }

	// Operators
	inline bool		operator==(const Array& other)					{ if (Len != other.Len) return false; for (long i=0; i<Len; i++) if (Data[i] != other.Data[i]) return false; return true; }
	inline const Array& operator=(const Array& other)				{ if (&other == this) return *this; Len = other.Len; Max = other.Max; delete[] Data; Data = NULL; if (other.Len) { Data=new T [other.Max]; for (long i=0; i<other.Len; i++) Data[i]=other.Data[i]; } return *this; }

	// Properties													
	inline bool		IsEmpty() const									{ return Len==0; }
	inline long		MaxIndex() const								{ return Len-1; }
	inline long		MemoryUsage() const								{ return Max*sizeof(T); }
	inline long		Length() const									{ return Len; }
	inline long		Capacity() const								{ return Max; }
	inline void		SwapArray(Array &a)								{ long	tLen=Len, tMax=Max;T* tData=Data;Len=a.Len,Max=a.Max.Data=a.Data;a.Len=tLen,a.Max=tMax,a.Data=tData;}
		
	// Data access													
	inline T		*GetPtr()										{ return Data; }
	inline const T	*GetReadPtr() const								{ return Data; }
	inline T&		First()											{ return Data[0]; }
	inline const T&	First() const									{ return Data[0]; }
	inline T&		Last()											{ return Data[Len-1]; }
	inline const T&	Last() const									{ return Data[Len-1]; }

	// Simple operations											
	inline void		Swap(long pos1, long pos2)						{ bounds(pos1, 0, Len-1); bounds(pos2, 0, Len-1); if (pos1 != pos2) swap(Data[pos1], Data[pos2]); }
	inline void		Clear(bool Resize = true)						{ if (Resize) AllocBuffer(0); else Len=0; }
	inline void		Resize(long n, bool keep_contents = false)		{ AllocElements(n, n, keep_contents ? min(n, Len) : 0); }

	// Element initialization										
	inline void		Add()											{ AssureSize(Len+1); }
	inline void		Add(const T& x)									{ AssureSize(Len+1); Data[Len-1]=x; }
	inline void		Add(const Array<T>& a)							{ long l = Len; AssureSize(Len+a.Len); for (long i=0; i<a.Length(); i++) Data[l+i]=a[i]; }
	inline void		PopBack()										{ assert(Len>0); Replace(Len-1, T()); Len--; }
	inline void		Remove(long pos, long num = 1)					{ assert(num>=0); Len-=num; for (; pos<Len; pos++) Data[pos] = Data[pos+num]; }
	inline void		SwapRemove(long pos)								{ bounds(pos, 0, Len-1); Swap(pos, Len-1); Len--; }
	inline void		Insert(long pos, const T& x = T())				{ bounds(pos, 0, Len); AssureSize(Len+1); for (long i=Len-1; i>pos; i--) Data[i] = Data[i-1]; Data[pos]=x; }
	inline long		Find(const T& x) const							{ for (long i=0; i<Len; i++) { if (Data[i] == x) return i; } return -1; }
	inline bool		Contains(const T& x) const						{ return (Find(x) != -1); }

	// Sort, requires operators "<" and "==" or a sort function
	inline void		Sort(CmpFunc cmp)								{ InnerSort(0, Len-1, cmp); }
	inline void		Sort(long first=0, long last=-1, CmpFunc cmp=StdCmp)
	{
		if (last==-1) last=Len-1;
		InnerSort(first, last, cmp);
	}
	inline void		InnerSort(long first, long last, CmpFunc cmp)
	{
		if (first>=last) return;
		long split=Partition(first, last, cmp);
		InnerSort(first, split-1, cmp);
		InnerSort(split+1, last, cmp);
	}

	// Element access
	inline T&		GetElem(long pos)								{ bounds(pos, 0, Len-1); return Data[pos]; }
	inline const T&	GetElem(long pos) const							{ bounds(pos, 0, Len-1); return Data[pos]; }
	inline T&		operator[](long pos)							{ bounds(pos, 0, Len-1); return Data[pos]; }
	inline const T& operator[](long pos) const						{ bounds(pos, 0, Len-1); return Data[pos]; }

	// Serialization
	inline friend class Archive& operator<<(class Archive& file, const Array<T>& arr) { file << arr.Len; for (long i=0; i<arr.Len; i++) file << arr.Data[i]; return file; }
	inline friend class Archive& operator>>(class Archive& file, Array<T>& arr)	{ long len; file >> len; arr.Resize(len); for (long i=0; i<arr.Len; i++) file >> ((T&)arr.Data[i]); return file; }

protected:
	inline void Replace(long pos, const T& el)						{ bounds(pos, 0, Len-1); Data[pos]=el; }

	// Allocate buffer for <n> elements
	void AllocBuffer(long n)											{ long needed = n+((n+32)/8); long arr_elements = min(n, Len); assert(Len<=Max); AllocElements(n, needed, arr_elements); }
	void AllocElements(long len_els, long alloc_els, long copy_els)
	{
		T* newData = new T [alloc_els];
		for (long i=0; i<copy_els; i++) 
			newData[i] = Data[i]; 
		delete[] Data; 
		Data = newData; 
		Max  = alloc_els; 
		Len  = len_els;
	}

	// Partition part of array (for sorting)
	long Partition(long left, long right, CmpFunc cmp)
	{
		// Make the partition
		T& target = Data[right];
		long i    = left-1;
		long j    = right;
		while (1) {
			while (i < j) { if (cmp(Data[++i], target) >= 0) break; }
			while (j > i) { if (cmp(Data[--j], target) <= 0) break; }
			if (i >= j) break;
			swap(Data[i], Data[j]);
		}
		swap(Data[i], Data[right]);
		return i;
	}

	// Assure that we have enough buffer space for <n> elements and set length to <n>
	void AssureSize(long n)											{ if (Max < n) AllocBuffer(n); Len=n; }

	// Data
	long	Len, Max;
	T*		Data;
};

//---------------------------------------------------------------------------
//
// List node base type
//
struct ListNode
{
	ListNode *pPrev, *pNext;
	ListNode(ListNode *p, ListNode *n) : pPrev(p), pNext(n)			{ }
};


//---------------------------------------------------------------------------
//
// Double linked list template
//
template<class T>
class List
{
protected:
	// List node
	struct Node : ListNode
	{
		Node( ListNode* pr=NULL, ListNode* nxt=NULL) :  ListNode(pr, nxt) { }
		T		Value;										// Value of this item
	};

public:
	// List iterator
	class Iter
	{
	public:
		// Init
		inline Iter()												: pNode(NULL) { }
		inline Iter(const Iter& init)								: pNode(init.pNode) { }
		inline Iter(ListNode *init)									: pNode(init) { }
		

		// Preceeding/Succeeding elements
		inline Iter& operator++(int)								{ assert(pNode); pNode = pNode->pNext; return *this; }
		inline Iter& operator--(int)								{ assert(pNode); pNode = pNode->pPrev; return *this; }
		inline Iter& operator++()									{ return (*this)++; }
		inline Iter& operator--()									{ return (*this)--; }
		inline Iter& operator+=(int num)							{ while (num-->0) (*this)++; return *this; }
		inline Iter& operator-=(int num)							{ while (num-->0) (*this)--; return *this; }
		inline friend Iter operator+(const Iter& i, int n)			{ Iter ni(i); ni+=n; return ni; }
		inline friend Iter operator-(const Iter& i, int n)			{ Iter ni(i); ni-=n; return ni; }

		// Operators
		inline bool IsValid()										{ return (pNode != NULL); }
		inline bool operator==(const Iter& other)					{ return (pNode == other.pNode); }
		inline bool operator!=(const Iter& other)					{ return (pNode != other.pNode); }
		inline const Iter& operator=(const Iter& other)				{ pNode = other.pNode; return *this; }
		inline const Iter& operator=(const ListNode *node)			{ pNode = const_cast<ListNode *>(node); return *this; }

		// Object access
		inline T* operator->() const								{ assert(pNode); return &(Ptr()->Value); }
		inline operator T*() const									{ assert(pNode); return &(Ptr()->Value); }
		inline T& Obj() const										{ assert(pNode); return Ptr()->Value; }
		inline Node *Ptr() const									{ return (Node *)pNode; }

	protected:
		ListNode	*pNode;
	};

	// Initialization
	inline List()													: iHead(NULL), iTail(NULL), Len(0) {offset=0; }
	inline List(int num)											:iHead(NULL), iTail(NULL), Len(0) { for (int i=0; i<num; i++) Add(); offset=0;}
	inline List(List& other)										: iHead(NULL), iTail(NULL), Len(0) { for (List::Iter i=other.Begin(); i!=other.End(); i++) PushBack();offset=0; }
	inline ~List()													{ Clear(); }
	inline const List& operator=(const List& other)
	{
		// Check for assignment to self
		if (&other == this) return *this;
		
		// Wipe out, then reload
		Clear();
		for (List::Iter i=other.Begin(); i; i++) PushBack();

		return *this;
	}

	// Properties
	inline bool		IsEmpty() const									{ return Len==0; }
	inline int		Length() const									{ return Len; }

	// Simple operations
	inline void		Clear()											{ Remove(iHead, Len); assert(Len==0 && iHead.Ptr()==NULL && iTail.Ptr()==NULL); }

	// Element initialization
	inline Iter		PushFront()
	{
		ListNode *new_node = new Node(NULL, iHead.Ptr());
		if (iHead.Ptr()) iHead.Ptr()->pPrev = new_node;
		SetHead(new_node);
		Len++;

		return Iter(new_node);
	};
	inline Iter		PushBack()					
	{
		ListNode *new_node = new Node(iTail.Ptr(), NULL);
		if (iTail.Ptr()) iTail.Ptr()->pNext = new_node;
		SetTail(new_node);
		Len++;

		return Iter(new_node);
	};

	inline Iter		Add()								{ return PushBack(); }
	inline Iter		Remove(Iter pos, int num = 1, bool delete_objs = true)
	{
		offset=0;
		assert(num>=0);
		ListNode *cur = pos.Ptr();
		while (num-- > 0 && cur!=NULL)
		{
		
			ListNode *next = cur->pNext;
			ListNode *prev = cur->pPrev;

			if (next)
				next->pPrev = prev;
			else
				iTail = prev;
			
			if (prev)
				prev->pNext = next;
			else
				iHead = next;

			if (delete_objs)delete (Node *)cur;
			cur = next;
			Len--;
		}
		return Iter(cur);
	}

	//inline Iter		Find(const T& x)								{ for (Iter i=iHead; i.IsValid(); i++) if ((*i) == x) return i; return Iter(NULL);  }
	//inline bool		Contains(const T& nx)							{ return (Find(x) != Iter(NULL)); }

	inline Iter		InsertBefore(Iter pos)
	{
		ListNode *next = pos.Ptr();
		ListNode *prev = next->pPrev;
		ListNode *new_node = new Node(prev, next);

		next->pPrev = new_node;
		if (prev)
			prev->pNext = new_node;
		else
			iHead = new_node;
		
		Len++;
		return Iter(new_node);
	}

	inline Iter		InsertAfter(Iter pos)
	{
		ListNode *prev = pos.Ptr();
		ListNode *next = prev->pNext;
		ListNode *new_node = new Node(prev, next);

		prev->pNext = new_node;
		if (next)
			next->pPrev = new_node;
		else
			iTail = new_node;

		Len++;
		return Iter(new_node);
	}

	inline Iter &MoveAfter(Iter &what,Iter &where)							
	{
		if (what.IsValid() && iHead!=iTail && what!=where) // is movable
		{
			if (!where.IsValid()) // move to top?
			{
				// untie old place
				if (what!=iHead)// don't move if this is the first
				{
					if (iTail==what)iTail=what-1; // set new tail
					if(	what.Ptr()->pPrev !=NULL) what.Ptr()->pPrev->pNext=what.Ptr()->pNext;
					if( what.Ptr()->pNext !=NULL) what.Ptr()->pNext->pPrev=what.Ptr()->pPrev;
					
					// tie in new place
					Iter n=iHead;
					iHead=what;
					n.Ptr()->pPrev=what.Ptr();
					what.Ptr()->pNext=n.Ptr();
					what.Ptr()->pPrev=NULL; // is head
				}
			}
			else if (where!=(what-1)) // not the same, not top
			{
				
				if(iHead==what)iHead=(what+1);
				if(iTail==what)iTail=(what-1);
				if(	what.Ptr()->pPrev !=NULL) what.Ptr()->pPrev->pNext=what.Ptr()->pNext;
				if( what.Ptr()->pNext !=NULL) what.Ptr()->pNext->pPrev=what.Ptr()->pPrev;
				// tie in new place

				if(where==iTail) // its the last one
				{
					iTail=what;
					what.Ptr()->pNext=NULL;
				}
				else
				{
					where.Ptr()->pNext->pPrev=what.Ptr();
					what.Ptr()->pNext=where.Ptr()->pNext;
				}
				what.Ptr()->pPrev=where.Ptr();
				where.Ptr()->pNext=what.Ptr();
			}
		}
		return what;
	}

	inline Iter& MoveUp(Iter &what)									{ Iter d=what;d-=2;MoveAfter(what,d);return what;}
	inline Iter& MoveDown(Iter &what)								{ Iter d=what;d++ ;MoveAfter(what,d);return what;}

	// Iterators
	inline const Iter&	Head() const								{ return iHead; }
	inline const Iter&	Tail() const								{ return iTail; }
	inline const Iter&	Begin() const								{ return iHead; }
	inline const Iter	End() const									{ return Iter(NULL); }

	// Data access
	inline T&		First()											{ return iHead.Obj(); }
	inline const T&	First() const									{ return iHead.Obj(); }
	inline T&		Last()											{ return iTail.Obj(); }
	inline const T&	Last() const									{ return iTail.Obj(); }

	// Searching
	inline Iter 	Find(T *o)										{Iter i=iHead;while (i.IsValid()){if (&i.Obj()==o)return i;i++;}Iter t;return t;}
	inline long	    EnumI(Iter &i)									{if (!i.IsValid())return(-1);long c=0;Iter j=iHead;while(j.IsValid()){if (j.Ptr()==i.Ptr())return(c);c++;j++;};return(-1);}
	inline Iter		EnumN(long n)									{if (n<0)return(iHead);Iter i=iHead;while (i.IsValid()){if(n==0)return i;n--;i++;};return(iTail);}

	long			offset; // offset counter in list

protected:
	// List management
	inline void		SetHead(ListNode *new_head)						{ iHead=new_head; if (iTail.Ptr() == NULL) iTail=iHead; }
	inline void		SetTail(ListNode *new_tail)						{ iTail=new_tail; if (iHead.Ptr() == NULL) iHead=iTail; }

	// List data
	Iter	iHead,	iTail;											// Head and tail iterators
	int		Len;													// Objects in list
};

//---------------------------------------------------------------------------------------------

///
// Double linked list template, but now with pointers
//
template<class T>
class PtrList
{
protected:
	// List node
	struct Node : ListNode
	{
		Node(T *val, ListNode* pr=NULL, ListNode* nxt=NULL) : Value(val), ListNode(pr, nxt) { }
		T	 *Value;										// Value of this item
		inline void DeleteObj()								{if (Value!=NULL)delete Value;Value=NULL;}	
	};

public:
	// List iterator
	class Iter
	{
	public:
		// Init
		inline Iter()												: pNode(NULL) { }
		inline Iter(const Iter& init)								: pNode(init.pNode) {}
		inline Iter(ListNode *init)									: pNode(init) {}

		// Preceeding/Succeeding elements
		inline Iter& operator++(int)								{ assert(pNode); pNode = pNode->pNext; return *this; }
		inline Iter& operator--(int)								{ assert(pNode); pNode = pNode->pPrev; return *this; }
		inline Iter& operator++()									{ return (*this)++; }
		inline Iter& operator--()									{ return (*this)--; }
		inline Iter& operator+=(int num)							{ while (num-->0) (*this)++; return *this; }
		inline Iter& operator-=(int num)							{ while (num-->0) (*this)--; return *this; }
		inline friend Iter operator+(const Iter& i, int n)			{ Iter ni(i); ni+=n; return ni; }
		inline friend Iter operator-(const Iter& i, int n)			{ Iter ni(i); ni-=n; return ni; }

		// Operators
		inline bool IsValid()										{ return (pNode != NULL); }
		inline bool operator==(const Iter& other)					{ return (pNode == other.pNode); }
		inline bool operator!=(const Iter& other)					{ return (pNode != other.pNode); }
		inline const Iter& operator=(const Iter& other)				{ pNode = other.pNode; return *this; }
		inline const Iter& operator=(const ListNode *node)			{ pNode = const_cast<ListNode *>(node); return *this; }

		// Object access
		inline void DeleteObj()										{ assert(pNode); Ptr()->DeleteObj();}
		inline T* operator->() const								{ assert(pNode); return (Ptr()->Value); }
		inline operator T*() const									{ assert(pNode); return (Ptr()->Value); }
		inline T& rObj() const										{ assert(pNode); return *Ptr()->Value; }
		inline T* Obj() const										{ assert(pNode); return (Ptr()->Value); }
		inline Node *Ptr() const									{ return (Node *)pNode; }
		inline T *SetObj(T *ptr) const								{ T* p=Ptr()->Value;Ptr()->Value=ptr;return (p); }
	
	protected:
		ListNode	*pNode;
	};
	// Initialization
	inline PtrList()												: iHead(NULL), iTail(NULL), Len(0) {offset=0;  }
	inline PtrList(T *e, int num)									: iHead(NULL), iTail(NULL), Len(0) { offset=0; for (int i=0; i<num; i++) Add(&e[i]); }
	inline PtrList(PtrList<T> &other)								: iHead(NULL), iTail(NULL), Len(0) {offset=0;  for (PtrList::Iter i=other.Begin(); i!=other.End(); i++) PushBack(i); }
	inline ~PtrList()												{Clear();}
	inline const PtrList& operator=(const PtrList<T> &other)
	{
		// Check for assignment to self
		if (&other == this) return *this;
		
		// Wipe out, then reload
		Clear();
		for (PtrList::Iter i=other.Begin(); i; i++) PushBack(i.Obj());

		return *this;
	}

	// Properties
	inline bool		IsEmpty() const									{ return Len==0; }
	inline long		Length() const									{ return Len; }

	// Simple operations
	inline void		Clear(bool delobjs=false)						{ Remove(iHead, Len,delobjs); assert(Len==0 && iHead.Ptr()==NULL && iTail.Ptr()==NULL); }

	// Element initialization
	inline Iter		PushFront(T *val)
	{
		//assert(val==NULL);
		//if (val==NULL)val=new T;
		ListNode *new_node = new Node(val, NULL, iHead.Ptr());
		if (iHead.Ptr()) iHead.Ptr()->pPrev = new_node;
		SetHead(new_node);
		Len++;
		return Iter(new_node);
	};
	inline Iter		PushBack(T *val)					
	{
		//assert(NULL);
		//if (val==NULL)val=new T;
		ListNode *new_node = new Node(val, iTail.Ptr(), NULL);
		if (iTail.Ptr()) iTail.Ptr()->pNext = new_node;
		SetTail(new_node);
		Len++;

		return Iter(new_node);
	};

	inline Iter		Add(T *val=NULL)								{ return PushBack(val); }

	inline Iter		MoveToFront(Iter pos)
	{
		T *item = (T*)pos;
		Remove(pos, 1, false);
		PushFront(item);
		return iHead;
	}

	inline Iter		MoveToBack(Iter pos)
	{
		T *item = (T*)pos;
		Remove(pos, 1, false);
		PushBack(item);
		return iTail;
	}

	inline Iter &MoveAfter(Iter &what,Iter &where)							
	{
		if (what.IsValid() && iHead!=iTail && what!=where) // is movable
		{
			if (!where.IsValid()) // move to top?
		{
				// untie old place
				if (what!=iHead)// don't move if this is the first
				{
					if (iTail==what)iTail=what-1; // set new tail
					if(	what.Ptr()->pPrev !=NULL) what.Ptr()->pPrev->pNext=what.Ptr()->pNext;
					if( what.Ptr()->pNext !=NULL) what.Ptr()->pNext->pPrev=what.Ptr()->pPrev;
					
					// tie in new place
					Iter n=iHead;
					iHead=what;
					n.Ptr()->pPrev=what.Ptr();
					what.Ptr()->pNext=n.Ptr();
					what.Ptr()->pPrev=NULL; // is head
				}
			}
			else if (where!=(what-1)) // not the same, not top
			{
				
				if(iHead==what)iHead=(what+1);
				if(iTail==what)iTail=(what-1);
				if(	what.Ptr()->pPrev !=NULL) what.Ptr()->pPrev->pNext=what.Ptr()->pNext;
				if( what.Ptr()->pNext !=NULL) what.Ptr()->pNext->pPrev=what.Ptr()->pPrev;
				// tie in new place

				if(where==iTail) // its the last one
				{
					iTail=what;
					what.Ptr()->pNext=NULL;
				}
				else
				{
					where.Ptr()->pNext->pPrev=what.Ptr();
					what.Ptr()->pNext=where.Ptr()->pNext;
				}
				what.Ptr()->pPrev=where.Ptr();
				where.Ptr()->pNext=what.Ptr();
			}
		}
		return what;
	}

	inline Iter& MoveUp(Iter &what)									{ Iter d=what;d-=2;MoveAfter(what,d);return what;}
	inline Iter& MoveDown(Iter &what)								{ Iter d=what;d++ ;MoveAfter(what,d);return what;}


	inline Iter		Remove(Iter pos, int num = 1, bool delete_objs = false)
	{
		offset=0;
		assert(num>=0);
		ListNode *cur = pos.Ptr();
		while (num-- > 0)
		{
			ListNode *next = cur->pNext;
			ListNode *prev = cur->pPrev;

			if (next)
				next->pPrev = prev;
			else
				iTail = prev;
			
			if (prev)
				prev->pNext = next;
			else
				iHead = next;

			if (delete_objs)((Node *)cur)->DeleteObj();
			delete cur;
			cur = next;
			Len--;
		}
		return Iter(cur);
	}

	inline Iter		Find(T *x)								{ for (Iter i=iHead; i.IsValid(); i++) if (i.Obj() == x) return i; return Iter(NULL);  }
	inline long	    EnumI(Iter &i)							{if (!i.IsValid())return(-1);long c=0;Iter j=iHead;while(j.IsValid()){if (j.Ptr()==i.Ptr())return(c);c++;j++;};return(-1);}
	inline Iter		EnumN(long n)							{if (n<0)return(iHead);Iter i=iHead;while (i.IsValid()){if(n==0)return i;n--;i++;};return(iTail);}

	inline bool		Contains(T *x)							{ return (Find(x) != Iter(NULL)); }

	inline Iter		InsertBefore(Iter pos, T *x=NULL)
	{
		if (x==NULL)x=new T();
		ListNode *next = pos.Ptr();
		ListNode *prev = next->pPrev;
		ListNode *new_node = new Node(x, prev, next);

		next->pPrev = new_node;
		if (prev)
			prev->pNext = new_node;
		else
			iHead = new_node;
		
		Len++;
		return Iter(new_node);
	}

	inline Iter		InsertAfter(Iter pos, T *x=NULL)
	{
		if (x==NULL)x=new T();
		ListNode *prev = pos.Ptr();
		ListNode *next = prev->pNext;
		ListNode *new_node = new Node(x, prev, next);

		prev->pNext = new_node;
		if (next)
			next->pPrev = new_node;
		else
			iTail = new_node;

		Len++;
		return Iter(new_node);
	}

	// Iterators
	inline const Iter	Enum(long pos) const						{ Iter b=iHead;b+=pos;return b;}
	inline const Iter&	Head() const								{ return iHead; }
	inline const Iter&	Tail() const								{ return iTail; }
	inline const Iter&	Begin() const								{ return iHead; }
	inline const Iter	End() const									{ return Iter(NULL); }

	// Data access
	inline T* 		First()											{ return iHead.Obj(); }
	inline const T*	First() const									{ return iHead.Obj(); }
	inline T*		Last()											{ return iTail.Obj(); }
	inline const T*	Last() const									{ return iTail.Obj(); }

	// serialization
	inline void Write(rArchive arc)										{Iter i=Head();while (i.IsValid()){arc<<true;arc<<*i.Obj();i++;}arc<<false;}
	inline void Read(rArchive arc)										{bool next;arc>>next;while (next){arc>>*Add().Obj();arc>>next;};}
	inline friend rArchive operator<<(rArchive arc, PtrList<T> &v){	v.Write(arc);return(arc);}
	inline friend rArchive operator>>(rArchive arc, PtrList<T> &v){	v.Read(arc); return(arc);}

	long			offset; // offset counter in list

protected:
	// List management
	inline void		SetHead(ListNode *new_head)						{ iHead=new_head; if (iTail.Ptr() == NULL) iTail=iHead; }
	inline void		SetTail(ListNode *new_tail)						{ iTail=new_tail; if (iHead.Ptr() == NULL) iHead=iTail; }

	// List data
	Iter	iHead,	iTail;											// Head and tail iterators
	int		Len;													// Objects in list
};

//---------------------------------------------------------------------------------------------

//
// fixed pointer stack
//
#define MAXSTACKSIZE 4096
template<class T>
class PtrStack
{
	public:
	PtrStack(){t=0;};
	
	inline T   *Push(T *obj)        {if (t>=MAXSTACKSIZE)log("Internal Stackoverflow in PtrStack!!");else{stack[t]=obj;t++;};return(obj);}
	inline T   *PushBottom(T *obj) {memmove(&stack[1],&stack[0],(t+1)*sizeof(long));stack[0]=obj;t++;return(obj);};
	inline T   *Pop()			    {if (t>0){t--;return(stack[t]);};return(NULL);}
	inline T   *PopBottom()	        {if (t>0){T *obj=stack[0];memmove(&stack[0],&stack[1],t*sizeof(long));t--;return(obj);};return(NULL);}
	inline T   *PushPop(T *obj)     {T *out=Pop();Push(obj);return(out);}
	inline T   *Get(long i)			{ if (i<t && i>=0)return(stack[i]);else return(NULL);}
	inline T   *Set(long i,T *obj)	{ if (i<t && i>=0)stack[i]=obj;return(obj);}
	inline void Clear()		        {t=0;}
	inline long Size()		    	{return(t);}
	inline bool HasElements()	    {return (Size()>0);};

	bool Lock; // multitasking stack lock
	long flag; // flag the stack with something

protected:
	long t;
	T *stack[MAXSTACKSIZE]; 
};


//---------------------------------------------------------------------------------------------

#define STACKSTEPPING 10
template<class T>
class PtrStackFl // flexible stack.. (size stepped by STACKSTEPPING)
{
	public:
	PtrStackFl(){t=0;length=STACKSTEPPING;stack=new long[length];};
	~PtrStackFl(){delete stack;};
	
	inline T   *Push(T *obj)           {if (t>=length)StackResize(length+STACKSTEPPING);stack[t]=(long)obj;t++;return(obj);}
	inline T   *PushChk(T *obj)        {for (long i=0;i<t;i++)if (stack[i]==(long)obj)return(obj);if (t>=length)StackResize(length+STACKSTEPPING);stack[t]=(long)obj;t++;return(obj);}
	inline T   *Pop()			       {if (t>0){t--;return((T *)stack[t]);};return(NULL);}
	inline T   *Get(long i)			   {if (i<t && i>=0)return((T *)stack[i]);else return(NULL);}
	inline T   *Set(long i,T *obj)	   {if (i<t && i>=0)stack[i]=(long)obj;return(obj);}
	inline void Remove(long i)		   {if (i<t && i>=0){memmove (&stack[i],&stack[i+1],(t-i)*sizeof(long));t--;}} 
	inline void Clear()		           {t=0;}
	inline void Copy(PtrStackFl &o)    {if (&o!=this)for (long i=0;i<o.Size();i++)Push((T *)o.stack[i]);}
	inline void CopyChk(PtrStackFl &o) {if (&o!=this)for (long i=0;i<o.Size();i++){PushChk((T *)o.stack[i]);}}
	inline long Size()		    	   {return(t);}
	inline bool HasElements()	       {return (Size()>0);};
	inline void StackFit()			   {StackResize(t+1);}
	inline void StackResize(long l)    {if (l>0){if (l<=t)Clear();long *source=stack;stack=new long[l];memcpy(stack,source,t*sizeof(long));delete source;}	}

	bool Lock; // multitasking stack lock

protected:
	long length;
	long t;
	long *stack; 
};

//
// Circular stack
//

//---------------------------------------------------------------------------------------------

template<class T>
class PtrLoopStack // looped read/write buffer
{
	public:
	PtrLoopStack()					   {len=0;readindex=0;writeindex=0;};
	inline void  Add(T *obj)           {stack[len]=obj;len++;}
	inline void  Add(T *obj,long n)    {while (n>0){stack[len]=&obj[n-1];n--;len++;};}
	inline void  Remove()			   {if (len>0)len--;};
	inline void  Flush()			   {lock(&Lock);readindex=0;writeindex=0;unlock(&Lock);}
	inline T  *GetRead()      
	{
		if (len==1)return(stack[0]);
		lock(Lock);

		if (writeindex%len>readindex%len)writeindex%=len,readindex%=len;

		if (readindex<writeindex-1) 
		{
			T *obj=stack[readindex%len];
			readindex++;
			unlock(Lock);
			return(obj);
		}
		unlock(Lock);
		return(NULL);
	}
	inline T  *GetWrite()
	{
		if (len==1)return(stack[0]);
		lock(Lock);
		if (writeindex%len>readindex%len)writeindex%=len,readindex%=len;

		if ( writeindex-readindex<len )// fits in wrap buffer?
		{
			T *obj=stack[writeindex%len];
			writeindex++;
			unlock(Lock);
			return(obj);
		}
		unlock(Lock);
		return(NULL);
	}

	MultithreadLock Lock; // multitasking stack lock

	inline long GetReadIndex() {return(readindex);}
	inline long GetWriteIndex(){return(writeindex);}
	inline long GetLength()	   {return(len);}
protected:
	long len;
	long readindex;
	long writeindex;
	T *stack[MAXSTACKSIZE]; 
};

//---------------------------------------------------------------------------------------------

#endif