/*
---------------------------------------------------------------------------
|						Generic tool functions
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/

// And just a very frequently used function.. copies a string, and returns the new string pointer (MUST BE DELETED BY YOU)

#ifndef _H_TOOLS
#define _H_TOOLS

// Macros
#define INITIALIZE_START(initname)	struct _init_##initname { _init_##initname(){ DLLRESOURCES;	// switches to dll bitmaps..etc.	
#define INITIALIZE_END(initname)	}}init_##initname;

//------------------------------------------------------------
// Bound checking macro

#define bounds(x,y,z) {assert(x>=y);assert(x<=z);}

//------------------------------------------------------------

inline float fix(float a,float b){	if (b==0)return(a);long r;if (a>0)r=(long)(((a)/b)+0.5),a=(float)r;	else r=(long)(((a)/b)-0.5),a=(float)r;	return(a*b);}
inline long	 cast32bit(double c) {	if (c>4294967295.0 || c<-4294967295.0) return(0xffffffff);return((long)c);}

//------------------------------------------------------------
inline long	 snappow2(long v,long min=4,long max=1024){if(v>max)return max;if(v<min)return min;long s=max;while((float)s/(float)v>=1.5 && s>min)s=s>>1;return s;}
inline float limitvar(float v,float min,float max,bool clipmin=true,bool clipmax=true){	if (clipmin==true)if (v<min)v=min;if (clipmax==true)if (v>max)v=max;	return(v);}
inline short  sabs(short s){if (s<0)return(-s);return(s);}
inline float  flabs(float s){if (s<0)return(-s);return(s);}
inline double dabs(double s){if (s<0)return(-s);return(s);}

//------------------------------------------------------------

inline void  flipbool(bool *target){if (*target==true)*target=false;else *target=true;}
inline bool  IsNonzero(long v){if (v!=0)return(true);else return(false);}
//------------------------------------------------------------

// Standard tiny inline functions..

template<class T> const T& max(const T& x, const T& y)				{ return (x>y) ? x : y; }
template<class T> const T& min(const T& x, const T& y)				{ return (x<y) ? x : y; }

inline float  fmax3(float a,float b,float c)						{ float x=(a>b) ? a : b;return (x>c) ? x : c; }
inline float  fmin3(float a,float b,float c)						{ float x=(a<b) ? a : b;return (x<c) ? x : c; }

// max(x, y)
inline double dmax(double  x, double y)								{ return (x>y) ? x : y; }
//inline float max(double x, float  y)								{ return (x>y) ? x : y; }

// min(x, y)
inline double dmin(double  x, double y)								{ return (x<y) ? x : y; }
//inline float min(double x, float  y)								{ return (x<y) ? x : y; }

// abs(x)
//inline float abs(float x)											{ if (x<0.0) return (-x); else return (x); }

// sign(x)
inline float sign(float x)											{ return ((x<0.0f) ? -1.0f : 1.0f); }

// clamp(val, low, high)
inline float  clamp_f(float  val, float  low, float  high)				{ if (val<low) return low; else if (val>high) return high; else return val; }
inline double clamp_d(double val, double low, double high)				{ if (val<low) return low; else if (val>high) return high; else return val; }
inline long   clamp_l(long    val, long    low, long    high)				{ if (val<low) return low; else if (val>high) return high; else return val; }
inline short  clamp_s(short  val, short  low, short  high)				{ if (val<low) return low; else if (val>high) return high; else return val; }

// inrange(val, low, high) - check to see if (val>=low && val<=high)
inline bool inrange(float  val, float   low, float high)				{ return ((val>=low) && (val<=high)); }
inline bool inrange(double val, double  low, double high)				{ return ((val>=low) && (val<=high)); }
inline bool inrange(int    val, int     low, int    high)				{ return ((val>=low) && (val<=high)); }
inline bool inrange_c(char   val, char    low, char   high)				{ return ((val>=low) && (val<=high)); }
inline bool inset(const char *set,char c)								{for (long i=0;set[i]!='0';i++)if(set[i]==c)return(true);return(false);}
// swap (a, b)
inline void  swapd(double *a, double *b){	double temp;temp = *a;*a = *b;*b = temp;}
inline void  swapf(double &a, double &b){	float  temp;temp =  a; a =  b; b = temp;}
inline void  swapl(long   *a, long   *b){	long   temp;temp = *a;*a = *b;*b = temp;}
inline void  swaps(short  *a, short  *b){	short  temp;temp = *a;*a = *b;*b = temp;}
inline void  swapc(char   *a, char   *b){	char   temp;temp = *a;*a = *b;*b = temp;}

inline long  b2l(bool b){ if (b)return(0);return(1);}
inline bool  l2b(long b){ if (b==0)return(true);return(false);}

// Check if character is of a certain type
inline bool  isSpace(char c)	{ return (c==' ' || c=='\t' || c=='\n' || c=='\r'); }
inline bool  isAlpha(char c)	{ return (inrange_c(c, 'a', 'z') || inrange_c(c, 'A', 'Z')); }
inline bool  isNumeric(char c)	{ return inrange_c(c, '0', '9'); }
inline bool  isLower(char c)	{ return inrange_c(c, 'a', 'z'); }
inline bool  isUpper(char c)	{ return inrange_c(c, 'A', 'Z'); }
inline char  toLower(char c)	{ return isUpper(c) ? (c-('A'-'a')) : c; }
inline char  toUpper(char c)	{ return isLower(c) ? (c-('a'-'A')) : c; }

// Byte, word, dword, qword element access
inline byte  gLoByte(word w)								{ return (w & 0xFF); }
inline byte  gHiByte(word w)								{ return (w >> 8); }
inline word  gLoWord(dword d)								{ return ((word)(d & 0xFFFF)); }
inline word  gHiWord(dword d)								{ return ((word)(d >> 16)); }
inline dword gHiDWord(qword q)								{ return ((dword)(q >> 32)); }
inline dword gLoDWord(qword q)								{ return ((dword)q); }

inline word  gMakeWord(byte low, byte hi)					{ return (low | (hi<<8)); }
inline dword gMakeDWord(word low, word hi)					{ return (low | (hi<<16)); }
inline dword gMakeDWord(byte b0, byte b1, byte b2, byte b3)	{ return (b0 | (b1<<8) | (b2<<16) | (b3<<24)); }
inline qword gMakeQWord(word w0, word w1, word w2, word w3)	{ return (((qword)w0) | (((qword)w1)<<16) | (((qword)w2)<<32) | (((qword)w3)<<48)); }
inline qword gMakeQWord(dword low, dword hi)				{ return ((qword)low | (((qword)hi)<<32)); }
inline qword gMakeQWord(byte b0, byte b1, byte b2, byte b3, byte b4, byte b5, byte b6, byte b7)  { return (((qword)b0) | (((qword)b1)<<8) | (((qword)b2)<<16) | (((qword)b3)<<24) | (((qword)b4)<<32) | (((qword)b5)<<40) | (((qword)b6)<<48) | (((qword)b7)<<56)); }

//Random number generator stuff..

extern  char   ctrand();
extern  short  strand();
extern  long   ltrand();
extern  double ftrand();
extern  float  ffrand(float feed);
extern  float  flrand(long feed);
extern  long   llrand(long  feed);
extern  long	  lfrand(float feed);

//#define ma

#endif